/*
 * dafit.c: PT VibXpert specific tweaks
 *
 * Copyright (C) 2001  Erik Mouw (J.A.K.Mouw@its.tudelft.nl)
 * Copyright (C) 2001,2002,2003  Stefan Eletzhofer 
 *                               (stefan.eletzhofer@eletztrick.de)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ident "$Id: dafit.c,v 1.5 2003/09/04 17:39:00 russd Exp $"

/**********************************************************************
 * includes
 */

#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/main.h>
#include <blob/arch.h>
#include <blob/errno.h>
#include <blob/error.h>
#include <blob/util.h>
#include <blob/reboot.h>
#include <blob/serial.h>
#include <blob/flash.h>
#include <blob/init.h>
#include <blob/command.h>
#include <blob/uucodec.h>
#include <blob/time.h>
#include <blob/partition.h>
#include <blob/led.h>

#include <blob/arch.h>
#include <blob/proc/sa1111.h>

#include <blob/generic_io.h>

#include <blob/pcmcia.h>
#include <blob/cf.h>
#include <blob/ide.h>
#include <blob/tar.h>
#include <net/ether.h>

/**********************************************************************
 * defines
 */

/* more readable IMHO */
#define MEM( x )        (*((u32 *)(x)))
#define SET(reg,bit)	((reg) |= (1<<(bit)))
#define RST(reg,bit)	((reg) &= ~(1<<(bit)))

#define DAFIT_DEBUG 1

#if DAFIT_DEBUG
#	define DBG( x, args... )	if ( dbg>=x ) printf( args )
#else
#	define DBG( x, args... )
#endif

/**********************************************************************
 * globals
 */

extern blob_status_t blob_status;

/**********************************************************************
 * module globals
 */

#if DAFIT_DEBUG
static int dbg = 1;
#else
static int sys3_dbg = 0;
#endif

/* flash descriptor for System3 flash. */
/* System 3 uses 2xINTEL e28F640 Chips */
static const flash_descriptor_t dafit_flash_descriptors[] =
{
	{
		size: 2 * 128 * 1024,
		num: 64,
		lockable: 1
	},
	{
		/* NULL block */
	},
};

/* default partition table for DAFIT

System3 flash layout
~~~~~~~~~~~~~~~~~~~~

16Mb Flash
256k erase size

0x00000000 BLOB		256k
0x00040000 CONFIG	256k
0x00080000 KERNEL	1024k
0x00180000 INITRD	1536k
0x00300000 ROOT		6M
0x00900000 DATA		7M
0x01000000 total	16M

Note:
	- dafit partitions are named like the files they're flashed
	  with. This is to keep updating easy: Just loop through all
	  files that are on our update media and flash them on the partition
	  named like the file.

 */
static const blob_partition_t dafit_default_partition_table[] = {
	{
		/* start of table */
		magic:	BLOB_DEFAULT_PART_TABLE_MAGIC,
		next:	sizeof(blob_partition_t),
		offset:	0x00000000,	/* absolute base address */
		size:	16 * 1024 * 1024 /* total size of the flash */
	},
	{
		/* blob itself */
		magic:		BLOB_PART_VALID_MAGIC,
		next:		sizeof(blob_partition_t),
		offset:		BLOB_FLASH_BASE,
		size:		BLOB_FLASH_LEN,
		name:		"blob",
		mem_base:	BLOB_RAM_BASE
	},
	{
		/* system config area */
		magic:	BLOB_PART_VALID_MAGIC,
		next:	sizeof(blob_partition_t),
		offset:	CONFIG_FLASH_BASE,
		size:	CONFIG_FLASH_LEN,
		name:	"config.tar",
		flags:	0
	},
	{
		/* kernel */
		magic:		BLOB_PART_VALID_MAGIC,
		next:		sizeof(blob_partition_t),
		offset: 	KERNEL_FLASH_BASE,
		size:		KERNEL_FLASH_LEN,
		name:		"zImage",
		flags:		BLOB_PART_FLAG_LOAD | BLOB_PART_FLAG_EXEC,
		mem_base:	KERNEL_RAM_BASE,
		entry_point:	KERNEL_RAM_BASE
	},
	{
		/* ramdisk */
		magic:		BLOB_PART_VALID_MAGIC,
		next:		sizeof(blob_partition_t),
		offset:		RAMDISK_FLASH_BASE,
		size:		RAMDISK_FLASH_LEN,
		name:		"initrd.gz",
		flags:		BLOB_PART_FLAG_LOAD,
		mem_base:	RAMDISK_RAM_BASE
	},
	{
		/* root (cramfs) */
		magic:		BLOB_PART_VALID_MAGIC,
		next:		sizeof(blob_partition_t),
		offset:		ROOTFS_FLASH_BASE,
		size:		ROOTFS_FLASH_LEN,
		name:		"cramfs",
		flags:		BLOB_PART_FLAG_CRAMFS,
	},
	{
		/* data (cramfs) */
		magic:		BLOB_PART_VALID_MAGIC,
		next:		sizeof(blob_partition_t),
		offset:		DATAFS_FLASH_BASE,
		size:		DATAFS_FLASH_LEN,
		name:		"data.img",
		flags:		BLOB_PART_FLAG_CRAMFS,
	},
	{
		/* last entry */
		magic:	BLOB_PART_LAST_MAGIC
	}
};

/**********************************************************************
 * static functions
 */
static char module_version[] = "$Id: dafit.c,v 1.5 2003/09/04 17:39:00 russd Exp $";

/*********************************************************************
 * init_dafit_flash - inits system 3 flash driver
 */
static void init_dafit_flash_driver(void)
{
	flash_descriptors = dafit_flash_descriptors;
	flash_driver = &intel32_flash_driver;
}
__initlist(init_dafit_flash_driver, INIT_LEVEL_DRIVER_SELECTION);

/*********************************************************************
 * dafit_init_hardware - inits system 3 LL hardware stuff
 */
static void dafit_init_hardware(void)
{
	/* init on-board CPLD */
	MEM(DAFIT_CTRL_0) = 0x03;
	MEM(DAFIT_CTRL_1) = 0x00;
	MEM(DAFIT_CTRL_2) = 0x00;

	/* initialize GPIO states */
	GPDR = 0x0ffdbb80;
	GPSR |=
		GPIO_GPIO13 |
		GPIO_GPIO15 |
		GPIO_GPIO26;
	GPCR |= 
		GPIO_GPIO7 |
		GPIO_GPIO8 |
		GPIO_GPIO9 |
		GPIO_GPIO11 |
		GPIO_GPIO12 |
		GPIO_GPIO16 |
		GPIO_GPIO18 |
		GPIO_GPIO19 |
		GPIO_GPIO20 |
		GPIO_GPIO21 |
		GPIO_GPIO22 |
		GPIO_GPIO23 |
		GPIO_GPIO24 |
		GPIO_GPIO25;

	/* activate SYSCLCK for keyboard controller */
	RST( MEM(TUCR), 31 );
	GAFR |= GPIO_32_768kHz;
	GPDR |= GPIO_32_768kHz;

	/* 3.6864 MHz SYSCLK */
	TUCR = TUCR_3_6864MHz;

	msleep( 1 );

	MEM(DAFIT_CTRL_1) = 0x04;

	/* tweak blob config */
	blob_status.boot_delay = 1;

	/* select drivers */
	reboot_driver = &sa11x0_reboot_driver;
	serial_driver = &sa11x0_serial_driver;
	led_driver = &sa11x0_gpio_led_driver;
	ether_driver = &smc9196_ether_driver;
	timer_driver = &intelarm_timer_driver;
}
__initlist(dafit_init_hardware, INIT_LEVEL_DRIVER_SELECTION);

/**********************************************************************
 * dafit default partition table
 */
static void dafit_set_partition_table(void)
{
	/* the default partition table */
	default_partition_table = dafit_default_partition_table;
	
	/* we don't know where the flash partition table will live, so
	 * put the pointer at 0x0000000 and let the partition table
	 * parser figure out.
	 */
	flash_partition_table = (blob_partition_t*)0x00000000;
}

__initlist(dafit_set_partition_table, INIT_LEVEL_OTHER_STUFF);
/* FIXME: need to think if this is the correct init level, we might
 * want to have this before the param list stuff starts running */


/*****************************************************************
 * sysver - print system version
 */
static int sysver_cmd( int argc, char *argv[] )
{
	printf( "version: '%s'\n", module_version );
	switch ( PT_ID_BOARDID ) {
		case PT_BOARD_ALPHA:
			printf( "ALPHA Board" );
			break;
		case PT_BOARD_SYS3:
			printf( "SYSTEM 3 Board" );
			break;
		case PT_BOARD_DAFIT:
			printf( "DAFIT Board" );
			break;
		default:
			printf( "UNKNOWN Board (ID 0x%X)", PT_ID_BOARDID );
			break;

	}
	printf( " revision 0x%02X.\n", PT_ID_REVISION ); 
	return 0;
}
static char sysver_help[] = "print BLOB and dafit version number\n";
__commandlist(sysver_cmd, "sysver", sysver_help);

/**********************************************************************
 * setdbg - set debug level
 */
int setdbg_cmd( int argc, char *argv[] )
{
	extern int cf_dbg_set(int);
	int ret;
	u32 lvl;

	if ( argc < 2 ) return -EINVAL;

	ret = strtou32(argv[1], &lvl);
	if ( ret < 0 ) return -EINVAL;

	dbg = lvl;

#if defined(CONFIG_PCMCIA_SUPPORT)
	if ( lvl>5 ) {
		cf_dbg_set( 5 );
		tar_dbg_set( 5 );
	} else {
		cf_dbg_set( 0 );
		tar_dbg_set( 0 );
	}

	if ( lvl>10 ) {
		io_dbg_set( 5 );
		pcmcia_dbg_set( 5 );
	} else {
		io_dbg_set( 0 );
		pcmcia_dbg_set( 0 );
	}
#endif

	printf( "setdbg: debug level set to %d\n", dbg );
	return 0;
}
static char setdbg_help[] = "setdbg level\n";
__commandlist(setdbg_cmd, "setdbg", setdbg_help);


#if defined(CONFIG_PCMCIA_SUPPORT)
/*****************************************************************
 * pcmcia_init - initialize PCMCIA
 */
int pcmcia_init()
{
	printf( "PT VibXpert CF init\n" );
	return 0;
}
#endif



/*****************************************************************
 * misc utility funcs
 */

/*****************************************************************
 * manually reference flash and download commands until they
 * are in libblob.
 */
#if 0
extern int cf_test_module( int argc, char *argv[] );
extern char cf_help[];
__commandlist(cf_test_module, "cftest", cf_help);
#endif

#if 1
extern int fwrite_cmd(int argc, char *argv[]);
extern char fwrite_help[];
__commandlist(fwrite_cmd, "fwrite", fwrite_help);

extern int ferase_cmd(int argc, char *argv[]);
extern char ferase_help[];
__commandlist(ferase_cmd, "ferase", ferase_help);
#endif

extern int dlfile_cmd(int argc, char *argv[]);
extern char dlfile_help[];
__commandlist(dlfile_cmd, "dlfile", dlfile_help);


