/*-------------------------------------------------------------------------
 * Filename:      flash-commands.c
 * Version:       $Id: flash-commands.c,v 1.7 2003/03/11 19:36:25 seletz Exp $
 * Copyright:     Copyright (C) 2002, Hewlett-Packard Company
 * Author:        Christopher Hoover <ch@hpl.hp.com>
 * Description:   Flash commands for blob
 * Created at:    Tue Jan 15 02:27:30 UTC 2002
 *-----------------------------------------------------------------------*/
/*
 * flash-commands.c: Flash commands for blob
 *
 * Copyright (C) 2002  Hewlett-Packard Company
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ident "$Id: flash-commands.c,v 1.7 2003/03/11 19:36:25 seletz Exp $"

#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/arch.h>
#include <blob/errno.h>
#include <blob/error.h>
#include <blob/util.h>
#include <blob/serial.h>
#include <blob/flash.h>
#include <blob/partition.h>
#include <blob/main.h>

#include <blob/command.h>

int parse_partition(char *s, u32 **addr, u32 *length)
{
	const blob_partition_t *p = pt_find_by_name(s);
	
	if (p == 0)
		return -EINVAL;

	*addr = (u32 *) pt_flash_start(p);
	*length = p->size;

	return 0;
}



static int Flash(int argc, char *argv[])
{
	int rv;
	u32 *src;
	u32 *dst;
	u32 numBytes = 0;
	u32 maxSize = 0;
	u32 nwords;
	char *dest_part;
	block_source_t type = fromFlash;
	
	if(argc < 2)
		return -ENOPARAMS;

	if ( argc>2 ) {
		dest_part = argv[2];
	} else {
		dest_part = argv[1];
	}

	rv = parse_partition(dest_part, &dst, &maxSize);
	if (rv < 0) {
		printerror(rv, dest_part);
		return 0;
	}

	if(strncmp(argv[1], "blob", 5) == 0) {
		src = (u32 *)BLOB_RAM_BASE;
		numBytes = blob_status.blobSize;
		type = blob_status.blobType;
	} else if(strncmp(argv[1], "param", 6) == 0) {
		src = (u32 *)PARAM_RAM_BASE;
		numBytes = blob_status.paramSize;
		type = blob_status.paramType;
	} else if(strncmp(argv[1], "kernel", 7) == 0) {
		src = (u32 *)KERNEL_RAM_BASE;
		numBytes = blob_status.kernelSize;
		type = blob_status.kernelType;
	} else if(strncmp(argv[1], "ramdisk", 8) == 0) {
		src = (u32 *)RAMDISK_RAM_BASE;
		numBytes = blob_status.ramdiskSize;
		type = blob_status.ramdiskType;
	} else {
		printerror(EINVAL, argv[1]);
		return 0;
	}

	if(type == fromFlash) {
		printerrprefix();
		printf("%s not downloaded\n", argv[1]);
		return -EINVAL;
	}

	if(numBytes > maxSize) {
		printerrprefix();
		printf("image too large for flash: 0x%08x > 0x%08x\n",
		       (unsigned int)numBytes, (unsigned int)maxSize);

		return -ETOOLONG;
	}

	nwords = (numBytes + sizeof(u32) - 1) / sizeof(u32);

	printf("Saving %s to flash\n", argv[1]);

	flash_write_region(dst, src, nwords);

	return 0;
}

static char flashhelp[] = "flash {blob|param|kernel|ramdisk} [part]\n"
"Write <argument> image to flash using partition <part>. If <part> is omitted\n"
"assume <part>==<argument>\n";

__commandlist(Flash, "flash", flashhelp);



static int Lock(int argc, char *argv[])
{
	int rv;
	u32 *addr, length;

	if(argc != 2)
		return -ENOPARAMS;
	
	if (!flash_driver->lock_block) {
		printerror(EINVAL, argv[1]);
		return 0;
	}

	rv = parse_partition(argv[1], &addr, &length);
	if (rv < 0) {
		printerror(rv, argv[1]);
		return 0;
	}

	rv = flash_lock_region(addr, length >> 2);
	if (rv < 0) {
		printerror(rv, argv[1]);
		return 0;
	}

	return 0;
}

static char lockhelp[] = "lock partition\n"
"Lock <partition> region of flash\n";

__commandlist(Lock, "lock", lockhelp);




static int Unlock(int argc, char *argv[])
{
	int rv;
	u32 *addr, length;

	if(argc != 2)
		return -ENOPARAMS;
	
	if (!flash_driver->unlock_block) {
		printerror(EINVAL, argv[1]);
		return 0;
	}

	rv = parse_partition(argv[1], &addr, &length);
	if (rv < 0) {
		printerror(rv, argv[1]);
		return 0;
	}

	rv = flash_unlock_region(addr, length >> 2);
	if (rv < 0) {
		printerror(rv, argv[1]);
		return 0;
	}

	return 0;
}

static char unlockhelp[] = "unlock partition\n"
"Unlock <partition> region of flash\n";

__commandlist(Unlock, "unlock", unlockhelp);



static int Erase(int argc, char *argv[])
{
	int rv;
	u32 *addr, length;

	if(argc != 2)
		return -ENOPARAMS;
	
	if (!flash_driver->erase) {
		printerror(EINVAL, argv[1]);
		return 0;
	}

	rv = parse_partition(argv[1], &addr, &length);
	if (rv < 0) {
		printerror(rv, argv[1]);
		return 0;
	}

	rv = flash_erase_region(addr, length >> 2);
	if (rv < 0) {
		printerror(rv, argv[1]);
		return 0;
	}

	return 0;
}

static char erasehelp[] = "erase partition\n"
"Erase <partition> region of flash\n";

__commandlist(Erase, "erase", erasehelp);
