/*
 * linux.c: support functions for booting a kernel
 *
 * $Id: linux.c,v 1.18 2003/09/04 17:27:33 russd Exp $
 *
 * Copyright (C) 2001 2002  Erik Mouw (J.A.K.Mouw@its.tudelft.nl)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ident "$Id: linux.c,v 1.18 2003/09/04 17:27:33 russd Exp $"

#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/linux.h>
#include <blob/arch.h>
#include <blob/command.h>
#include <blob/init.h>
#include <blob/main.h>
#include <blob/memory.h>
#include <blob/util.h>
#include <blob/errno.h>
#include <blob/error.h>

#include <asm-arm/setup.h>

#ifdef ATAG_BOOT_INFO
#define POWERUP_REASON_ENABLE
#else
#undef POWERUP_REASON_ENABLE
#endif //ATAG_BOOT_INFO

#define HW_VERSION_ENABLE
#define PRODUCT_HW_VERSION_BASE (*(unsigned int *)(0x5c010000 - 4))
#define PRODUCT_MACH_BASE (*(unsigned int *)(0x5c010000 - 8))

static void setup_start_tag(void);
static void setup_memory_tags(void);
static void setup_commandline_tag(int argc, char *argv[]);
//static void setup_ramdisk_tag(void);
static void setup_initrd_tag(void);
static void setup_end_tag(void);
#ifdef POWERUP_REASON_ENABLE
static void setup_powerupreason_tag(void);
#endif	//POWERUP_REASON_ENABLE
#ifdef HW_VERSION_ENABLE
static void setup_hwversion_tag(void);
#endif	//HW_VERSION_ENABLE


static struct tag *params;


static void (*theKernel)(int zero, int arch, u32 params) = 
//#ifdef KERNEL_XIP_BASE
//	(void (*)(int, int, u32)) KERNEL_XIP_BASE;
//#else
	(void (*)(int, int, u32)) KERNEL_RAM_BASE;
//#endif

static void (*ramKernel)(int zero, int arch, u32 params) =
	(void (*)(int, int, u32)) KERNEL_RAM_BASE;

#ifndef CONFIG_NAND_SUPPORT
static void (*xipKernel)(int zero, int arch, u32 params) =
	(void (*)(int, int, u32)) KERNEL_XIP_BASE;
#endif 

#ifdef MACH_TYPE_ENABLE
int mach_type = ARCH_NUMBER;

static int cmd_machtype(int argc, char *argv[])
{
  int ret = 0;
  unsigned int new_machtype;

  if( argc != 2 ){
    printf( "Mach type = %d\n", mach_type );
    return 0;
  }

  ret = strtou32( argv[1], &new_machtype );
  if ( ret < 0 ) {
    return -1;
  }

  printf( "Old mach type = %d, ",mach_type );
  mach_type = new_machtype;
  printf( "New mach type = %d\n ",mach_type );
 
  return 0;
}

static char machtypehelp[] = "machtype\n"
"Set mach_type of Linux\n";
__commandlist(cmd_machtype, "machtype", machtypehelp);

#endif	//MACH_TYPE_ENABLE
static int boot_linux(int argc, char *argv[])
{

	setup_start_tag();
	setup_memory_tags();
	setup_commandline_tag(argc, argv);
	if (blob_status.load_ramdisk) {
		setup_initrd_tag();
//		setup_ramdisk_tag();
	}
#ifdef POWERUP_REASON_ENABLE
	setup_powerupreason_tag();
#endif	//POWERUP_REASON_ENABLE
#ifdef HW_VERSION_ENABLE
	setup_hwversion_tag();
#endif	//HW_VERSION_ENABLE
	setup_end_tag();

	/* we assume that the kernel is in place */
#ifdef MACH_TYPE_ENABLE
	printf("\nStarting kernel at 0x%08lx mach_type=%d ...\n\n", (unsigned long)ramKernel, mach_type );
#else
	printf("\nStarting kernel at 0x%08lx...\n\n", (unsigned long)ramKernel);
#endif	//MACH_TYPE_ENABLE
	//serial_flush_output();

	/* disable subsystems that want to be disabled before kernel boot */
	exit_subsystems();

	//i = 0xffffffff;
	//while(i--);

	/* start kernel */
#ifdef MACH_TYPE_ENABLE
	ramKernel(0, mach_type, BOOT_PARAMS);
#else
	ramKernel(0, ARCH_NUMBER, BOOT_PARAMS);
#endif	//MACH_TYPE_ENABLE

	printf("Hey, the kernel returned! This should not happen.\n");

	return 0;
}

static char boothelp[] = "boot [kernel options]\n"
"Boot Linux with optional kernel options\n";

__commandlist(boot_linux, "boot", boothelp);


#ifndef CONFIG_NAND_SUPPORT
static int boot_linux_xip(int argc, char *argv[])
{
  int i;

	setup_start_tag();
	setup_memory_tags();
	setup_commandline_tag(argc, argv);
	if (blob_status.load_ramdisk) {
		setup_initrd_tag();
		setup_ramdisk_tag();
	}
	setup_end_tag();

	/* we assume that the kernel is in place */
	printf("\nStarting kernel at 0x%08x...\n\n", xipKernel);
	//serial_flush_output();

	/* disable subsystems that want to be disabled before kernel boot */
	exit_subsystems();

	//i = 0xffffffff;
	//while(i--);

	/* start kernel */
	xipKernel(0, ARCH_NUMBER, BOOT_PARAMS);

	printf("Hey, the kernel returned! This should not happen.\n");

	return 0;
}

static char xipboothelp[] = "xipboot [kernel options]\n"
"Boot Linux with optional kernel options\n";

__commandlist(boot_linux_xip, "xipboot", xipboothelp);
#endif


static int cmd_kernel(int argc, char *argv[])
{
	unsigned int adr;
	int ret;

	if ( argc < 2 ) goto out;

	ret = strtou32(argv[1], &adr);
	if ( ret < 0 ) return -EINVAL;

	theKernel = (void (*)(int, int, u32)) adr;

out:
	printf("Kernel @ 0x%08lX Now\n", (unsigned long)theKernel);

	return 0;
}

static char kernelhelp[] = "kernel hex_adr\n"
"Change the Linux kernel base\n";

__commandlist(cmd_kernel, "kernel", kernelhelp);


static void setup_start_tag(void)
{
	params = (struct tag *)BOOT_PARAMS;

	params->hdr.tag = ATAG_CORE;
	params->hdr.size = tag_size(tag_core);

	params->u.core.flags = 0;
	params->u.core.pagesize = 0;
	params->u.core.rootdev = 0;

	params = tag_next(params);
}


static void setup_memory_tags(void)
{
	int i;

	for(i = 0; i < NUM_MEM_AREAS; i++) {
		if(memory_map[i].used) {
			params->hdr.tag = ATAG_MEM;
			params->hdr.size = tag_size(tag_mem32);

			params->u.mem.start = memory_map[i].start;
			params->u.mem.size = memory_map[i].len;

			params = tag_next(params);
		}
#if defined( NEPONSET )
		/* When neponset board is present dont report
		 * the 2nd ram area to the kernel. This area
		 * is on the neponset board and would require
		 * NUMA support in the kernel.*/
		break;
#endif
	}
}


static void setup_commandline_tag(int argc, char *argv[])
{
	char *p;
	int i;

	/* initialise commandline */
	params->u.cmdline.cmdline[0] = '\0';

	/* copy default commandline from parameter block */
	if(blob_status.cmdline[0] != '\0')
		strlcpy(params->u.cmdline.cmdline, blob_status.cmdline,
			COMMAND_LINE_SIZE);

	/* copy commandline */
	if(argc >= 2) {
		p = params->u.cmdline.cmdline;

		for(i = 1; i < argc; i++) {
			strlcpy(p, argv[i], COMMAND_LINE_SIZE);
			p += strlen(p);
			*p++ = ' ';
		}

		p--;
		*p = '\0';

		/* technically spoken we should limit the length of
		 * the kernel command line to COMMAND_LINE_SIZE
		 * characters, but the kernel won't copy longer
		 * strings anyway, so we don't care over here.
		 */
	}

	if(strlen(params->u.cmdline.cmdline) > 0) {
		params->hdr.tag = ATAG_CMDLINE;
		params->hdr.size = (sizeof(struct tag_header) +
				    strlen(params->u.cmdline.cmdline) + 1 + 4) >> 2;

		params = tag_next(params);
	}
}


static int ramdisk_len = 0;
int get_ramdisk_len( void )
{
	return ramdisk_len;
}

void set_ramdisk_len( int len )
{
	if( len > (int)( RAMDISK_FLASH_LEN - sizeof( ramdisk_len ) ) )
		ramdisk_len = 0;
	else
		ramdisk_len = len;

	printf( "ramdisk_len = %d\n", ramdisk_len );
}

static void setup_initrd_tag(void)
{
	if( !get_ramdisk_len() )
		return;

	/* an ATAG_INITRD node tells the kernel where the compressed
	 * ramdisk can be found. ATAG_RDIMG is a better name, actually.
	 */
	params->hdr.tag = ATAG_INITRD;
	params->hdr.size = tag_size(tag_initrd);

	params->u.initrd.start = RAMDISK_RAM_BASE + 0x20000000;
	params->u.initrd.size = get_ramdisk_len();	//RAMDISK_FLASH_LEN;

	params = tag_next(params);

	/* Uhoh, we got it all wrong. ATAG_INITRD takes a *virtual* address
	 * not a physical address, so it nicely breaks on PXA machines.
	 * This was fixed in linux-2.4.18-rmk6, which added the
	 * ATAG_INITRD2 node. This one really takes a physical address.
	 * Pre 2.4.18-rmk6 SA11x0 kernels will just ignore this node and
	 * skip over it. Post 2.4.18-rmk6 kernels will parse both the
	 * INITRD and the INITRD2 node, but the latter will supersede
	 * the former and everything will be all right. -- Erik
	 */

#ifndef ATAG_INITRD2 /* Quick fix for headers without it */
#define ATAG_INITRD2 0x54420005
#endif

	params->hdr.tag = ATAG_INITRD2;
	params->hdr.size = tag_size(tag_initrd);

	params->u.initrd.start = RAMDISK_RAM_BASE + 0x20000000;
	params->u.initrd.size = get_ramdisk_len();	//RAMDISK_FLASH_LEN;

	params = tag_next(params);
}

#if 0
static void setup_ramdisk_tag(void)
{
	if( !get_ramdisk_len() )
		return;

	/* an ATAG_RAMDISK node tells the kernel how large the
	 * decompressed ramdisk will become.
	 */
	params->hdr.tag = ATAG_RAMDISK;
	params->hdr.size = tag_size(tag_ramdisk);

	params->u.ramdisk.start = 0;
	params->u.ramdisk.size = RAMDISK_SIZE;
	params->u.ramdisk.flags = 1;	/* automatically load ramdisk */

	params = tag_next(params);
}
#endif
#ifdef HW_VERSION_ENABLE
extern int calc_hardware_ver( void ); 

unsigned int get_hw_version_id(void)
{
	unsigned int product_mach_type;
	unsigned int product_hw_version;
#if defined(TAVOR_EVB2)
	product_hw_version = 0xFFFF & calc_hardware_ver();
	product_mach_type = 0xFFFF & PRODUCT_MACH_BASE;
#else
	product_hw_version = 0xFFFF & PRODUCT_HW_VERSION_BASE;
	product_mach_type = 0xFFFF & PRODUCT_MACH_BASE;
#endif
	printf("hwversion:0x%x, mach:0x%x\n", product_hw_version, 
			product_mach_type);

	return product_hw_version;
}

static int set_product_hw_version(unsigned int version)
{
	PRODUCT_HW_VERSION_BASE = 0xFFFF & version;
}

static int cmd_set_hwversion(int argc, char *argv[])
{
	int ret = 0;
	unsigned int new_hw_version;

	if( argc != 2 ){
		get_hw_version_id();
	}
	else{
		ret = strtou32( argv[1], &new_hw_version );
		if( ret < 0 )
			return -1;

		set_product_hw_version( new_hw_version );
		get_hw_version_id();
	}

	return 0;
}

static char hw_versionhelp[] = "show hw version\n"
"set hw version\n";
__commandlist(cmd_set_hwversion, "hwversion", hw_versionhelp);

static void setup_hwversion_tag(void)
{
	params->hdr.tag = ATAG_REVISION;
	params->hdr.size = tag_size(tag_revision);

	params->u.revision.rev = get_hw_version_id();//calc_hardware_ver();

	params = tag_next(params);
}
#endif	//HW_VERSION_ENABLE

#ifdef POWERUP_REASON_ENABLE

inline unsigned get_boot_info( void )
{
	return (int)PSPR; 
}

static void setup_powerupreason_tag( void )
{
	params->hdr.tag = ATAG_BOOT_INFO;
	params->hdr.size = tag_size(tag_boot_info);

	params->u.boot_info.boot_info = get_boot_info();

	params = tag_next(params);
}
#endif //POWERUP_REASON_ENABLE

static void setup_end_tag(void)
{
	params->hdr.tag = ATAG_NONE;
	params->hdr.size = 0;
}

#ifdef POWERUP_REASON_ENABLE

unsigned int get_powerup_reason( void )
{
	unsigned int powerup_reason;
	powerup_reason = (PSPR & 0x0000ffff);
	printf( "powerup_reason = %d\n", powerup_reason );

	return (powerup_reason);
}

void set_powerup_reason( unsigned int new )
{
	new &= 0x0000ffff;
	new |= (PSPR & 0xffff0000);
	printf( "boot_info = %d\n", new );

	PSPR = new;
}


static int cmd_powerup_reason(int argc, char *argv[])
{
	int ret = 0;
	unsigned int new_powerup_reason;

	if( argc != 2 ){
		get_powerup_reason();
	}
	else{
		get_powerup_reason();

		ret = strtou32( argv[1], &new_powerup_reason );
		if( ret < 0 )
			return -1;

		set_powerup_reason( new_powerup_reason );
		get_powerup_reason();
	}

	return 0;
}

static char powerup_reasonhelp[] = "powerup_reason\n"
"Show power up reason\n";
__commandlist(cmd_powerup_reason, "powerup_reason", powerup_reasonhelp);

#endif	//POWERUP_REASON_ENABLE


