/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *(C) Copyright 2006 Marvell International Ltd.  
 * All Rights Reserved 
 */

/*++

Module Name:  $Workfile: xllp_ov2620_hw.c $

Abstract:  
     contains all hardware related functions for OV2620

Notes:Only valid for processor code named Monahans.

--*/

#include "xllp_camera_os_depend.h"
#include "xllp_camera.h"
#include "xllp_gpio_plat.h"
#include "xllp_gpio_proc.h"
#include "xllp_ost.h"
#include "xllp_ov2620_hw.h"
#include "xllp_serialization.h"

/***********************************************************************
*  Register Settings
***********************************************************************/
const static XLLP_UINT8_T ov2620InitSetting[]=
{
    OV2620_COMH,     0x80,        // Initiates system reset
    OV2620_COMF,     0x05,        
    OV2620_CLKRC,    0x81,
    OV2620_COMH,     0x24,
    OV2620_FVO,      0x0c,
    OV2620_CHLF,     0x37,
    OV2620_VCHG,     0x4c,
    OV2620_R39,      0xf3,
    OV2620_AEW,      0x80,
    OV2620_AEB,      0x70,
    OV2620_COMI,     0x8f,
    OV2620_REGEND,   0x00
};

/***********************************************************************
*  Register Settings, Get from OmniVision
***********************************************************************/
static XLLP_UINT8_T UXGA[] =
{
    OV2620_COMH,    0x80,
    OV2620_COMF,    0x05,
    OV2620_COMH,    0x04, 
    OV2620_FVO,     0x0c, 
    OV2620_CHLF,    0x07, 
    OV2620_VCHG,    0x4c,
    OV2620_R39,     0xf3,
    OV2620_CLKRC,   0x81,
    OV2620_COML,    0x00,
    OV2620_FRARL,   0x00,
    OV2620_REGEND,  0x00
};

static XLLP_UINT8_T SVGA[] =
{
    OV2620_COMH,    0x80,
    OV2620_COMF,    0x05,
    OV2620_COMH,    0x44, 
    OV2620_FVO,     0x0c, 
    OV2620_CHLF,    0x07, 
    OV2620_VCHG,    0x4c,
    OV2620_R39,     0xf3,
    OV2620_CLKRC,   0x81,
    OV2620_COML,    0x10,
    OV2620_FRARL,   0x00,
    OV2620_HREFST,  0x3c,
    OV2620_HREFEND, 0xa0,
    OV2620_REGEND,  0x00
};

static XLLP_UINT8_T CIF[] =
{
    OV2620_COMH,    0x80,
    OV2620_COMF,    0x05,
    OV2620_COMH,    0x24, 
    OV2620_FVO,     0x0c, 
    OV2620_CHLF,    0x07, 
    OV2620_VCHG,    0x4c,
    OV2620_R39,     0xf3,
    OV2620_CLKRC,   0x81,
    OV2620_COML,    0x40,
    OV2620_FRARL,   0x10,
    OV2620_REGEND,  0x00
};

const static XLLP_UINT8_T gSensorSlaveAddr = 0x30;

/***********************************************************************
*  Private/helper api
***********************************************************************/
#ifdef XLLP_DEBUG_PARAM_CHECK
static XLLP_STATUS_T PrvGetRegValue( XLLP_UINT8_T *regP, XLLP_UINT8_T regAddr, XLLP_UINT8_T *regValueP )
{
    unsigned int index = 0;
    XLLP_UINT8_T curReg = 0;
    
    while( curReg < OV2620_REGEND )
    {
        curReg = regP[index << 1];
        if( curReg == regAddr )
        {
            *regValueP = regP[(index << 1) + 1];
            return XLLP_STATUS_SUCCESS;
        }    
        index ++;
    } 
        
    return XLLP_STATUS_FAILURE;

}

static XLLP_STATUS_T PrvSetRegValue( XLLP_UINT8_T *regP, XLLP_UINT8_T regAddr, XLLP_UINT8_T regValue )
{
    unsigned int index = 0;
    XLLP_UINT8_T curReg = 0;
    
    while( curReg < OV2620_REGEND )
    {
        curReg = regP[index << 1];
        if( curReg == regAddr )
        {
            regP[(index << 1) + 1] = regValue;
            return XLLP_STATUS_SUCCESS;
        }    
        index ++;
    } 
        
    return XLLP_STATUS_FAILURE;

}
#endif

/***********************************************************************
*  Sensor read/write 
***********************************************************************/
static int PrvReadSensorReg( const XLLP_UINT8_T subAddress, XLLP_UINT8_T *bufP )
{
	XLLP_UINT8_T buffer[1];
	int status;
	buffer[0] = subAddress;

	XllpLock(XLLP_RESOURCE_I2C);
	status = OS_I2CMasterWriteData( gSensorSlaveAddr, buffer, 1);
	if (!status) {
		status = OS_I2CMasterReadData( gSensorSlaveAddr, buffer, 1);
		*bufP = buffer[0];
	}
	XllpUnlock(XLLP_RESOURCE_I2C);
	return status;
}

static int PrvWriteSensorReg( const XLLP_UINT8_T subAddress, XLLP_UINT8_T *bufP )
{
	XLLP_UINT8_T buffer[2];
	int status;
	buffer[0] = subAddress;
	buffer[1] = *bufP;

	XllpLock(XLLP_RESOURCE_I2C);
	status = OS_I2CMasterWriteData( gSensorSlaveAddr, buffer, 2);
	XllpUnlock(XLLP_RESOURCE_I2C);
	return status;
}

static int PrvRMWSensorReg( const XLLP_UINT8_T subAddress, XLLP_UINT8_T *bufP, XLLP_UINT8_T andMask, XLLP_UINT8_T orMask ) 
{
    int status;
    status = PrvReadSensorReg( subAddress, bufP );
    if (!status) {
        *bufP &= andMask;
        *bufP |= orMask;
        status = PrvWriteSensorReg( subAddress, bufP );
    }
    return status;
}

int OV2620ReadSensorReg( const XLLP_UINT8_T subAddress, XLLP_UINT8_T *bufP )
{
    return PrvReadSensorReg(subAddress, bufP);
}

int OV2620WriteSensorReg( const XLLP_UINT8_T subAddress, XLLP_UINT8_T *bufP )
{
    return PrvWriteSensorReg(subAddress, bufP);
}

XLLP_STATUS_T OV2620SetRegs( const XLLP_UINT8_T *regP )
{
    XLLP_UINT32_T    curReg = 0;
    XLLP_STATUS_T    status = XLLP_STATUS_SUCCESS;
        
    // The list is a register number followed by the value.
    while( regP[curReg << 1] < OV2620_REGEND )
    {
        XLLP_UINT8_T regVal = regP[(curReg << 1) + 1];

        status = (PrvWriteSensorReg( regP[curReg << 1], &regVal ) == 0) ?
                 XLLP_STATUS_SUCCESS : XLLP_STATUS_FAILURE;
        
        if( curReg == 0 )
            OV2620Wait( 5 );
            
        curReg++;
    }
        
    return status;
}

XLLP_STATUS_T OV2620ReadAllRegs( XLLP_UINT8_T *bufP, XLLP_UINT32_T numRegs )
{
    XLLP_UINT32_T    curReg;
    
    for( curReg = 0; curReg < numRegs; curReg++, bufP++ )
        PrvReadSensorReg( (XLLP_UINT8_T)curReg, bufP );


    return XLLP_STATUS_SUCCESS;
}

/***********************************************************************
*  Power & Reset
***********************************************************************/
void OV2620PowerDown(unsigned int gpio_reg_base, unsigned int ost_reg_base, XLLP_UINT8_T powerMode )
{
    // OV2620 PWRDWN, 0 = NORMAL, 1=POWER DOWN
    if( powerMode == XLLP_CAMERA_POWER_OFF )
        XllpGpioSetOutputLevel( (P_XLLP_GPIO_T)gpio_reg_base, XLLP_GPIO_CAMERA_HI_PWDN, XLLP_HI );
    else
        XllpGpioSetOutputLevel( (P_XLLP_GPIO_T)gpio_reg_base, XLLP_GPIO_CAMERA_HI_PWDN, XLLP_LO );
    
   XllpOstDelayMilliSeconds( (P_XLLP_OST_T)ost_reg_base, 100 );
}

void OV2620Reset( )
{
    OV2620SetRegs(ov2620InitSetting);
    return;
}

void OV2620Wait( int ms )
{
    OS_Delay( ms );
}


/***********************************************************************
*  Settings
***********************************************************************/
int OV2620VersionRevision(XLLP_UINT8_T * pCmRevision, XLLP_UINT8_T *pSensorRevision)
{
    PrvReadSensorReg( OV2620_PIDH, pCmRevision );
    PrvReadSensorReg( OV2620_PIDL, pSensorRevision );
    return XLLP_STATUS_SUCCESS;
}

void OV2620SetHSYNC()
{
    XLLP_UINT8_T    val;
    
    // Makes HREF become HSYNC
    PrvReadSensorReg( OV2620_COMK, &val );
    val |= 0x40;
    PrvWriteSensorReg( OV2620_COMK, &val );
}

void OV2620AutoFunctionOn()
{
    XLLP_UINT8_T val;
    PrvReadSensorReg( OV2620_COMI, &val );
    val |= 0x07;    // don't disturb AWB
    PrvWriteSensorReg( OV2620_COMI, &val );
}

void OV2620AutoFunctionOff()
{
    XLLP_UINT8_T val;
    PrvReadSensorReg( OV2620_COMI, &val );
    val &= ~0x07;    // don't disturb AWB
    PrvWriteSensorReg( OV2620_COMI, &val );
}


/***********************************************************************
*  Viewfinder, still 
***********************************************************************/
int OV2620ViewfinderOn()
{
    XLLP_UINT8_T    com3;
    
    PrvReadSensorReg( OV2620_COMD, &com3 );
    com3 &= ~0x01;
    PrvWriteSensorReg( OV2620_COMD, &com3 );
        
    return OV_ERR_NONE;
}


int OV2620ViewfinderOff()
{
    XLLP_UINT8_T    com3;
    
    PrvReadSensorReg( OV2620_COMD, &com3 );
    com3 |= 0x01;
    PrvWriteSensorReg( OV2620_COMD, &com3 );
    
    return OV_ERR_NONE;
}


int OV2620HaltVideoOutput()
{
    XLLP_UINT8_T    com3;
    
    // Set the camera to only output 1 frame.
    PrvReadSensorReg( OV2620_COMD, &com3 );
    com3 |= 1;
    PrvWriteSensorReg( OV2620_COMD, &com3 );
    
    return OV_ERR_NONE;
}

int OV2620ResumetoFullOutputMode()
{
    XLLP_UINT8_T    mode;
    
    // Output still frames continuously
    // Turn off single capture mode COM3.
    PrvRMWSensorReg( OV2620_COMD, (&mode), ((XLLP_UINT8_T) ~1), 0 );
    return OV_ERR_NONE;
}

int OV2620GetSingleImage()
{
    XLLP_UINT8_T    mode;
    
    PrvRMWSensorReg( OV2620_COMD, &mode, (XLLP_UINT8_T) ~1, 1 );
    return OV_ERR_NONE;
}

/***********************************************************************
*  Format 
***********************************************************************/
XLLP_STATUS_T OV2620SetFormat(XLLP_UINT32_T captureWidth, 
                              XLLP_UINT32_T captureHeight, 
                              XLLP_UINT32_T *winStartX, 
                              XLLP_UINT32_T *winStartY, 
                              XLLP_UINT32_T *winEndX, 
                              XLLP_UINT32_T *winEndY)
{
    OV2620_MODE mode;
    unsigned short hStart, hEnd;
    unsigned short vStart, vEnd;
    unsigned char regVal;
	unsigned short hMiddle, hProtection, vProtection;
	int bConfigROI = 1;
	
	hProtection = 1;
	vProtection =1;

    //
    // let the sensor work on proper mode
    //
	if((captureWidth <= 403) && (captureHeight <= 301)){
        mode = OV2620_CIF;
    } else if((captureWidth <= 807) && (captureHeight <= 603)) {
        mode = OV2620_SVGA;
    } else if((captureWidth <= 1615) && (captureHeight <= 1207)) {
        mode = OV2620_UXGA;
    } else {
        return XLLP_STATUS_WRONG_PARAMETER;
    }
 
    if (mode == OV2620_CIF){
        OV2620SetRegs(CIF);
    }else if (mode == OV2620_SVGA){
       OV2620SetRegs(SVGA);
    }else{
       OV2620SetRegs(UXGA);
    }

    //
    // set cropping window
    //
    if (mode == OV2620_CIF) {
        captureWidth *= 2;
    }
    
   
	if (mode == OV2620_CIF){
		hMiddle = 888;
        hStart = (unsigned short)(hMiddle - captureWidth/2);
		if(1 == hStart%2) {
			hStart++;
		}
		hEnd = (unsigned short)(hStart+captureWidth+hProtection);
		vStart = 1;
	    vEnd = (unsigned short)(vStart + captureHeight/2)+vProtection;

		if((hStart<=488)||(hEnd>=1288)||(vEnd>=151)) {
			bConfigROI = 0;
		}
		
    }else if (mode == OV2620_SVGA){
   		hMiddle = 880;
       	hStart = (unsigned short)(hMiddle - captureWidth/2);
		if(1 == hStart%2) {
			hStart++;
		}
		hEnd = (unsigned short)(hStart+captureWidth+hProtection);
		vStart = 1;
	    vEnd = (unsigned short)(vStart + captureHeight/2)+vProtection;

		if((hStart<=480)||(hEnd>=1280)||(vEnd>=303)) {
			bConfigROI = 0;
		}
    
    }else{
    	hMiddle = 1140;
       	hStart = (unsigned short)(hMiddle - captureWidth/2);
		if(1 == hStart%2) {
			hStart++;
		}
		hEnd = (unsigned short)(hStart+captureWidth+hProtection);
        vStart = 1;
	    vEnd = (unsigned short)(vStart + captureHeight/2)+vProtection;

		if((hStart<=340)||(hEnd>=1940)||(vEnd>=603)) {
			bConfigROI = 0;
		}
    }

    
    if(bConfigROI){
	    // set Horizontal Window Start
	    regVal = hStart>>3;
	    PrvWriteSensorReg(OV2620_HREFST, &regVal);
	    PrvReadSensorReg(OV2620_COMM, &regVal);
	    regVal &= ~0x07;
	    regVal |= hStart & 0x07;
	    PrvWriteSensorReg(OV2620_COMM, &regVal);
	
	    // set Horizontal Window End
	    regVal = hEnd>>3;
	    PrvWriteSensorReg(OV2620_HREFEND, &regVal);
	    PrvReadSensorReg(OV2620_COMM, &regVal);
	    regVal &= ~(0x07 << 3);
	    regVal |= (hEnd & 0x07) << 3;
	    PrvWriteSensorReg(OV2620_COMM, &regVal);
	
	    // set Vertical Window Start
	    regVal = vStart>>2;
	    PrvWriteSensorReg(OV2620_VSTRT, &regVal);
	    PrvReadSensorReg(OV2620_COMA, &regVal);
	    regVal &= ~0x03;
	    regVal |= vStart & 0x03;
	    PrvWriteSensorReg(OV2620_COMA, &regVal);
	
	    // set Vertical Window End
	    regVal = vEnd>>2;
	    PrvWriteSensorReg(OV2620_VEND, &regVal);
	    PrvReadSensorReg(OV2620_COMA, &regVal);
	    regVal &= ~(0x03 << 2);
	    regVal |= (vEnd & 0x03) << 2;
	    PrvWriteSensorReg(OV2620_COMA, &regVal);

    }

    //
    // return window region
    //
    
    *winStartX = hStart;
    *winStartY = vStart;
    *winEndX   = hEnd;
    *winEndY   = vEnd;
    
    return XLLP_STATUS_SUCCESS;
}

/***********************************************************************
 *
 * Contrast
 *
 ***********************************************************************/
// FIX ME: TBD
const static XLLP_UINT8_T ContrastLowestSettings[] = {
    0x6C, 0x80,
    0x6D, 0xa0,
    0x6E, 0x78,
    0x6F, 0x50,
    0x70, 0x48,
    0x71, 0x40,
    0x72, 0x48,
    0x73, 0x40,
    0x74, 0x40,
    0x75, 0x40,
    0x76, 0x40,
    0x77, 0x40,
    0x78, 0x3e,
    0x79, 0x3c,
    0x7A, 0x3c,
    0x7B, 0x28,
    0x7C, 0x8,
    0x7D, 0x12,
    0x7E, 0x21,
    0x7F, 0x35,
    0x80, 0x3e,
    0x81, 0x46,
    0x82, 0x4f,
    0x83, 0x57,
    0x84, 0x5f,
    0x85, 0x67,
    0x86, 0x77,
    0x87, 0x87,
    0x88, 0xa6,
    0x89, 0xc4,
    0x8A, 0xe2,
    OV2620_REGEND,     0x00        // End of list delimiter.
};

const static XLLP_UINT8_T ContrastLowSettings[] = {
    0x6C, 0x70,
    0x6D, 0x80,
    0x6E, 0x40,
    0x6F, 0x54,
    0x70, 0x58,
    0x71, 0x60,
    0x72, 0x60,
    0x73, 0x60,
    0x74, 0x50,
    0x75, 0x58,
    0x76, 0x44,
    0x77, 0x3c,
    0x78, 0x30,
    0x79, 0x28,
    0x7A, 0x22,
    0x7B, 0x44,
    0x7C, 0x7,
    0x7D, 0xf,
    0x7E, 0x17,
    0x7F, 0x2c,
    0x80, 0x37,
    0x81, 0x43,
    0x82, 0x4f,
    0x83, 0x5b,
    0x84, 0x65,
    0x85, 0x70,
    0x86, 0x81,
    0x87, 0x90,
    0x88, 0xa8,
    0x89, 0xbc,
    0x8A, 0xcd,
    OV2620_REGEND,     0x00        // End of list delimiter.
};

const static XLLP_UINT8_T ContrastMiddleSettings[] = {
    0x6C, 0x40,
    0x6D, 0x30,
    0x6E, 0x4B,
    0x6F, 0x60,
    0x70, 0x70,
    0x71, 0x70,
    0x72, 0x70,
    0x73, 0x70,
    0x74, 0x60,
    0x75, 0x60,
    0x76, 0x50,
    0x77, 0x48,
    0x78, 0x3A,
    0x79, 0x2E,
    0x7A, 0x28,
    0x7B, 0x22,
    0x7C, 0x4,
    0x7D, 0x7,
    0x7E, 0x10,
    0x7F, 0x28,
    0x80, 0x36,
    0x81, 0x44,
    0x82, 0x52,
    0x83, 0x60,
    0x84, 0x6C,
    0x85, 0x78,
    0x86, 0x8C,
    0x87, 0x9E,
    0x88, 0xBB,
    0x89, 0xD2,
    0x8A, 0xE6,
    OV2620_REGEND,     0x00        // End of list delimiter.
};

const static XLLP_UINT8_T ContrastHighSettings[] = {
    0x6c, 0x10,
    0x6d, 0x10,
    0x6e, 0x18,
    0x6f, 0x5c,
    0x70, 0x90,
    0x71, 0x90,
    0x72, 0x90,
    0x73, 0x90,
    0x74, 0x80,
    0x75, 0x80,
    0x76, 0x60,
    0x77, 0x5c,
    0x78, 0x44,
    0x79, 0x24,
    0x7a, 0x1a,
    0x7b, 0x10,
    0x7c, 0x1,
    0x7d, 0x2,
    0x7e, 0x5,
    0x7f, 0x1c,
    0x80, 0x2e,
    0x81, 0x40,
    0x82, 0x52,
    0x83, 0x64,
    0x84, 0x74,
    0x85, 0x84,
    0x86, 0x9c,
    0x87, 0xb3,
    0x88, 0xd5,
    0x89, 0xe7,
    0x8a, 0xf4,
    OV2620_REGEND,     0x00        // End of list delimiter.
};

const static XLLP_UINT8_T ContrastHighestSettings[] = {
    0x6c, 0x20,
    0x6d, 0x40,
    0x6e, 0x10,
    0x6f, 0x38,
    0x70, 0x80,
    0x71, 0xe0,
    0x72, 0xd0,
    0x73, 0xe8,
    0x74, 0xa0,
    0x75, 0x80,
    0x76, 0x80,
    0x77, 0x54,
    0x78, 0x30,
    0x79, 0x22,
    0x7a, 0x8,
    0x7b, 0x1,
    0x7c, 0x2,
    0x7d, 0x6,
    0x7e, 0x8,
    0x7f, 0x16,
    0x80, 0x26,
    0x81, 0x42,
    0x82, 0x5c,
    0x83, 0x79,
    0x84, 0x8d,
    0x85, 0x9d,
    0x86, 0xbd,
    0x87, 0xd2,
    0x88, 0xea,
    0x89, 0xfb,
    0x8a, 0xff,
    OV2620_REGEND,     0x00        // End of list delimiter.
};
    
XLLP_STATUS_T OV2620SetContrast(XLLP_UINT32_T value)
{
    const XLLP_UINT8_T *regP;

    regP = NULL;
    switch(value) {
        case XLLP_SENSOR_CONTRAST_LOWEST:
            regP = ContrastLowestSettings;
            break;
        case XLLP_SENSOR_CONTRAST_LOW:
            regP = ContrastLowSettings;
            break;
        case XLLP_SENSOR_CONTRAST_MIDDLE:
            regP = ContrastMiddleSettings;
            break;
        case XLLP_SENSOR_CONTRAST_HIGH:
            regP = ContrastHighSettings;
            break;
        case XLLP_SENSOR_CONTRAST_HIGHEST:
            regP = ContrastHighestSettings;
            break;
        default:
            regP = ContrastMiddleSettings;
            break;
    }
    
    // set hw
    if (regP)        
        OV2620SetRegs(regP);
    return XLLP_STATUS_SUCCESS;
}

/***********************************************************************
 *
 * Exposure
 *
 ***********************************************************************/
// FIX ME: TBD
const static XLLP_UINT8_T ExposureSettings[] = {
    0x40, 0x30, 0x81, // EV-2
    0x58, 0x48, 0x91, // EV-1
    0x88, 0x7c, 0x93, // EV0
    0xa0, 0x90, 0xb4, // EV+1
    0xc0, 0xb0, 0xd6, // EV+2
};

XLLP_STATUS_T OV2620SetExposure(XLLP_UINT32_T value)
{
    XLLP_UINT8_T aew, aeb, vv;
    int index = -1;

    switch(value) {
        case XLLP_SENSOR_EXPOSURE_LOWEST:
            index = 0;
            break;
        case XLLP_SENSOR_EXPOSURE_LOW:
            index = 3;
            break;
        case XLLP_SENSOR_EXPOSURE_MIDDLE:
            index = 6;
            break;
        case XLLP_SENSOR_EXPOSURE_HIGH:
            index = 9;
            break;
        case XLLP_SENSOR_EXPOSURE_HIGHEST:
            index = 12;
            break;
        default:
            break;
    }

    aew = aeb = vv = 0;

    if (index != -1){
        aew = ExposureSettings[index];
        aeb = ExposureSettings[index + 1];
        vv  = ExposureSettings[index + 2];
    }
    
    // set hw
    if( aew || aeb || vv){
        OV2620WriteSensorReg( OV2620_AEW, &aew);
        OV2620WriteSensorReg( OV2620_AEB, &aeb);
        OV2620WriteSensorReg( OV2620_VV, &vv);
    }
    
    return XLLP_STATUS_SUCCESS;
}

/***********************************************************************
 *
 * Auto White Balance
 *
 ***********************************************************************/
// FIX ME: TBD
const static XLLP_UINT8_T AWBAuto[] = {
    0x13, 0xad,
    0x01, 0x80,
    0x02, 0x80,
    0x60, 0x14,
    0x5f, 0x05,
    0x13, 0xaf,
    OV2620_REGEND,     0x00        // End of list delimiter.
};

const static XLLP_UINT8_T AWBFluorescent[] = {
    0x13, 0xad,
    0x01, 0x6c,
    0x02, 0x2e,
    0x5f, 0x05,
    OV2620_REGEND,     0x00        // End of list delimiter.
};

const static XLLP_UINT8_T AWBOutdoor[] = {
    0x13, 0xad,
    0x01, 0x44,
    0x02, 0x44,
    0x5f, 0x05,
    OV2620_REGEND,     0x00        // End of list delimiter.
};

const static XLLP_UINT8_T AWBIncandescent[] = {
    0x13, 0xad,
    0x01, 0x6c,
    0x02, 0x20,
    0x5f, 0x05,
    OV2620_REGEND,     0x00        // End of list delimiter.
};


XLLP_STATUS_T OV2620SetWhiteBalance(XLLP_UINT32_T value)
{
    const XLLP_UINT8_T *regP;

    regP = NULL;
    switch(value) {
        case XLLP_SENSOR_WHITEBALANCE_AUTO:                //Auto:
            regP = AWBAuto;
            break;
        case XLLP_SENSOR_WHITEBALANCE_INCANDESCENT:        //Incandescent:
            regP = AWBIncandescent;
            break;
        case XLLP_SENSOR_WHITEBALANCE_SUNNY:               //Sunny:
            regP = AWBOutdoor;
            break;
        case XLLP_SENSOR_WHITEBALANCE_FLUORESCENT:         //Fluorescent:
            regP = AWBFluorescent;
            break;
        default:
            break;
    }

    // set hw
    if (regP) {
        OV2620SetRegs(regP);
    }
    return XLLP_STATUS_SUCCESS;
}

