/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *(C) Copyright 2006 Marvell International Ltd.  
 * All Rights Reserved 
 */

/**
**
**  FILENAME:       xllp_2d_gcu.c
**
**  PURPOSE: Contains all primitive functions for Monahans 2D Graphics
**           Controller register access and control
**                  
******************************************************************************/

#include "xllp_2d_gcu.h"

/* 2D Instruction Template */
static unsigned long instruction_template[]= {
	0x00000000,		/* GC_BBST */	
	0x01000000,		/* GC_BBEND */
	0x02000000,		/* GC_BUFFI */
	0x04000000,		/* GC_LREG */
	0x05000000,		/* GC_NOP */
	0x06000000, 		/* GC_DBFLIP */
	0x07000001,		/* GC_STREG */
	0x08000000,		/* GC_INT */
	0x09000000,		/* GC_WAIT */

	0x40000000,		/* GC_CFILL */
	0x41000000,		/* GC_CKBLT */
	0x42000000,		/* GC_LINE */
	0x43000000,		/* GC_AALINE */
	0x44000000,		/* GC_SCBLT */
	0x45000000,		/* GC_STRBLT */
	0x46000000,		/* GC_CKSTRBLT */
	0x47000000,		/* GC_ABLND */
	0x48000000,		/* GC_SCALE */
	0x49000000,		/* GC_BIAS */
	0x4A000000,		/* GC_ROT */
	0x4B000000,		/* GC_RAST */
	0x4C000000,		/* GC_PATT */
	0x4D000000,		/* GC_DECBLT */
};

/*****************************************************************************
 * Help Routines
 */

/* 
 * 
 * Returns:
 * 	0 : ok
 *      others: failure
 */
static int validate_2d_instruction( XLLP_GCU_BLITPARMS_T *pBlitParms)
{
	int len=0;

	/* Invalid instruction opcode */
	if (pBlitParms->opcode > (sizeof(instruction_template)>>2)) 
		return -1;
	
	switch(pBlitParms->opcode) {
	case GC_BUFFI:	/* Buffer Info */

		len = 3;
		/* reserved */
		if ( pBlitParms->instruction.buffi.buff_addr > 0xA) return -1;

		break;

	case GC_LREG:

		len = 2;
		/* GC_LREG <register>, <imm>, <format>, data */
		if (pBlitParms->instruction.lreg.imm) {
			
			if ( pBlitParms->instruction.lreg.pixel_format > 0xB) return -1;

			if ( (pBlitParms->instruction.lreg.pixel_format == 0xA) 
				|| (pBlitParms->instruction.lreg.pixel_format == 0xB)) 
				len = 3;

		}
		break;

	case GC_NOP:
	case GC_INT:
	case GC_DBFLIP:

		len = 1;
		break;

	case GC_STREG:
		len = 2;
		
		if ( pBlitParms->instruction.streg.reg > 0x8) return -1;
		if ( pBlitParms->instruction.streg.pixel_format > 0xB ) return -1;
		break;
		
	case GC_WAIT:
	
		len = 1;
		
		if ( pBlitParms->instruction.wait.event_id > 0x13) return -1;
		break;

	case GC_CFILL:
		/* FIXME: pls check width>0, height>0, otherwise GCU halts */
		len = 4;
		if (pBlitParms->instruction.cfill.pixel_format > 0xB) return -1;
		if (pBlitParms->instruction.cfill.imm) {
			len = 5;
			
			if ( (pBlitParms->instruction.cfill.pixel_format == 0xA) 
				|| (pBlitParms->instruction.cfill.pixel_format == 0xB)) 
				len = 6;
		} 
		break;

	case GC_LINE:

		len = 5;

		/* reserved */
		if (pBlitParms->instruction.line.pixel_format > 0xB) return -1;
		if (pBlitParms->instruction.line.imm) {
			len = 6;

			if ( (pBlitParms->instruction.line.pixel_format == 0xA) 
				|| (pBlitParms->instruction.line.pixel_format==0xB) ) {
				len = 7;
			}
		}
		break;

	case GC_AALINE:

		len = 6;

		/* reserved */
		if (pBlitParms->instruction.aaline.pixel_format > 0xB) return -1;
		if (pBlitParms->instruction.aaline.imm) {
			len = 7;

			if ( (pBlitParms->instruction.aaline.pixel_format == 0xA) 
				|| (pBlitParms->instruction.aaline.pixel_format==0xB) ) {
				len = 8;
			}
		}
		break;

	case GC_CKSTRBLT:
		/* FIXME : check Source 0 Block width restricts */
		len = 0xb;
		break;

	case GC_ABLND:

		len = 8;
		if (pBlitParms->instruction.ablnd.imm) len = 9;
		break;


	case GC_SCALE:
		
		len = 6;
		if (pBlitParms->instruction.scale.imm) len = 7;
		break;

	case GC_BIAS:

		len = 6;
		if (pBlitParms->instruction.bias.imm) len = 7;
		break;

	case GC_SCBLT:
	case GC_ROT:

		len = 6;
		break;

	case GC_CKBLT:
		len = 7;
		break;

	case GC_PATT:
		/* FIXME : check Source 0 Block width restricts */
		len = 7;
		break;

	case GC_RAST:
		len = 8;
		break;

	case GC_STRBLT:
		/* FIXME : check Source 0 Block width restricts */
		len = 9;
		break;

	case GC_DECBLT:
		len = 9;
		break;
	}

	pBlitParms->word_count = len ;

	return 0;
}

/* 
 * Returns:
 * 	=0 : no space
 *      >0 : instruction word count
 *
 */
static int encode_2d_instruction( XLLP_GCU_CONTEXT_T *pGCU, XLLP_GCU_BLITPARMS_T *pBlitParms, unsigned long *w)
{
	unsigned long len=0;

	/* encode instruction */
	w[0] = instruction_template[pBlitParms->opcode];

	switch(pBlitParms->opcode) {
	case GC_BUFFI:	/* Buffer Info */

		len = 2;

		w[0] |= (pBlitParms->instruction.buffi.buff_addr & 0x1f)<<4;
		w[1] = pBlitParms->instruction.buffi.base_addr;

		/* WD2 - 22:19 : pixel format, 18:5 : stride, 3:0 : step */
		w[2] = ((pBlitParms->instruction.buffi.pixel_format & 0xf)<<19) | 
			((pBlitParms->instruction.buffi.stride & 0x3fff)<<5) | 
			(pBlitParms->instruction.buffi.step & 0xf); 
		break;

	case GC_LREG:

		len = 1;

		w[0] |= ((pBlitParms->instruction.lreg.reg & 0xf) << 20) |
			((pBlitParms->instruction.lreg.pixel_format & 0xf) << 8) |
			((pBlitParms->instruction.lreg.imm & 0x1) << 4);
		w[1] = pBlitParms->instruction.lreg.data[0];

		/* GC_LREG <register>, <imm>, <format>, data */
		if (pBlitParms->instruction.lreg.imm) {
			/* 8 byte ? */
			if ( (pBlitParms->instruction.lreg.pixel_format == 0xA) || 
				(pBlitParms->instruction.lreg.pixel_format <= 0xB)) {
				len = 2;
				w[2] = pBlitParms->instruction.lreg.data[1];
			}
		}
		break;

	case GC_NOP:

		len = 0;
		w[0] |= (pBlitParms->instruction.nop.nop_id & 0xfffff) << 4;
		break;

	case GC_DBFLIP:

		len = 0;
		w[0] |= (pBlitParms->instruction.dbflip.mode & 0x3) << 5;
		break;

	case GC_STREG:

		len = 1;
		
		w[0] |= ((pBlitParms->instruction.streg.reg & 0xf) << 20) |
			((pBlitParms->instruction.streg.pixel_format & 0xf) << 8);
		w[1] = pBlitParms->instruction.streg.addr;
		break;
		
	case GC_INT:

		len = 0;
		w[0] |= ((pBlitParms->instruction.intr.int_id & 0xfffff) << 4);
		break;

	case GC_WAIT:
	
		len = 0;
		w[0] |= ((pBlitParms->instruction.wait.event_id & 0x1f)<<4);
		break;

	case GC_CFILL:

		len = 3;

		w[0] |= (pBlitParms->instruction.cfill.pixel_format & 0xf) << 8;
		w[1] = pBlitParms->instruction.cfill.dx0;
		w[2] = pBlitParms->instruction.cfill.dy0;
		w[3] = (pBlitParms->instruction.cfill.height << 16) |
			(pBlitParms->instruction.cfill.width);
		
		if (!pBlitParms->instruction.cfill.imm) 
			w[0] |= (pBlitParms->instruction.lreg.reg & 0xf) << 20;
		else {
			//w[0] |= 0x10;	//bit4: imm 

			len = 4;
			w[4] = pBlitParms->instruction.cfill.data[0];
			
			if ( (pBlitParms->instruction.cfill.pixel_format == 0xA) 
				|| (pBlitParms->instruction.cfill.pixel_format == 0xB)) {
				len = 5;
				w[5] = pBlitParms->instruction.cfill.data[1];
			}
		} 
		break;

	case GC_CKBLT:
		
		len = 6;

		w[0] |= (pBlitParms->instruction.ckblt.invert & 0x1) << 4;
		w[1] = pBlitParms->instruction.ckblt.dx0;
		w[2] = pBlitParms->instruction.ckblt.dy0;
		w[3] = pBlitParms->instruction.ckblt.sx0;
		w[4] = pBlitParms->instruction.ckblt.sy0;
		w[5] = (pBlitParms->instruction.ckblt.height <<16 ) | (pBlitParms->instruction.ckblt.width);
		w[6] = ((pBlitParms->instruction.ckblt.key & 0x3) << 17) |
			(pBlitParms->instruction.ckblt.threshold & 0x1ff);
		break;

	case GC_LINE:

		len = 4;

		w[0] |= 0x300;//((pBlitParms->instruction.line.pixel_format & 0xf) << 8);
		w[1] = pBlitParms->instruction.line.dx0;
		w[2] = pBlitParms->instruction.line.dy0;
		w[3] = ((pBlitParms->instruction.line.x_dir & 0x1) << 31) |
			((pBlitParms->instruction.line.x_is_1 & 0x1) << 30) |
			((pBlitParms->instruction.line.delx & 0x3fff) << 16) |
			((pBlitParms->instruction.line.y_dir & 0x1) << 15) |
			((pBlitParms->instruction.line.y_is_1 & 0x1) << 14) |
			(pBlitParms->instruction.line.dely & 0x3fff);
		w[4] = ((pBlitParms->instruction.line.width & 0xff) << 16) | 
			(pBlitParms->instruction.line.length & 0x7ff);

		if (!pBlitParms->instruction.line.imm)
			w[0] |= ((pBlitParms->instruction.line.reg & 0xf) << 20);
		else {

			len = 5;

			w[5] = pBlitParms->instruction.line.data[0];
			// lihui
			w[0] |=  ((pBlitParms->instruction.line.imm)<<4);

			/* 48, 64 bit */
			if ( (pBlitParms->instruction.line.pixel_format == 0xA) 
				|| (pBlitParms->instruction.line.pixel_format==0xB) ) {
				len = 6;
				w[6] = pBlitParms->instruction.line.data[1];
			}
		}
		break;
	case GC_AALINE:

		len = 5;

		w[0] |= ((pBlitParms->instruction.aaline.pixel_format & 0xf) << 8);
		w[1] = pBlitParms->instruction.aaline.dx0;
		w[2] = pBlitParms->instruction.aaline.dy0;
		w[3] = ((pBlitParms->instruction.aaline.x_dir & 0x1) << 31) |
			((pBlitParms->instruction.aaline.x_is_1 & 0x1) << 30) |
			((pBlitParms->instruction.aaline.delx & 0x3fff) << 16) |
			((pBlitParms->instruction.aaline.y_dir & 0x1) << 15) |
			((pBlitParms->instruction.aaline.y_is_1 & 0x1) << 14) |
			(pBlitParms->instruction.aaline.dely & 0x3fff);
		w[4] = ((pBlitParms->instruction.aaline.width & 0xff) << 16) | 
			(pBlitParms->instruction.aaline.length & 0x7ff);
		w[5] = (pBlitParms->instruction.aaline.address&0xfff8);

		if (!pBlitParms->instruction.aaline.imm)
			w[0] |= ((pBlitParms->instruction.aaline.reg & 0xf) << 20);
		else {

			len = 6;

			w[6] = pBlitParms->instruction.aaline.data[0];
			w[0] |=  ((pBlitParms->instruction.line.imm)<<4);

			/* 48, 64 bit */
			if ( (pBlitParms->instruction.aaline.pixel_format == 0xA) 
				|| (pBlitParms->instruction.aaline.pixel_format==0xB) ) {
				len = 7;
				w[7] = pBlitParms->instruction.aaline.data[1];
			}
		}
		break;

	case GC_SCBLT:
		
		len = 5;

		w[0] |= ((pBlitParms->instruction.scblt.rop & 0xff) << 16);
		w[1] = pBlitParms->instruction.scblt.dx0;
		w[2] = pBlitParms->instruction.scblt.dy0;
		w[3] = pBlitParms->instruction.scblt.sx0;
		w[4] = pBlitParms->instruction.scblt.sy0;
		w[5] = (pBlitParms->instruction.scblt.height << 16) | (pBlitParms->instruction.scblt.width);
		break;


	case GC_STRBLT:
		len = 8;

		w[1] = pBlitParms->instruction.strblt.dx0;
		w[2] = pBlitParms->instruction.strblt.dy0;
		w[3] = pBlitParms->instruction.strblt.sx0;
		w[4] = pBlitParms->instruction.strblt.sy0;
		w[5] = (pBlitParms->instruction.strblt.s_height << 16) | 
			(pBlitParms->instruction.strblt.s_width & 0x1ff);
		w[6] = (pBlitParms->instruction.strblt.d_height << 16) |
			(pBlitParms->instruction.strblt.d_width);
		w[7] = (pBlitParms->instruction.strblt.x_str);
		w[8] = (pBlitParms->instruction.strblt.y_str);
		
		break;

	case GC_CKSTRBLT:

		len = 0xa;


		w[1] = pBlitParms->instruction.ckstrblt.dx0;
		w[2] = pBlitParms->instruction.ckstrblt.dy0;
		w[3] = pBlitParms->instruction.ckstrblt.s0x0;
		w[4] = pBlitParms->instruction.ckstrblt.s0y0;
		w[5] = pBlitParms->instruction.ckstrblt.s1x0;
		w[6] = pBlitParms->instruction.ckstrblt.s1y0;
		w[7] = (pBlitParms->instruction.ckstrblt.s_height << 16) | 
			(pBlitParms->instruction.ckstrblt.s_width & 0x1ff);
		w[8] = (pBlitParms->instruction.ckstrblt.d_height << 16) |
			(pBlitParms->instruction.ckstrblt.s_width);
		w[9] = ((pBlitParms->instruction.ckstrblt.x_str & 0x3ff) << 16) |
			(pBlitParms->instruction.ckstrblt.y_str & 0x3ff);

		w[10] = ((pBlitParms->instruction.ckstrblt.key & 0x3) << 17) |
			(pBlitParms->instruction.ckstrblt.threshold & 0x1ff);
		break;

	case GC_ABLND:
		len = 7;
		w[0] |= ((pBlitParms->instruction.ablnd.single_alpha & 0x1) << 8) |
			((pBlitParms->instruction.ablnd.blend_op & 0x3) << 6) |
			((pBlitParms->instruction.ablnd.imm & 0x3) << 4);
		w[1] = pBlitParms->instruction.ablnd.dx0;
		w[2] = pBlitParms->instruction.ablnd.dy0;
		w[3] = pBlitParms->instruction.ablnd.s0x0;
		w[4] = pBlitParms->instruction.ablnd.s0y0;
		w[5] = pBlitParms->instruction.ablnd.s1x0;
		w[6] = pBlitParms->instruction.ablnd.s1y0;
		w[7] = (pBlitParms->instruction.ablnd.height << 16) |
			(pBlitParms->instruction.ablnd.width);

		if (pBlitParms->instruction.ablnd.imm) {
			len = 8;
			w[8] = 0;
			if (pBlitParms->instruction.ablnd.imm & 0x1) 
				w[8] |= (pBlitParms->instruction.ablnd.alpha0<<16);
			if (pBlitParms->instruction.ablnd.imm & 0x2)
				w[8] |= (pBlitParms->instruction.ablnd.alpha1);
		}
		break;


	case GC_SCALE:
		
		len = 5;

		w[0] |= ((pBlitParms->instruction.scale.imm & 0x1) << 4);
		w[1] = pBlitParms->instruction.scale.dx0;
		w[2] = pBlitParms->instruction.scale.dy0;
		w[3] = pBlitParms->instruction.scale.sx0;
		w[4] = pBlitParms->instruction.scale.sy0;
		w[5] = (pBlitParms->instruction.scale.height << 16) | pBlitParms->instruction.scale.width;
	
		if (pBlitParms->instruction.scale.imm) {
			len = 6;
			w[6] = pBlitParms->instruction.scale.scf & 0xffff;
		}
		break;

	case GC_BIAS:
		len = 5;

		w[0] |= ((pBlitParms->instruction.bias.imm & 0x1) << 4);
		w[1] = pBlitParms->instruction.bias.dx0;
		w[2] = pBlitParms->instruction.bias.dy0;
		w[3] = pBlitParms->instruction.bias.sx0;
		w[4] = pBlitParms->instruction.bias.sy0;
		w[5] = (pBlitParms->instruction.bias.height << 16) | pBlitParms->instruction.bias.width;
	
		if (pBlitParms->instruction.bias.imm) {
			len = 6;
			w[6] = pBlitParms->instruction.bias.bias& 0xffff;
		}
		break;

	case GC_ROT:
		len = 5;

		w[0] |= (pBlitParms->instruction.rot.rot & 0x3) << 4;
		w[1] = pBlitParms->instruction.rot.dx0;
		w[2] = pBlitParms->instruction.rot.dy0;
		w[3] = pBlitParms->instruction.rot.sx0;
		w[4] = pBlitParms->instruction.rot.sy0;
		w[5] = (pBlitParms->instruction.rot.height << 16) | pBlitParms->instruction.rot.width;
		break;

	case GC_RAST:

		len = 7;

		w[0] |= (pBlitParms->instruction.rast.rop & 0xff) << 16;
		w[1] = pBlitParms->instruction.rast.dx0;
		w[2] = pBlitParms->instruction.rast.dy0;
		w[3] = pBlitParms->instruction.rast.s0x0;
		w[4] = pBlitParms->instruction.rast.s0y0;
		w[5] = pBlitParms->instruction.rast.s1x0;
		w[6] = pBlitParms->instruction.rast.s1y0;
		w[7] = (pBlitParms->instruction.rast.height << 16) | pBlitParms->instruction.rast.width;
		break;

	case GC_PATT:

		len = 6;

		w[0] |= (pBlitParms->instruction.patt.opt & 0x1) << 4;
		w[1] = pBlitParms->instruction.patt.dx0;
		w[2] = pBlitParms->instruction.patt.dy0;
		w[3] = pBlitParms->instruction.patt.sx0;
		w[4] = pBlitParms->instruction.patt.sy0;
		w[5] = (pBlitParms->instruction.patt.s_height << 16) | pBlitParms->instruction.patt.s_width;
		w[6] = (pBlitParms->instruction.patt.d_height << 16) | pBlitParms->instruction.patt.d_width;
		break;

	case GC_DECBLT:
		
		len = 8;

		w[1] = pBlitParms->instruction.decblt.dx0;
		w[2] = pBlitParms->instruction.decblt.dy0;
		w[3] = pBlitParms->instruction.decblt.sx0;
		w[4] = pBlitParms->instruction.decblt.sy0;
		w[5] = (pBlitParms->instruction.decblt.s_height << 16) | (pBlitParms->instruction.decblt.s_width & 0x1ff);
		w[6] = (pBlitParms->instruction.decblt.d_height << 16) | (pBlitParms->instruction.decblt.d_width & 0x1ff);
		w[7] = (pBlitParms->instruction.decblt.x_dec_int << 16) | pBlitParms->instruction.decblt.x_dec_frac;
		w[8] = (pBlitParms->instruction.decblt.y_dec_int << 16) | pBlitParms->instruction.decblt.y_dec_frac;
		break;
	}

	/* set word count */
	w[0] |=  len;

	return (len+1);
}

/*****************************************************************************
 * Exported XLLP Routines
 */

/*
 * XLLP_STATUS_T XllpGCU_Enable(XLLP_GCU_CONTEXT_T *pGCU)
 *
 * Description:
 *	Intialize the 2D Graphics Controller
 *
 * Input Parameters:
 *	XLLP_GCU_CONTEXT_T *pGCU  the GCU device handle
 *
 * Output Parameters:
 *	None
 *
 * Returns:
 *	XLLP_GCU_SUCCESS upon successful GCU enable.
 *	XLLP_GCU_FAILURE upon unsuccessful GCU enable.
 */

XLLP_STATUS_T XllpGCU_Init(XLLP_GCU_CONTEXT_T *pGCU)
{
	volatile P_XLLP_GCU_REG_T pGCRegs = (P_XLLP_GCU_REG_T) pGCU->GCU;

	/* Ring buffer length has be >=256bytes */
	if (pGCU->ring.length <= 0xFF) 
		return XLLP_STATUS_FAILURE;

	/* 256-bytes align */
	pGCU->ring.addr= (pGCU->ring.addr+ 0xFF) & (~0xFF);
	pGCU->ring.base = (pGCU->ring.base + 0xFF) & (~0xFF);
	pGCU->ring.head = pGCU->ring.base;
	pGCU->ring.tail = pGCU->ring.base;

	/* 256-bytes align, up to 256K */
	pGCU->ring.length = (pGCU->ring.length) & (~0xFF);
	pGCU->ring.length = (pGCU->ring.length > (1U<<18)) ? (1U<<18) : pGCU->ring.length;
	
	/* FIXME : Needs XLLP Clock Manager */
	/* Enable 2D Controller Clock */
	/* D0CKENA |= D0CKENA_GCCEN7; */

	/* Reset the Graphics controller completely */
	pGCRegs->gccr = GCCR_ABORT | (pGCRegs->gccr & 0xF);

	/* De-assert ABORT to allow writes to ring buffer configuration register */
	pGCRegs->gccr = GCCR_STOP | (pGCRegs->gccr & 0xF);

	/* Setup ring buffer */
	pGCRegs->gcrblr = 0;
	pGCRegs->gcrbbr = pGCU->ring.base;
	pGCRegs->gcrbtr = pGCU->ring.tail;
	pGCRegs->gcrblr = pGCU->ring.length;

	return XLLP_STATUS_SUCCESS;
}

/*
 * XLLP_STATUS_T XllpGCU_Enable(XLLP_GCU_CONTEXT_T *pGCU)
 *
 * Description:
 *	Enable the 2D Graphics Controller after initialization. 
 *
 * Input Parameters:
 *	XLLP_GCU_CONTEXT_T *pGCU  the GCU device handle
 *
 * Output Parameters:
 *	None
 *
 * Returns:
 *	XLLP_GCU_SUCCESS upon successful GCU enable.
 *	XLLP_GCU_FAILURE upon unsuccessful GCU enable.
 */

XLLP_STATUS_T XllpGCU_Enable(XLLP_GCU_CONTEXT_T *pGCU)
{
	volatile P_XLLP_GCU_REG_T pGCRegs = (P_XLLP_GCU_REG_T) pGCU->GCU;
	pGCRegs->gcrbtr = pGCU->ring.tail;
	pGCRegs->gciecr = pGCU->interrupt_mask;

	/* start the ontroller */
	pGCRegs->gccr = (pGCRegs->gccr & (~GCCR_STOP)) & GCCR_MASK;
	
	return XLLP_STATUS_SUCCESS;
}

/* 
 * XLLP_STATUS_T XllpGCU_Disable(XLLP_GCU_CONTEXT_T *pGCU)
 *
 * Description:
 * 	Disable the 2D Graphics Controller Unit to enter a low power state.  
 *
 * Input Parameters:
 * 	XLLP_GCU_CONTEXT_T *pGCU  the GCU device handle
 *
 * Output Parameters:
 * 	None
 * 
 * Returns:
 *	XLLP_GCU_SUCCESS upon successful GCU disable.
 *	XLLP_GCU_FAILURE upon unsuccessful GCU disable
 *
 */

XLLP_STATUS_T XllpGCU_Disable(XLLP_GCU_CONTEXT_T *pGCU)
{
	volatile P_XLLP_GCU_REG_T pGCRegs = (P_XLLP_GCU_REG_T) pGCU->GCU;

	if ( ! (pGCRegs->gccr & GCCR_STOP) )
		pGCRegs->gccr = (pGCRegs->gccr & GCCR_MASK) | GCCR_STOP;

	/* update GCU context */
	pGCU->ring.head = pGCRegs->gcrbhr;
	pGCU->ring.exhead = pGCRegs->gcrbexhr;

	return XLLP_STATUS_SUCCESS;
}

/* 
 * XLLP_STATUS_T XllpGCU_SubmitInstruction(XLLP_GCU_CONTEXT_T *pGCU, XLLP_GCU_BLITPARMS_T *pBlitParms, XLLP_UINT32_T mode);
 *
 * Description:
 *
 * 	Load a single instruction to the graphics controller's active 
 *	instruction list.  It handles a single BLT or Line Draw operation
 *	in one function call.  Use XllpGCU_SubmitInstructionList() to 
 *	efficiently handle the submission of multiple instructions.  
 *	XllpGCU_SubmitInstruction() may be used for this purpose but will 
 * 	have higher overhead when used in this way.
 *
 *	Set the mode flag according to the needs of the graphics driver.
 *	  o Set mode to GCU_SYNCHRONOUS_RETURN to instruct the API to wait
 *	    until the instruction executes before returning.
 *	  o Set mode to GCU_ASYNCHRONOUS_RETURN to instruct the API to return
 *	    as soon as the instruction is queued.
 *
 * Input Parameters:
 *
 *	XLLP_GCU_CONTEXT_T *pGCU : the GCU device handle
 * 	XLLP_GCU_BLITPARMS_T *pBlitParms : the 2D Graphics instruction. Caller configures the instruction type, the correct parameters for the corresponding structure, and sets the source 0 and source 1 buffer type as appropriate.  The buffer type may be XLLP_BUFFER_IS_PHYSICALLY_CONTIGUOUS or XLLP_BUFFER_NOT_PHYSICALLY_CONTIGUOUS.
 *	XLLP_UINT32_T mode : above
 *
 * Output Parameters:
 *	None
 *
 * Returns:
 *	XLLP_GCU_SUCCESS upon successful instruction submission.
 * 	XLLP_GCU_FAILURE upon unsuccessful instruction submission.
 *
 */

XLLP_STATUS_T XllpGCU_SubmitInstruction(XLLP_GCU_CONTEXT_T *pGCU, XLLP_GCU_BLITPARMS_T *pBlitParms, XLLP_UINT32_T mode)
{
	volatile P_XLLP_GCU_REG_T pGCRegs = (P_XLLP_GCU_REG_T) pGCU->GCU;
	int len, tail, wordcnt[2];
	XLLP_GCU_BLITPARMS_T nop;

	nop.opcode = GC_NOP;
	nop.instruction.nop.nop_id = 0;

	/* Update Ring Buffer head now */
	pGCU->ring.head = pGCRegs->gcrbhr;
	pGCU->ring.exhead = pGCRegs->gcrbexhr;
	pGCU->ring.tail = pGCRegs->gcrbtr;
	/* Validate instruction */
	if ( validate_2d_instruction(pBlitParms) )
		return XLLP_STATUS_FAILURE;
	
	/* Check Ring Buffer space */
	if (pGCU->ring.tail >= pGCU->ring.exhead) {
		wordcnt[0] = pGCU->ring.base + pGCU->ring.length - pGCU->ring.tail;
		wordcnt[1] = pGCU->ring.exhead - pGCU->ring.base;
	}
	else {
		wordcnt[0] = pGCU->ring.exhead - pGCU->ring.tail;
		wordcnt[1] = 0;
	}

	wordcnt[0] = wordcnt[0] >> 2;
	wordcnt[1] = wordcnt[1] >> 2;
	
	len = pBlitParms->word_count;
	if (wordcnt[0] >= len) {
		encode_2d_instruction( pGCU, pBlitParms, (unsigned long*) (pGCU->ring.addr + (pGCU->ring.tail - pGCU->ring.base)));

		tail = pGCU->ring.tail + (len<<2);
	}
	else if (wordcnt[1] >= len) {
		/* Fill with NOP */
		while (wordcnt[0]) {
			encode_2d_instruction( pGCU, &nop, (unsigned long*) (pGCU->ring.addr + pGCU->ring.length - (wordcnt[0]<<2)));
			wordcnt[0] --;
		}
		encode_2d_instruction( pGCU, pBlitParms, (unsigned long*) pGCU->ring.addr);
		
		tail = pGCU->ring.addr + (len<<2);
	}
	else return XLLP_STATUS_FAILURE;
	
	/* Update GCU Ring Buffer tail */
	pGCRegs->gcrbtr = pGCU->ring.tail = tail;

	/* Return when Ring Buffer is empty */
	if ( mode == GCU_SYNCHRONOUS_RETURN ) {
		XllpGCU_Enable(pGCU);
		while ( pGCRegs->gcrbexhr != pGCRegs->gcrbtr) {
			/* FIXME: delay for a while */
			int i=1000;
			while(i) i--;
		}
	}
	
	return XLLP_STATUS_SUCCESS;
}

/*
 * XLLP_STATUS_T XllpGCU_SubmitInstructionList(XLLP_GCU_CONTEXT_T *pGCU, XLLP_GCU_INSTRUCTION_LIST_T *pInstructionList);
 *
 * Description:
 *
 *	Load an instruction list into the graphics controllers active instruction
 * 	list.  
 *
 *	Set mode to GCU_SYNCHRONOUS_RETURN to instruct the API to wait until the 
 *	instruction executes before returning.
 *
 * 	Set mode to GCU_ASYNCHRONOUS_RETURN to instruct the API to return as soon
 *	as the instruction is queued.
 *
 *	Load an instruction list into the graphics controller's active instruction
 *	list.  It handles multiple BLT or Line Draw operation in one function call.
 *	Use XllpGCU_SubmitInstruction() to efficiently handle the submission of 
 *	single instructions.  XllpGCU_SubmitInstructionList() may be used for this
 *	purpose but will have higher overhead when used in this way.
 *
 *	Set the mode flag according to the needs of the graphics driver.
 *
 *	 o Set mode to GCU_SYNCHRONOUS_RETURN to instruct the API to wait until 
 *	   the instruction executes before returning.
 *	 o Set mode to GCU_ASYNCHRONOUS_RETURN to instruct the API to return as
 *	   soon as the instruction is queued.
 *
 *	If the results of this computation must be wordcntable by the time this
 *	function call returns then the driver must set the mode to 
 *	GCU_SYNCHRONOUS_RETURN.
 *
 * Input Parameters:
 *	XLLP_GCU_CONTEXT_T *pGCU : the GCU device handle
 * 	XLLP_GCU_INSTRUCTION_LIST_T *pInstructionList : the instructions to load into graphics controller
 *
 * Output Parameters:
 *	None
 *
 * Returns:
 *	XLLP_GCU_SUCCESS upon successful graphics list load.
 *	XLLP_GCU_FAILURE upon unsuccessful graphics list load.
 */

XLLP_STATUS_T XllpGCU_SubmitInstructionList(XLLP_GCU_CONTEXT_T *pGCU, XLLP_GCU_INSTRUCTION_LIST_T *pInstructionList, XLLP_UINT32_T mode)
{
	volatile P_XLLP_GCU_REG_T pGCRegs = (P_XLLP_GCU_REG_T) pGCU->GCU;
	int i, len, valid, no_space, wordcnt[2];
	unsigned long tail;
	XLLP_GCU_BLITPARMS_T nop;

	nop.opcode = GC_NOP;
	nop.instruction.nop.nop_id = 0;

	/* Update Ring Buffer head now */
	pGCU->ring.head = pGCRegs->gcrbhr;
	pGCU->ring.exhead = pGCRegs->gcrbexhr;

	/* Validate the instructions, and check ring buffer for space */
	no_space = 0;
	tail = pGCU->ring.tail;

	for ( i=0; i < pInstructionList->num_instr; i++ ) {

		/* Check instruction */
		valid = !validate_2d_instruction(pInstructionList->instructions + i);
		len = pInstructionList->instructions[i].word_count;
		
		/* Check Ring Buffer space */
		if ( valid && (!no_space)) {
			if (tail >= pGCU->ring.exhead) {
				wordcnt[0] = pGCU->ring.length - tail;
				wordcnt[1] = pGCU->ring.exhead - pGCU->ring.base;
			}
			else {
				wordcnt[0] = pGCU->ring.exhead - tail;
				wordcnt[1] = 0;
			}

			wordcnt[0] = (wordcnt[0] >> 2);
			wordcnt[1] = (wordcnt[1] >> 2);
	
			if (wordcnt[0] >= len) tail += (len<<2);
			else if (wordcnt[1] >= len) tail = pGCU->ring.addr + (len<<2);
			else no_space = 1;
		}
	}

	if (!valid) return XLLP_STATUS_FAILURE;
	if (no_space) return XLLP_STATUS_FAILURE;

	/* Encode the instructions into ring buffer */
	tail = pGCU->ring.tail;

	for ( i=0; i < pInstructionList->num_instr; i++ ) {

		len = pInstructionList->instructions[i].word_count;

		if (tail >= pGCU->ring.exhead) {
			wordcnt[0] = pGCU->ring.base + pGCU->ring.length - tail;
			wordcnt[1] = pGCU->ring.exhead - pGCU->ring.base;
		}
		else {
			wordcnt[0] = pGCU->ring.exhead - tail;
			wordcnt[1] = 0;
		}

		wordcnt[0] = wordcnt[0] >> 2;
		wordcnt[1] = wordcnt[1] >> 2;
	
		if (wordcnt[0] >= len) {
			encode_2d_instruction( pGCU, pInstructionList->instructions + i, (unsigned long*)(pGCU->ring.addr + (tail - pGCU->ring.base)));
			tail += len<<2;
		}
		else if (wordcnt[1] >= len) {
			/* fill with NOP */
			while (wordcnt[0]) {
				encode_2d_instruction( pGCU, &nop, (unsigned long*) (pGCU->ring.addr + pGCU->ring.length - (wordcnt[0]<<2)));
				wordcnt[0] --;
			}
			encode_2d_instruction( pGCU, pInstructionList->instructions + i, (unsigned long*) (pGCU->ring.addr));
			tail = pGCU->ring.addr + (len<<2);
		}
	}

	/* Update GCU Ring Buffer tail */
	pGCRegs->gcrbtr = pGCU->ring.tail = tail;

	/* sync mode */
	if ( mode == GCU_SYNCHRONOUS_RETURN ) {
		XllpGCU_Enable(pGCU);
		while ( pGCRegs->gcrbexhr != pGCRegs->gcrbtr) {
			/* FIXME: delay for a while */
			int i=1000;
			while(i) i--;
		}
	}

	return XLLP_STATUS_SUCCESS;
}
