/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *(C) Copyright 2006 Marvell International Ltd.  
 * All Rights Reserved 
 */

/**
**  FILENAME:   xllp_rtc.c
**
**  PURPOSE:    contains XLLP RTC primitives.
**
******************************************************************************/

#include "xllp_defs.h"
#include "xllp_rtc.h"

//
/// Configures the RTC Alarm to cause an interrupt at the designated time.
/// Upon completion of this routine, the RTC alarm is armed and enabled.
///
/// @param
///  pRTCReg		RTC base register addresses.
///  alarmtime		the unsigned 32-bit value to assign to the RTC alarm register.
///
/// @remarks
///		Not touch the Interrupt Controller, so enabling Interrupt Controller should be done by Caller. 
///		In Monahan, interrupt controller can be accessed only by coprocessor 6. So need supervise mode. 
///		Palm OS has interrupt controller node, which also dont expected bypass and directly call 
///		another module xllp function
///
void XllpRtcConfigureAlarm (P_XLLP_RTC_T pRTCReg, XLLP_UINT32_T alarmtime)
{
    P_XLLP_RTC_T      pRTCRegs =  pRTCReg;
    

    //disable the RTC alarm interrupt
    pRTCRegs->rtsr &= (~(XLLP_RTSR_MASK | XLLP_RTSR_RESERVED_BITS | XLLP_RTSR_ALE));

    //clear any current RTC alarm interrupt
    pRTCRegs->rtsr &= (~(XLLP_RTSR_MASK | XLLP_RTSR_RESERVED_BITS) | XLLP_RTSR_AL);

    //
    // Set the Alarm value
    //
    pRTCRegs->rtar = alarmtime;

    // Enable the RTC alarm interrupt
    pRTCRegs->rtsr = (pRTCRegs->rtsr & ~(XLLP_RTSR_MASK | XLLP_RTSR_RESERVED_BITS)) | XLLP_RTSR_ALE;
    return;
}
//
/// Configures the RTC Alarm to cause an interrupt at the designated time.
/// Upon completion of this routine, the RTC alarm is armed and enabled.
///
/// @param
///  pRTCReg		RTC base register addresses.
///  alarmtime	the time between the interrupt and now .
///
/// @remarks
///		Not touch the Interrupt Controller, so enabling Interrupt Controller should be done by Caller. 
///		In Monahan, interrupt controller can be accessed only by coprocessor 6. So need supervise mode. 
///		Palm OS has interrupt controller node, which also dont expected bypass and directly call 
///		another module xllp function
///
void XllpRtcConfigureIncreaseAlarm (P_XLLP_RTC_T pRTCReg, XLLP_UINT32_T increase_time)
{
	XllpRtcConfigureAlarm (pRTCReg, pRTCReg->rcnr + increase_time);
}

/// Save state before system enter sleep state
///
/// @param 
///     pVirtual    RTC base register addresses
///     pSave       Buff to save status
///
/// @return
///     XLLP_STATUS_SUCCESS   -   Successful
///     XLLP_STATUS_WRONG_PARAMETER   -   Fail
XLLP_STATUS_T XllpRTCSave(P_XLLP_RTC_T pVirtual, P_XLLP_RTC_T pSave)
{
    if(!pVirtual)
        return XLLP_STATUS_WRONG_PARAMETER;

    if(!pSave)
        return XLLP_STATUS_WRONG_PARAMETER;

    *pSave = *pVirtual;

    return XLLP_STATUS_SUCCESS;
}

/// restore state After system wakeup
///
/// @param 
///     pVirtual    RTC base register addresses
///     pSave       Buff to save status
///
/// @return
///     XLLP_STATUS_SUCCESS   -   Successful
///     XLLP_STATUS_WRONG_PARAMETER   -   Fail
XLLP_STATUS_T XllpRTCRestore(P_XLLP_RTC_T pVirtual, P_XLLP_RTC_T pSave)
{
    if(!pVirtual)
        return XLLP_STATUS_WRONG_PARAMETER;

    if(!pSave)
        return XLLP_STATUS_WRONG_PARAMETER;

    *pVirtual=*pSave;

    return XLLP_STATUS_SUCCESS;
}
