/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *(C) Copyright 2006 Marvell International Ltd.  
 * All Rights Reserved 
 */

/**
**  FILENAME:       xllp_uart_proc.c
**
**  PURPOSE:        Definitions the low-level primitives that access the UART
**
******************************************************************************/

#include "xllp_defs.h"
#include "xllp_serialization.h"

#include "xllp_uart_proc.h"

/******************************************************************************
 * 	UartProcInit
 * 	This function initialize the part related with processor.
 *	This function is internal function.
 * 	
 * 	Param:
 * 		Handle: pointer to XLLP_UART_HANDLE_T
 * 	Return Value:
 * 		XLLP_STATUS_T	
 ******************************************************************************
 */
XLLP_STATUS_T	UartProcInit(P_XLLP_UART_HANDLE_T Handle)
{
	XLLP_VUINT32_T tmp;	// need by flushing registers.

	// UART, MFP register, MFP resource DB ID.
	P_XLLP_UART_T pUARTRegs = Handle->pUARTRegs;
	XLLP_UINT32_T	Mask = Handle->Mask;

	// Configure the UART registers to Reset Values.
	// Ignore TBR, IIR, FOR and ACR.
	pUARTRegs->lcr = 0;
	pUARTRegs->ier_dlh = 0;

	//Reset Tx/Rx FIFOs
	pUARTRegs->iir_fcr = (XLLP_FCR_RESETRF | XLLP_FCR_RESETTF);

	pUARTRegs->spr = 0;
	tmp = pUARTRegs->rbr_thr_dll;	// Read RBR to empty.
	tmp = pUARTRegs->lsr;	// Read LSR to clear status bits.
	tmp = pUARTRegs->msr;	// Read MSR to clear status bits.

	// Set Baud Rate.
	if (Mask & XLLP_DLL_DLH_SET) {
		//Access the DLL and DLH Registers.
		pUARTRegs->lcr |= XLLP_LCR_DLAB;	
		pUARTRegs->rbr_thr_dll = Handle->DLL_Value;
		pUARTRegs->ier_dlh = Handle->DLH_Value;
		pUARTRegs->lcr &= ~XLLP_LCR_DLAB;	
	}

	// Set LCR to user specified values.
	// Note Divisor Latch Access Bit, and 
	// Set Break fields are always cleared.
	if (Mask & XLLP_LCR_SET) {
		pUARTRegs->lcr = Handle->LCR_Value & (XLLP_LCR_STKYP 
			| XLLP_LCR_EPS | XLLP_LCR_PEN | XLLP_LCR_STB 
			| XLLP_LCR_WLS8);
	}

	// Set FCR to user specified values.
	// Note RESETRF and RESETTF fields can't be modified, 
	// but they were set earlier in this funcion.
	if (Mask & XLLP_FCR_SET) {
		pUARTRegs->iir_fcr = Handle->FCR_Value & (XLLP_FCR_ITL32 
			| XLLP_FCR_BUS | XLLP_FCR_TRAIL | XLLP_FCR_TIL 
			| XLLP_FCR_TRFIFOE);
	}

	// Set ABR to user specified values.
	if (Mask & XLLP_ABR_SET) {
		pUARTRegs->abr = Handle->ABR_Value & (XLLP_ABR_ABT 
			| XLLP_ABR_ABUP | XLLP_ABR_ABLIE | XLLP_ABR_ABE);
	}

	// Set MCR to user specified values.
	if (Mask & XLLP_MCR_SET) {
		pUARTRegs->mcr = Handle->MCR_Value & (XLLP_MCR_AFE 
			| XLLP_MCR_LOOP | XLLP_MCR_OUT2 | XLLP_MCR_RTS 
			| XLLP_MCR_DTR);
	}

	// Set ISR to user specified values.
	if (Mask & XLLP_ISR_SET) {
		pUARTRegs->isr = Handle->ISR_Value & (XLLP_ISR_XMITIR 
			| XLLP_ISR_RCVEIR | XLLP_ISR_XMODE | XLLP_ISR_TXPL 
			| XLLP_ISR_RXPL);
	}

	// Set IER to user specified values.
	if (Mask & XLLP_IER_SET) {
		pUARTRegs->ier_dlh = Handle->IER_Value & (XLLP_IER_DMAE 
			| XLLP_IER_UUE | XLLP_IER_NRZE | XLLP_IER_RTOIE 
			| XLLP_IER_MIE | XLLP_IER_RLSE | XLLP_IER_TIE
			| XLLP_IER_RAVIE);
	}

	return XLLP_STATUS_SUCCESS;
}

/******************************************************************************
 * 	XllpUart32BitTx
 * 	This function transmit data using 32 bit peripheral bus.
 * 	
 * 	Note: 
 *	1. This function just put the data to UART FIFO, doesn't make sure
 * 	   the data are sent out. 
 *	2. Using this function must disable UART irq.
 *	3. Before use this function to transmit data, make sure UART is 
 *	   under 32 Bit peripheral bus mode.
 * 	
 * 	Param:
 * 		Handle: pointer to XLLP_UART_HANDLE_T
 * 		pBuf:	pointer to the data
 * 		Length:	the length of data want to be transmitted
 * 	Return Value:
 * 		The length of data sent out actually
 ******************************************************************************
 */
XLLP_INT32_T XllpUart32BitTx(P_XLLP_UART_HANDLE_T Handle, P_XLLP_UINT8_T pBuf,
			     XLLP_INT32_T Length)
{
	XLLP_INT32_T len, i, j, k;
	XLLP_INT32_T tmp = (XLLP_FIFO_SIZE / 2);
	XLLP_VUINT32_T data;
	P_XLLP_UINT8_T buf = pBuf;
	P_XLLP_UART_T pUARTRegs = Handle->pUARTRegs;


	// Calculate the number of bytes before word-aligned starting address
	k = ((XLLP_UINT32_T)pBuf) % 4;
	k = (k) ? (4-k) : 0;

	// The max length of data sent out is uart_fifo_size/2 
	len = (Length > tmp) ? tmp : Length;
	k = (k<len) ? k : len;
	i = (len-k)/4;
	j = (len-k)%4;

	while (k > 0) {
		// for 8 bit access to uart THR
		*((P_XLLP_UINT8_T) (&(pUARTRegs->rbr_thr_dll))) = *buf;
		++buf;
		--k;
	}

	while (i > 0) {
		data = *(P_XLLP_UINT32_T) buf;
		pUARTRegs->rbr_thr_dll = data;
		--i;
		buf += 4;
	}

	while (j > 0) {
		// for 8 bit access to uart THR
		*((P_XLLP_UINT8_T) (&(pUARTRegs->rbr_thr_dll))) = *buf;
		++buf;
		--j;
	}
	return len;
}				// end of XllpUart32BitTx()

/******************************************************************************
 * 	XllpUart32BitRx
 * 	This function receive data using 32 bit peripheral bus.
 * 	
 * 	Note: 
 *	1. Using this function must disable UART irq.
 *	2. Before use this function to receive data, make sure UART is
 *         under 32 Bit peripheral bus mode.
 * 	
 * 	Param:
 * 		Handle: pointer to XLLP_UART_HANDLE_T
 * 		pBuf:	pointer to the data
 * 		Length:	the length of receive data 
 * 	Return Value:
 * 		The length of data received actually
 ******************************************************************************
 */
XLLP_INT32_T XllpUart32BitRx(P_XLLP_UART_HANDLE_T Handle, P_XLLP_UINT8_T pBuf,
			     XLLP_INT32_T Length)
{
	XLLP_INT32_T len, i, j, k;
	XLLP_VUINT32_T data;
	P_XLLP_UINT8_T buf = pBuf;
	P_XLLP_UART_T pUARTRegs = Handle->pUARTRegs;

	// Actually FOR register just has bit for 63 bytes instead of
	// 64 bytes. So we check whether data ready or not first.
	// If data ready and (0 == FOR&0x3f), we think FIFO has 64 bytes
	// data. FIXME: this maybe be moved to driver.
	if (!(pUARTRegs->lsr & XLLP_LSR_DR)) 
		return 0;
	len = pUARTRegs->forr & XLLP_FOR_BYTECOUNT;
	len = (0 == len) ? XLLP_FIFO_SIZE : len; 

	// Calulate the number of bytes before word-aligned starting address
	k = ((XLLP_UINT32_T)pBuf) % 4;
	k = (k) ? (4-k) : 0;

	// The actually receive data is the min of FIFO occupancy and 
	// buffer length
	len = (len > Length) ? Length : len;
	k = (k<len) ? k : len;
	i = (len-k)/4;
	j = (len-k)%4;

	while (k > 0) {
		// for 8 bit access to uart THR
		*buf = *((P_XLLP_UINT8_T) (&(pUARTRegs->rbr_thr_dll)));
		++buf;
		--k;
	}

	while (i > 0) {
		data = pUARTRegs->rbr_thr_dll;
		*(P_XLLP_UINT32_T) buf = data;
		--i;
		buf += 4;
	}

	while (j > 0) {
		// for 8 bit access to uart THR
		*buf = *((P_XLLP_UINT8_T) (&(pUARTRegs->rbr_thr_dll)));
		++buf;
		--j;
	}
	return len;
}				// end of XllpUart32BitRx()
