/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *(C) Copyright 2006 Marvell International Ltd.  
 * All Rights Reserved 
 */

/**
**  FILENAME:       xllp_udc.c
**
**  PURPOSE:   		This files contains the driver for USB Device controller 
**                  (UDC).
**
**
******************************************************************************/

/*
*******************************************************************************
*   HEADER FILES
*******************************************************************************
*/                                                                      

#include "xllp_udc.h"

/******************************************************************************

  Function Name: 
    XllpUdcConfigureEndpoints

  Description: 
    This function configures and enables all of the Endpoints that are
    used for each of the USB Configurations and Interfaces.

  Global Register Modified: 
    None 

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle

  Output Arguments:
    None
          
  Return Value: 
    XLLP_STATUS_T - error code if failed, zero if success

*******************************************************************************/

XLLP_STATUS_T XllpUdcConfigureEndpoints (P_XLLP_UDC_T pUdcHandle)
{
    P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;    
    P_XLLP_UDC_EP_CONFIG_TABLE_T pConfigTable = (P_XLLP_UDC_EP_CONFIG_TABLE_T)
                                                           pUdcHandle->pConfigTable;
    XLLP_UINT32_T i;

    // Program the Endpoint Configuration Registers:
    // Disable UDC controller by writing zero to the UDE bit (UDCCR.UDE = 0),
    // that enables the Endpoint Configuration Registers for write operation.
    pRegs->udccr = 0;
    // Endpoint 0 always maps to udc ep0
    pUdcHandle->usb2UDC[0] = 0;

    // Invalidate old USB/UDC endpoints mapping.
    for(i=1;i<XLLP_UDC_USB_MAX_EP_NUM;i++)pUdcHandle->usb2UDC[i] = ENDPOINT_INVALID;  

    // Clear the Endpoints transfer structures
//	memset (&pUdcHandle->udcEpXferTable[0], 0, (sizeof(XLLP_UDC_XFER_T) * XLLP_UDC_USB_MAX_EP_NUM));
       	
    i = 1;
    while (pConfigTable[i].usbEndpointNum != 0)
    {
        // Compute the values that would be loaded to the Endpoints Configuration 
        // Registers UDCCRA-UDCCRX, from the data in the Endpoints Configuration Table.
        // Write the configuration value to the Endpoints Configuration Registers UDCCRA-UDCCRX.
        // Remember UDCCRZ[0] is reserved and i starts from 1.
        pRegs->udccrz[i] =   ((pConfigTable[i].usbConfigNum)<<XLLP_UDC_UDCCRZ_CN_SHIFT)
        					|((pConfigTable[i].usbInterfaceNum)<<XLLP_UDC_UDCCRZ_IN_SHIFT)
        					|((pConfigTable[i].usbIntAltSettingsNum)<<XLLP_UDC_UDCCRZ_AISN_SHIFT)
        					|((pConfigTable[i].usbEndpointNum)<<XLLP_UDC_UDCCRZ_EN_SHIFT)
        					|((pConfigTable[i].usbEndpointType)<<XLLP_UDC_UDCCRZ_ET_SHIFT)
        					|((pConfigTable[i].usbEndpointDirection)<<XLLP_UDC_UDCCRZ_ED_SHIFT)
        					|((pConfigTable[i].maxPacketSize)<<XLLP_UDC_UDCCRZ_MPS_SHIFT)
        					|((pConfigTable[i].doubleBuffEnabled)<<XLLP_UDC_UDCCRZ_DE_SHIFT)
        					|((pConfigTable[i].endpointEnabled));
        
        // Assume no conflicts of ep number between configurations. 
        // It'd be better if SetConfigure/InterfaceCallback are used.
       	if(pUdcHandle->usb2UDC[pConfigTable[i].usbEndpointNum] == ENDPOINT_INVALID)
       		pUdcHandle->usb2UDC[pConfigTable[i].usbEndpointNum] = i;
       		
        ++i;    
    }

    return XLLP_STATUS_SUCCESS;
}

XLLP_STATUS_T XllpUdcEnable (P_XLLP_UDC_T pUdcHandle)
{
    P_XLLP_UDC_REGISTERS_T pRegs = pUdcHandle->pRegsBase;    
	
    // Clear all interrupt status.
    pRegs->udcisr0 = 0xffffffff;
    pRegs->udcisr1 = 0xf800ffff;
    
    // Clear configure information in pUdcHandle
  	pUdcHandle->usbConfigNumActive = 0;
    pUdcHandle->usbInterfaceNumChanged = 0;
    pUdcHandle->usbIntAltSettingsNumSelected = 0;
    
    // Enable UDC controller by writing one to the bit UDE (UDCCR.UDE = 1),
    // that disables the Endpoint Configuration Registers for write operation.
    if (!(pRegs->udccr & XLLP_UDC_UDCCR_UDE))
    {
	    // There's a write-1-clear bit (EMCE) in UDCCR.
	    // We do want to clear it now.
    	pRegs->udccr |= XLLP_UDC_UDCCR_UDE;
    }
	
    // FIXME: We don't know how long is necessary for UDC to update EMCE status.
   
    // Check if the Endpoint Memory Configuration is valid by reading the status of
    // the Endpoint Memory Configuration Error UDCCR.EMCE bit. 
    if (pRegs->udccr & XLLP_UDC_UDCCR_EMCE)
    {
        // This bit is set if the endpoint memory configuration has an error. 
        // The driver must clear UDCCR.EMCE bit by writing one to it before again enabling the UDC.
        pRegs->udccr = XLLP_UDC_UDCCR_EMCE;
        return XLLP_STATUS_WRONG_PARAMETER;
    }
    return XLLP_STATUS_SUCCESS;	
}

XLLP_STATUS_T XllpUdcDisable (P_XLLP_UDC_T pUdcHandle)
{
    // Disable UDC controller by writing zero to the UDE bit (UDCCR.UDE = 0),
    // There's a write-1-clear bit in UDCCR. Keep it.
    pUdcHandle->pRegsBase->udccr = 0;
  
    return XLLP_STATUS_SUCCESS;	
}

/******************************************************************************

  Function Name: 
    XllpUdcEnableInterrupt

  Description: 
    This function enables the Endpoints or Events interrupt.

  Global Register Modified: 
    None 

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle
    XLLP_UDC_INTERRUPTS_T udcEndpointNum - an endpoint or event number
    XLLP_UDC_EP_INTERRUPT_TYPE_T interruptType - PC or FIFO error interrupts

  Output Arguments:
    None
          
  Return Value: 
    None

*******************************************************************************/

void XllpUdcEnableInterrupt (P_XLLP_UDC_T pUdcHandle, 
                            XLLP_UDC_INTERRUPTS_T udcEndpointNum,
                            XLLP_UDC_EP_INTERRUPT_TYPE_T interruptType)
{
    XLLP_UINT32_T bitNum;
    volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;    
    
    // Set the corresponding bit in the UDC's interrupt control registers UDCICR0 and UDCICR1
    // to enable the endpoint's and event's interrupt requests.
    if (udcEndpointNum < INT_ENDPOINT_Q)
    {
 	    bitNum = (udcEndpointNum << 1) | interruptType;
   	    pRegs->udcicr0 |= (1 << bitNum);		  
 	}														  
    else if (udcEndpointNum < INT_ENDPOINT_X + 1)
    {														  
     	udcEndpointNum -= INT_ENDPOINT_Q;
     	bitNum = ((udcEndpointNum  & 0xff) << 1) | interruptType;
   	    pRegs->udcicr1 |= (1 << bitNum);		  
    }
    else if (udcEndpointNum >= INT_RESET)
    {
   	    bitNum = (udcEndpointNum & 0x1f);
   	    pRegs->udcicr1 |= (1 << bitNum);		  
    }
}

/******************************************************************************

  Function Name: 
    XllpUdcDisableInterrupt

  Description: 
    This function disables the Endpoints or Events interrupts.

  Global Register Modified: 
    None 

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle
    XLLP_UDC_INTERRUPTS_T udcEndpointNum - an endpoint or event number
    XLLP_UDC_EP_INTERRUPT_TYPE_T interruptType - PC or FIFO error interrupts

  Output Arguments:
    None
          
  Return Value: 
    None

*******************************************************************************/

void XllpUdcDisableInterrupt (P_XLLP_UDC_T pUdcHandle, 
                            XLLP_UDC_INTERRUPTS_T udcEndpointNum,
                            XLLP_UDC_EP_INTERRUPT_TYPE_T interruptType)
{
    XLLP_UINT32_T bitNum;
    volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;    

    // Clear the corresponding bit in the UDC's interrupt control registers UDCICR0 and UDCICR1
    // to disable the endpoint's and event's interrupt requests.
    if (udcEndpointNum < INT_ENDPOINT_Q)
    {
 	    bitNum = (udcEndpointNum << 1) | interruptType;
        pRegs->udcicr0 &= ~(1 << bitNum);		  
 	}														  
    else if (udcEndpointNum < (INT_ENDPOINT_X + 1))
    {														  
     	udcEndpointNum -= INT_ENDPOINT_Q;
     	bitNum = ((udcEndpointNum  & 0xff) << 1) | interruptType;
        pRegs->udcicr1 &= ~(1 << bitNum);		  
    }
    else if (udcEndpointNum >= INT_RESET)
    {
     	bitNum = (udcEndpointNum  & 0x1f);
        pRegs->udcicr1 &= ~(1 << bitNum);		  
    }
}

/******************************************************************************

  Function Name: 
    XllpUdcClearInterrupt

  Description: 
    This function clears the Endpoints or Events interrupts.
    It also clears the Endpoint FIFO Error bit in the Endpoint's UDCCSR register.

  Global Register Modified: 
    None 

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle
    XLLP_UDC_INTERRUPTS_T udcEndpointNum - an endpoint or event number
    XLLP_UDC_EP_INTERRUPT_TYPE_T interruptType - PC or FIFO error interrupts

  Output Arguments:
    None
          
  Return Value: 
    None

*******************************************************************************/

void XllpUdcClearInterrupt (P_XLLP_UDC_T pUdcHandle,
                            XLLP_UDC_INTERRUPTS_T udcEndpointNum,
                            XLLP_UDC_EP_INTERRUPT_TYPE_T interruptType)
{
    XLLP_UINT32_T bitNum;
    volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;    

    // Write one to the corresponding bit of the UDC's interrupt status registers 
    // UDCISR0 and UDCISR1 to clear the endpoint's and event's interrupt.
    if (udcEndpointNum < INT_ENDPOINT_Q)
    {
 	    bitNum = (udcEndpointNum << 1) | interruptType;
        pRegs->udcisr0 = (1 << bitNum);
        
        // Clear the Endpoint FIFO Error bit in the Endpoint's UDCCSR register
        pRegs->udccsr[udcEndpointNum] = 
          (pRegs->udccsr[udcEndpointNum] & XLLP_UDC_UDCCSR_DME) | XLLP_UDC_UDCCSR_EFE; 
 	}														  
    else if (udcEndpointNum < (INT_ENDPOINT_X + 1))
    {														  
     	udcEndpointNum -= INT_ENDPOINT_Q;
     	bitNum = ((udcEndpointNum  & 0xff) << 1) | interruptType;
        pRegs->udcisr1 = (1 << bitNum);
        
        // Clear the Endpoint FIFO Error bit in the Endpoint's UDCCSR register
        pRegs->udccsr[udcEndpointNum] = 
          (pRegs->udccsr[udcEndpointNum] & XLLP_UDC_UDCCSR_DME) | XLLP_UDC_UDCCSR_EFE; 
    }
    else if (udcEndpointNum >= INT_RESET)
    {
     	bitNum = (udcEndpointNum  & 0x1f);
        pRegs->udcisr1 = (1 << bitNum);
    }
}

/******************************************************************************

  Function Name: 
    XllpUdcGetInterruptStatus

  Description: 
    This function returns the status of the Endpoint's or Event's interrupts.

  Global Register Modified: 
    None 

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle
    XLLP_UDC_INTERRUPTS_T udcEndpointNum - an endpoint or event number
    XLLP_UDC_EP_INTERRUPT_TYPE_T interruptType - PC or FIFO error interrupts

  Output Arguments:
    P_XLLP_UINT32_T intStatus - the status of the Endpoint's or Event's interrupt.
                                One would be returned, if the corresponding bit in the
                                UDC interrupt status register is set.
  Return Value: 
    None

*******************************************************************************/

void XllpUdcGetInterruptStatus (P_XLLP_UDC_T pUdcHandle,
                               XLLP_UDC_INTERRUPTS_T udcEndpointNum,
                               XLLP_UDC_EP_INTERRUPT_TYPE_T interruptType,  
                               P_XLLP_UINT32_T intStatus)
{
    XLLP_UINT32_T bitNum;
    volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;    

    // Read status one of the corresponding bits in the UDC's interrupt status registers 
    // UDCISR0 and UDCISR1.
    if (udcEndpointNum < INT_ENDPOINT_Q)
    {
 	    bitNum = (udcEndpointNum << 1) | interruptType;
        *intStatus = (pRegs->udcisr0 >> bitNum) & 1;
 	}														  
    else if (udcEndpointNum < (INT_ENDPOINT_X + 1))
    {														  
     	udcEndpointNum -= INT_ENDPOINT_Q;
     	bitNum = ((udcEndpointNum  & 0xff) << 1) | interruptType;
 	    *intStatus = (pRegs->udcisr1 >> bitNum) & 1;

    }
    else if (udcEndpointNum >= INT_RESET)
    {
     	bitNum = (udcEndpointNum  & 0x1f);
 	    *intStatus = (pRegs->udcisr1 >> bitNum) & 1;
    }
}

/******************************************************************************

  Function Name: 
    XllpUdcWriteDwordFifo

  Description: 
    This function writes the specified number of bytes to the Endpoint's FIFO.

  Global Register Modified: 
    None 

  Input Arguments:
    P_XLLP_VUINT32_T pReg    - a pointer to the Endpoints FIFO
    void *src               - a pointer to the data buffer
    XLLP_INT32_T len        - number of bytes to write

  Output Arguments:
    None

  Return Value: 
    None

*******************************************************************************/

void XllpUdcWriteDwordFifo(P_XLLP_VUINT32_T pReg, P_XLLP_UINT8_T src, XLLP_INT32_T len)
{
	XLLP_INT32_T    i;
	XLLP_UINT32_T	temp[16];
	P_XLLP_UINT32_T p1;
    P_XLLP_VUINT8_T pBuff, pTemp;
	
	if (len == 0) {
		return;
	}
	
	if (len > sizeof(temp) && ((XLLP_UINT32_T)src & 0x3))
	{
		while(len > 0) {
			XllpUdcWriteDwordFifo(pReg, src, (len>sizeof(temp))?sizeof(temp):len);
			len -= sizeof(temp);
		}
		return;
	}
	
	if ((XLLP_UINT32_T) src & 0x3) {
		i = len;
		pBuff = (P_XLLP_VUINT8_T)src;
		pTemp = (P_XLLP_VUINT8_T)temp;
		while(i-- > 0) *(pTemp+i) = *(pBuff +i);
//		memcpy (temp, src, len);
		p1 = temp;
	} else {
		p1 = (P_XLLP_UINT32_T)src;
	}
	
	if (((XLLP_INT32_T) len & 0x3) == 0)
	{
		// 4 byte aligned
		for (i=0; i<(len>>2); i++)
		{
			*pReg = *p1++;
	    }
	}
	else 
	{
		for (i=0; i<(len>>2); i++)
		{
			// all words
			*pReg = *p1++;
	    }

        pBuff = (P_XLLP_VUINT8_T)p1;
        pTemp = (P_XLLP_VUINT8_T)pReg;
	    for (i=0; i<(len&0x3); i++)
	    {
		    *pTemp = *pBuff++;
	    }
	}	
}


/******************************************************************************

  Function Name: 
    XllpUdcReadDwordFifo

  Description: 
    This function reads the specified number of bytes to the Endpoint's FIFO.

  Global Register Modified: 
    None 

  Input Arguments:
    P_XLLP_VUINT32_T pReg    - a pointer to the Endpoints FIFO
    void *buf               - a pointer to the data buffer
    XLLP_INT32_T len        - number of bytes to read

  Output Arguments:
    None

  Return Value: 
    None

*******************************************************************************/

void XllpUdcReadDwordFifo (P_XLLP_VUINT32_T pReg, P_XLLP_UINT8_T buf, XLLP_INT32_T len)
{
	XLLP_INT32_T    i;
	XLLP_UINT32_T	temp[16], rdat;
	P_XLLP_UINT32_T p1;
    XLLP_BOOL_T     copyback;
    P_XLLP_VUINT8_T pBuff, pTemp;
    	
	if (((XLLP_UINT32_T) buf & 0x3) != 0  && len > sizeof(temp))
	{
		while(len > 0) {
			XllpUdcReadDwordFifo(pReg, buf, (len>sizeof(temp))?sizeof(temp):len);
			len -= sizeof(temp);
		}
		return;
	}
	
	if (((XLLP_UINT32_T)buf & 0x3) != 0)
	{
		p1 = temp;
		copyback = XLLP_TRUE;
	}
	else 
	{
		p1 = (P_XLLP_UINT32_T)buf;
		copyback = XLLP_FALSE;
	}
	
	
	if (((XLLP_INT32_T)len & 0x3) == 0)
	{
		// 4 byte aligned
		for (i=0; i<(len>>2); i++) 
		{
			*p1++ = *pReg;
	    }
	}
	else 
	{
		for (i=0; i<(len>>2); i++)
		{
			// all words
			*p1++ = *pReg;
	    }
	    rdat = *pReg;
        
        pBuff = (P_XLLP_VUINT8_T)p1;
        pTemp = (P_XLLP_VUINT8_T)&rdat;
	    
	    for (i=0; i<(len&0x3); i++)
	    {
           *pBuff++ = *pTemp++;
	    }
	}	
	
	if (copyback)
	{
		i = len;
		pBuff = (P_XLLP_VUINT8_T)buf;
		pTemp = (P_XLLP_VUINT8_T)temp;
		while(i-- > 0) *(pBuff + i) = *(pTemp + i);
//		memcpy (buf, temp, len);
	}
}

// This function should be called when config change interrupt occurs.
// It will update the USB/UDC endpoints mapping table.
void XllpUdcConfigChangeCallback (P_XLLP_UDC_T pUdcHandle)
{
	XLLP_UINT32_T i, config, udcCR;
    P_XLLP_UDC_EP_CONFIG_TABLE_T pConfigTable = (P_XLLP_UDC_EP_CONFIG_TABLE_T)
                                                           pUdcHandle->pConfigTable;

    // Clear USB endpoint to UDC endpoint mapping.
    for(i=1;i<XLLP_UDC_USB_MAX_EP_NUM;i++)pUdcHandle->usb2UDC[i] = ENDPOINT_INVALID;  

	udcCR = pUdcHandle->pRegsBase->udccr;

	// Read current active configuration number.
	config = (udcCR & XLLP_UDC_UDCCR_ACN_MASK)>>XLLP_UDC_UDCCR_ACN_SHIFT;
	pUdcHandle->usbConfigNumActive = ( XLLP_UINT8_T )config;
    
    // Read latest changed interface number.
    pUdcHandle->usbInterfaceNumChanged = (XLLP_UINT8_T) ((udcCR & XLLP_UDC_UDCCR_AIN_MASK) >> XLLP_UDC_UDCCR_AIN_SHIFT);
    
    // Read selected alt setting for that interface.
    pUdcHandle->usbIntAltSettingsNumSelected = (XLLP_UINT8_T) ((udcCR & XLLP_UDC_UDCCR_AAISN_MASK) >> XLLP_UDC_UDCCR_AAISN_SHIFT);

    // Switch Endpoint memory to active configuration
    pUdcHandle->pRegsBase->udccr |= XLLP_UDC_UDCCR_SMAC;

	// Setup a new mapping table.
	i = 1;
	while(pConfigTable[i].usbEndpointNum != 0)
	{
		if (pConfigTable[i].usbConfigNum == config)
			pUdcHandle->usb2UDC[pConfigTable[i].usbEndpointNum] = i;
		i++;
	}
	
	return; 
}

void XllpUdcInitTransfer (P_XLLP_UDC_T pUdcHandle,
				  XLLP_UDC_EP_T udcEndpointNum, 
				  P_XLLP_UINT8_T buf, 
				  XLLP_UINT32_T len,
				  XLLP_UINT32_T dmaChannel)
{
    // Initilize the UDC's endpoint xfer structure
    pUdcHandle->udcEpXferTable[udcEndpointNum].pDataEp = buf; 
    pUdcHandle->udcEpXferTable[udcEndpointNum].dmaChannel = dmaChannel; 
    pUdcHandle->udcEpXferTable[udcEndpointNum].xferLength = len;
    pUdcHandle->udcEpXferTable[udcEndpointNum].xferDataCounter = 0;
    pUdcHandle->udcEpXferTable[udcEndpointNum].maxPacketSize = 
    	pUdcHandle->pConfigTable[udcEndpointNum].maxPacketSize;
    pUdcHandle->udcEpXferTable[udcEndpointNum].xferComplete = XLLP_FALSE;
    return;
}
/******************************************************************************

  Function Name: 
    XllpUdcLoadFifo

  Description:
    This function is used to load the packet to the A-X Endpoint's FIFO.
    
  Global Register Modified: 
    None. 

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle         - pointer to the UDC's handle
    XLLP_UDC_EP_T udcEndpointNum    - the UDC endpoint number
    XLLP_BOOL_T enableZLP           - a flag that controls if ZLP should be sent to the USB Host

  Output Arguments:
    None
          
  Return Value: 
    None

  Note: According to USB specification for BULK and Interrupt transfer the transfer is complete
        when the endpoint dose one of the following:
        - Has transfered exactly the amount of data expected
        - Transfers a short packet or a zero-length packet
        The different OS use different termination of the BULK and Interrupt transfer.
        The enableZLP should be set to FALSE is USB Host expected the exact number of bytes
        and to TRUE otherwize.
        For ISO transfer enableZLP should be set to TRUE

*******************************************************************************/

XLLP_UDC_XFER_RESULT_T XllpUdcLoadFifo (P_XLLP_UDC_T pUdcHandle, 
                      					XLLP_UDC_EP_T udcEndpointNum, 
                      					XLLP_BOOL_T enableZLP)

{
    volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;    
    XLLP_UINT32_T packetLength;
    P_XLLP_UINT8_T pTxBuff;
    XLLP_UINT32_T status;

    if(udcEndpointNum==0)
		return UDC_XFER_UNSUPPORT_EP0;

	// Read the UDCCSRx register
    status = pRegs->udccsr[udcEndpointNum];
	
    do {
    
    // Check for errors
    if (status & XLLP_UDC_UDCCSR_PC)
    {
        // Clear UDCCSR.PC bit
        pRegs->udccsr[udcEndpointNum] = 
              (status & XLLP_UDC_UDCCSR_RW_BITS_MASK) | XLLP_UDC_UDCCSR_PC; 

        // Check if FIFO underun took place
        if (status & XLLP_UDC_UDCCSR_EFE)  
        {
            // Clear UDCCSR.EFE bit
            pRegs->udccsr[udcEndpointNum] = 
              (status & XLLP_UDC_UDCCSR_RW_BITS_MASK) | XLLP_UDC_UDCCSR_EFE; 
        }

        // If STALL is set just return and wait for CLEAR_FEATURE to clear UDCCSR.SST bit
        if (status & XLLP_UDC_UDCCSR_SST)
        {
            // Clear the Interrupt
            XllpUdcClearInterrupt (pUdcHandle, udcEndpointNum, PACKET_COMPL_INT);
            return UDC_XFER_CONTINUE;
        }
    }

    // Transfer is complete, we are leaving.
    if (pUdcHandle->udcEpXferTable[udcEndpointNum].xferComplete == XLLP_TRUE) {
      return UDC_XFER_FINISH;
    }

    if (pUdcHandle->udcEpXferTable[udcEndpointNum].xferLength >= 
            pUdcHandle->udcEpXferTable[udcEndpointNum].maxPacketSize)
      packetLength = pUdcHandle->udcEpXferTable[udcEndpointNum].maxPacketSize;
    else
      packetLength = pUdcHandle->udcEpXferTable[udcEndpointNum].xferLength;

    // Get the pointer (in bytes) to the Tx data buffer
    pTxBuff = pUdcHandle->udcEpXferTable[udcEndpointNum].pDataEp
    		 + pUdcHandle->udcEpXferTable[udcEndpointNum].xferDataCounter;

    // Write to the Endpoint's FIFO
    XllpUdcWriteDwordFifo(&pRegs->udcdr[udcEndpointNum], pTxBuff, packetLength);
    
    // Check if this is a short packet
    if (packetLength < pUdcHandle->udcEpXferTable[udcEndpointNum].maxPacketSize)
    {
        // Set UDCCSR.SP to signal UDC that a short packet was loaded into the FIFO
        pRegs->udccsr[udcEndpointNum] = 
                        (status & XLLP_UDC_UDCCSR_RW_BITS_MASK) | XLLP_UDC_UDCCSR_SP; 

        // This is a short or zero-length packet.
        // All content include ZLP if any has been sent into FIFO.
        // Set xfer complete flag
        if (enableZLP == XLLP_TRUE)
          pUdcHandle->udcEpXferTable[udcEndpointNum].xferComplete = XLLP_TRUE;
    }

    // Update xfer length
    pUdcHandle->udcEpXferTable[udcEndpointNum].xferLength -= packetLength;
    pUdcHandle->udcEpXferTable[udcEndpointNum].xferDataCounter += packetLength;
    if ((pUdcHandle->udcEpXferTable[udcEndpointNum].xferLength == 0) &&
        (enableZLP == XLLP_FALSE))
    {
        // No need to send a ZLP. So let's stop here.
        // Set xfer complete flag
        pUdcHandle->udcEpXferTable[udcEndpointNum].xferComplete = XLLP_TRUE;
        return UDC_XFER_FINISH;
    }
    // else if xferLength == 0 and ZLP enabled, then we'll go another round. 
    
    status = pRegs->udccsr[udcEndpointNum];

    } while ((status & (XLLP_UDC_UDCCSR_BNE | XLLP_UDC_UDCCSR_FS)) && 
               (pUdcHandle->udcEpXferTable[udcEndpointNum].xferComplete == XLLP_FALSE));
	return UDC_XFER_CONTINUE;
}

/******************************************************************************

  Function Name: 
    XllpUdcUnloadFifo

  Description:
    This function is used to unload the packet from the A-X Endpoint's FIFO 
    
  Global Register Modified: 
    None. 

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle         - pointer to the UDC's handle
    XLLP_UDC_EP_T udcEndpointNum    - the UDC endpoint number

  Output Arguments:
    None
          
  Return Value: 
    None

*******************************************************************************/

XLLP_UDC_XFER_RESULT_T XllpUdcUnloadFifo (P_XLLP_UDC_T pUdcHandle, 
                        					XLLP_UDC_EP_T udcEndpointNum) 
{
    volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;    
    P_XLLP_UINT8_T pRxBuff;
    XLLP_UINT32_T packetLength;
    XLLP_UINT32_T   status;
    
	if(udcEndpointNum==0)
		return UDC_XFER_UNSUPPORT_EP0;

    // Keep in mind that : 
    // FS will remain asserted until both buffers are empty.
    // BNE only represents current buffer.
    // UDCBCR holds the byte count for the current buffer.
    while (((status=pRegs->udccsr[udcEndpointNum]) & (XLLP_UDC_UDCCSR_PC | XLLP_UDC_UDCCSR_FS)) 
            && pUdcHandle->udcEpXferTable[udcEndpointNum].xferComplete == XLLP_FALSE)
    {
        // Set the pointer to where the data will be placed to
        pRxBuff = pUdcHandle->udcEpXferTable[udcEndpointNum].pDataEp 
        		 + pUdcHandle->udcEpXferTable[udcEndpointNum].xferDataCounter;
        
        // Check if this is zero-length packet, UDCCSRx.SP = 1 and UDCCSRx.BNE = 0
        if ((status & XLLP_UDC_UDCCSR_SP) && ((status & XLLP_UDC_UDCCSR_BNE) == 0))
        {
            // Clear UDCCSRx.PC bit
            pRegs->udccsr[udcEndpointNum] = 
                (status & XLLP_UDC_UDCCSR_RW_BITS_MASK) | XLLP_UDC_UDCCSR_PC; 
            // Update receive complete flag to signal the end of the transfer
            pUdcHandle->udcEpXferTable[udcEndpointNum].xferComplete = XLLP_TRUE;
            return UDC_XFER_FINISH;
        }
        // Check if this is a short packet, UDCCSRx.SP = 1 and UDCCSRx.BNE = 1
        else if ((status & XLLP_UDC_UDCCSR_SP) && ((status & XLLP_UDC_UDCCSR_BNE) > 0))
        {
            packetLength = pRegs->udcbcr[udcEndpointNum] & XLLP_UDC_UDCBCR_BC_MASK;
            XllpUdcReadDwordFifo (&pRegs->udcdr[udcEndpointNum], pRxBuff, packetLength);
            // Update xfer data counter
            pUdcHandle->udcEpXferTable[udcEndpointNum].xferDataCounter += packetLength;
            // Update receive complete flag to signal the end of the transfer
            pUdcHandle->udcEpXferTable[udcEndpointNum].xferComplete = XLLP_TRUE;
            // Clear UDCCSR.PC bit
            pRegs->udccsr[udcEndpointNum] = 
                  (status & XLLP_UDC_UDCCSR_RW_BITS_MASK) | XLLP_UDC_UDCCSR_PC; 
            return UDC_XFER_FINISH;
        }
        // Check if this is a max. size packet, UDCCSRx.SP = 0 and UDCCSRx.BNE = 1
        else if (((status & XLLP_UDC_UDCCSR_SP) == 0) && (status & XLLP_UDC_UDCCSR_BNE))
        {
            packetLength = pRegs->udcbcr[udcEndpointNum] & XLLP_UDC_UDCBCR_BC_MASK;
            XllpUdcReadDwordFifo ( &pRegs->udcdr[udcEndpointNum], pRxBuff, packetLength);
            // Update xfer data counter
            pUdcHandle->udcEpXferTable[udcEndpointNum].xferDataCounter += packetLength;
            
            // Clear UDCCSR.PC bit
            pRegs->udccsr[udcEndpointNum] = 
                  (status & XLLP_UDC_UDCCSR_DME) | XLLP_UDC_UDCCSR_PC; 
            
            // Check if the driver would have to notify upper layer to request more memory
            if (pUdcHandle->udcEpXferTable[udcEndpointNum].xferDataCounter >=
                                        pUdcHandle->udcEpXferTable[udcEndpointNum].xferLength)
            {
                pUdcHandle->udcEpXferTable[udcEndpointNum].xferComplete = XLLP_TRUE;
                return UDC_XFER_FINISH_AND_MORE;
            }
        }
    }
   	return UDC_XFER_CONTINUE;
}

/******************************************************************************

  Function Name: 
    XllpUdcForceEndpointStall

  Description:
    This function forces the Endpoint to send STALL to the USB Host.
    It sets the UDCCSR.FST bit so the STALL handshake will be issued to all IN tokens.

  Global Register Modified: 
    None. 

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle
    XLLP_UDC_EP_T udcEndpointNum    - the UDC endpoint number

  Output Arguments:
    None
          
  Return Value: 
    None

*******************************************************************************/

void XllpUdcForceEndpointStall (P_XLLP_UDC_T pUdcHandle,
                                XLLP_UDC_EP_T udcEndpointNum)
{
    volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;    
    
    pRegs->udccsr[udcEndpointNum] = 
          (pRegs->udccsr[udcEndpointNum] & XLLP_UDC_UDCCSR_RW_BITS_MASK) | XLLP_UDC_UDCCSR_FST; 
}

/******************************************************************************

  Function Name: 
    XllpUdcClearEndpointStall

  Description:
    This function clears the UDCCSR.SST bit, after the STALL handshake was issued 
    from the UDC.

  Global Register Modified: 
    None. 

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle
    XLLP_UDC_EP_T udcEndpointNum    - the UDC endpoint number

  Output Arguments:
    None
          
  Return Value: 
    None

*******************************************************************************/

void XllpUdcClearEndpointStall (P_XLLP_UDC_T pUdcHandle,
                                XLLP_UDC_EP_T udcEndpointNum)
{
    volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;    
    
    pRegs->udccsr[udcEndpointNum] = 
          (pRegs->udccsr[udcEndpointNum] & XLLP_UDC_UDCCSR_RW_BITS_MASK) | XLLP_UDC_UDCCSR_SST; 
}

/******************************************************************************

  Function Name: 
    XllpUdcFlushEndpointFifo

  Description:
    This function flashes the EndpointÂs FIFO. It sets the UDCCSR.FEF bit.

  Global Register Modified: 
    None. 

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle
    XLLP_UDC_EP_T udcEndpointNum    - the UDC endpoint number

  Output Arguments:
    None
          
  Return Value: 
    None

*******************************************************************************/

void XllpUdcFlushEndpointFifo (P_XLLP_UDC_T pUdcHandle,
                               XLLP_UDC_EP_T udcEndpointNum)
{
    volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;    
    
    pRegs->udccsr[udcEndpointNum] = 
          (pRegs->udccsr[udcEndpointNum] & XLLP_UDC_UDCCSR_RW_BITS_MASK) | XLLP_UDC_UDCCSR_FEF; 
}

/******************************************************************************

  Function Name: 
    XllpUdcStartDma

  Description:
    This function start UDC DMA transfer

  Global Register Modified: 
    None. 

  Input Arguments:
    P_XLLP_UDC_T pUdcHandle - a pointer to the UDC's handle
    XLLP_UDC_EP_T udcEndpointNum    - the UDC endpoint number

  Output Arguments:
    None
          
  Return Value: 
    None

*******************************************************************************/

void XllpUdcStartDma (P_XLLP_UDC_T pUdcHandle,
                               XLLP_UDC_EP_T udcEndpointNum)
{
    volatile P_XLLP_UDC_REGISTERS_T pRegs = (P_XLLP_UDC_REGISTERS_T)pUdcHandle->pRegsBase;
    P_XLLP_UDC_XFER_T pXfer = &pUdcHandle->udcEpXferTable[udcEndpointNum]; 
    P_XLLP_UDC_EP_CONFIG_TABLE_T pEpConfig = pUdcHandle->pConfigTable + udcEndpointNum;
    XLLP_DMAC_COMMAND_T cmd;
    
    cmd.aLen = (XLLP_INT16_T) pXfer->xferLength;
    cmd.aWidth = XLLP_DMAC_WIDTH_32;
    cmd.aSize = XLLP_DMAC_BURSTSIZE_32;
    cmd.aEndian = 0;
    cmd.aEndIrqEn = XLLP_TRUE; //Enable end interrupt
    cmd.aStartIrqEn = XLLP_FALSE;

    if (pEpConfig->usbEndpointDirection == XLLP_USB_IN) {
	    cmd.aFlowTrg = XLLP_TRUE;
	    cmd.aFlowSrc = XLLP_FALSE;
	    cmd.aIncTrgAddr = XLLP_FALSE;
	    cmd.aIncSrcAddr = XLLP_TRUE;
	    XllpDmacCfgChannelNoDescTransfer(pUdcHandle->pDmacRegsBase, 
			(XLLP_UINT32_T)pXfer->pDataEp,
			XLLP_UDC_UDCDR0_PHYS + udcEndpointNum * 4,
			&cmd, pXfer->dmaChannel, 
			pUdcHandle->pDmacRegsBase->drcmr1[XLLP_DMAC_USB_ENDPOINT_0+udcEndpointNum]
			);
    
    } else if (pEpConfig->usbEndpointDirection == XLLP_USB_OUT) {
	    cmd.aFlowTrg = XLLP_FALSE;
	    cmd.aFlowSrc = XLLP_TRUE;
	    cmd.aIncTrgAddr = XLLP_TRUE;
	    cmd.aIncSrcAddr = XLLP_FALSE;
	    /* Enable Short Packet Interrupt */
	    XllpUdcEnableInterrupt(pUdcHandle, udcEndpointNum, PACKET_COMPL_INT);
//	    XllpUdcEnableInterrupt(pUdcHandle, udcEndpointNum, FIFO_ERROR_INT);
            XllpDmacCfgChannelNoDescTransfer(pUdcHandle->pDmacRegsBase, 
			XLLP_UDC_UDCDR0_PHYS + udcEndpointNum * 4,
			(XLLP_UINT32_T)pXfer->pDataEp,
			&cmd, pXfer->dmaChannel, 
			pUdcHandle->pDmacRegsBase->drcmr1[XLLP_DMAC_USB_ENDPOINT_0+udcEndpointNum]
			);
    } 

    /* Enable DMA */
    pRegs->udccsr[udcEndpointNum] = XLLP_UDC_UDCCSR_DME;
    /* Call DMA controller function to confiure DMA transfer */
    XllpDmacStartTransfer(pUdcHandle->pDmacRegsBase, pXfer->dmaChannel);
}
void XllpUdcCancelDma (P_XLLP_UDC_T pUdcHandle,
                               XLLP_UDC_EP_T udcEndpointNum)
{
    P_XLLP_UDC_XFER_T pXfer = &pUdcHandle->udcEpXferTable[udcEndpointNum]; 

    XllpDmacStopTransfer(pUdcHandle->pDmacRegsBase, pXfer->dmaChannel);

    pUdcHandle->pRegsBase->udccsr[udcEndpointNum] = XLLP_UDC_UDCCSR_FEF;
}
