/*
 * sysupd.c: system update from CF/RAM card
 *
 * Update a board's firmware from a tar file located on a CF memory
 * card or in RAM.
 *
 * Copyright (C) 2001 Stefan Eletzhofer <stefan.eletzhofer@eletztrick.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ident "$Id: sysupd.c,v 1.4 2003/08/21 12:48:05 seletz Exp $"

/**********************************************************************
 * Includes
 */

#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#if defined( CONFIG_GIO_SUPPORT ) 

#include <blob/errno.h>
#include <blob/debug.h>
#include <blob/types.h>
#include <blob/util.h>

#include <blob/generic_io.h>

#include <blob/pcmcia.h>
#include <blob/cf.h>
#include <blob/ide.h>
#include <blob/tar.h>

/* FIXME: make this arch indep. */
/* these are the default images */
static char *update_files[] = {
	"blob",
	"config.tar",
	"zImage",
	"initrd.gz",
	"cramfs",
	"data.img",
	"xdata.img"
};


/**********************************************************************
 * defines
 */
#undef SYSUPD_DEBUG
#ifdef SYSUPD_DEBUG
static int dbg = 1;
#	define DBG( x, args... )	if ( dbg>=x ) printf( args )
#else
#	define DBG( x, args... )
#endif

/**********************************************************************
 * sysupd - update complete system. Flashes
 * 			blob, config, kernel, initrd and cramfs images.
 */

/* update_umage - Read image "image" from TAR io driver and flash it
 * into a BLOB partition
 */
int update_image( char *image )
{
	int ret;

	if ( !image )
		return -EINVAL;

	/* select tar file */
	ret = io_configure( "TAR", image );
	if ( ret ) {
		printf( "sysupd: can't find image '%s'. Skipping.\n", image );
		return -EINVAL;
	}

	/* select partition */
	ret = io_configure( "BLOB_PART", image );
	if ( ret ) {
		printf( "sysupd: can't find flash partition '%s'. Skipping.\n", image );
		return -EINVAL;
	}

#if !defined(SYSUPD_DUMMY_COPY)
	/* flash tha image, man! */
	ret = io_copy( "BLOB_PART", "TAR", 0 );
	if ( ret ) {
		/* ouch! */
		printf( "sysupd: error updating image %s (%d).\n", image, ret );
		return ret;
	}
#else
	DBG( 1, "%s: io_copy( BLOB_PART, TAR, 0 )\n", __FUNCTION__ );
#endif

	return 0;
}

int sysupd_cmd( int argc, char *argv[] )
{
	int ret;
	char **images = NULL;
	int nimages = sizeof(update_files)/sizeof(char *);
	int i=0;
	char *source = "CF";

	argv++;

	/* source io param */
	if ( argc>1 ) {
		source = *argv;
		argv++;
		argc--;
	}
	
	/* image list */
	if ( argc>1 ) {
		nimages = argc -1;
		images = &argv[0];
	} else {
		images = update_files;
	}

	/* configure CF card if CF is source */
	if ( strncmp( source, "CF", 2 ) == 0 ) {
		/* CF card is on slot 1 */
		ret = io_configure( "CF", (void *)1 );
		if ( ret ) {
			printf( "io_configure() error %d\n", ret );
			return -EINVAL;
		}
	}

	/* write chain: --->blob partition->flash */
	ret = io_chain_driver( "BLOB_PART", "FLASH" );
	if ( ret ) {
		DBG( 1, "%s: io_chain_driver() = %d.", __FUNCTION__, ret );
		return ret;
	}

	/* read chain: <---TAR<--source(CF or RAM) */
	ret = io_chain_driver( "TAR", source );
	if ( ret ) {
		DBG( 1, "%s: io_chain_driver( TAR, %s ) = %d.", __FUNCTION__, source, ret );
		return ret;
	}

	/* update all images in list */
	for ( i=0; i<nimages; i++ ) {
		DBG( 1, "images[%d]=%s\n", i, images[i] );
		ret = update_image( images[i] );
		if ( ret == 0 ) 
			printf( "sysupd: flashed image '%s'\n", images[i] );
		DBG( 1, "i=%d\n", i );
	}

	return 0;
}
char sysupd_help[] = "sysupd [CF|RAM] [image [image]...]\n\nUpdate board firmware from CF card or RAM.\n"
	"You have to have a update CF card inserted (for CF updates) _or_ uploaded a\n"
	"update image to ram.";

#endif

