/*
 *  arp.c
 *
 *  Copyright 1996-2001 Russell King
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
//#include "init.h"
#include <blob/util.h>
#include <blob/serial.h>
#include <net/net.h>
#include <net/ether.h>
#include <net/arp.h>

#undef ARP_DEBUG
#ifdef ARP_DEBUG
static int dbg = 1;
#	define DBG( x, args... )	if ( dbg>=x ) printf( args )
#else
#	define DBG( x, args... ) ;
#endif

static const unsigned char arp_txpacket[] =
	{					/* destination address			*/
						/* source address			*/
	 0x08, 0x06,				/* proto				*/
	 /* arp */
	 0x00, 0x01,				/* hardware format: ethernet		*/
	 0x08, 0x00,				/* protocol format: ip			*/
	 0x06,					/* hardware length			*/
	 0x04,					/* protocol length			*/
	 0x00, 0x01				/* arp request				*/
	};

static unsigned char arp_tx[64];
static unsigned char arp_rx[64];

int arp_rcv (unsigned char *thwaddress)
{
	unsigned int size;

	size = ether_rcv (arp_rx, sizeof(arp_rx));
	if (!size)
		return 0;

	if (!memeq (arp_rx + 12, arp_tx + 12, 9))
		return 0;			/* not valid reply			*/

	if (arp_rx[21] != 2)
		return 0;			/* reply				*/

	if (!memeq (arp_rx + 28, arp_tx + 38, 4))
		return 0;			/* target ip address			*/

	memcpy (thwaddress, arp_rx + 22, 6);
	return 1;
}

int do_arp(unsigned char *thwaddress, unsigned char *tipaddress)
{
	int i;
	unsigned char *p = arp_tx;
	int retry = 8;

	memzero (p, sizeof (arp_tx));
	memset  (p, 0xff, 6);
	memcpy  (p + 12, arp_txpacket, sizeof (arp_txpacket));
	memcpy  (p + 22, hwaddress, 6);
	memcpy  (p + 28, clientipaddress, 4);
	memcpy  (p + 38, tipaddress, 4);

	do {
		ether_snd (arp_tx, sizeof(arp_tx));

		//for (i = 0; i < 0x00040000; i++) {
		for (i = 0; i < 0x40; i++) {
			if (arp_rcv (thwaddress))
				return 0;
		}
	} while (--retry);

	return -1;
}

int do_arp_reply(unsigned char __attribute__((unused)) *rx_pkt, unsigned int __attribute__((unused)) size)
{
	struct ether_hdr *eth_request = (struct ether_hdr *)rx_pkt;
	struct arp_packet *request = (struct arp_packet *)(rx_pkt+sizeof(struct ether_hdr));
	u8 bcast[6] = { [0 ... 5] 0xff };
	struct arp_packet reply = {
		.hw_type = HTON( 0x0001 ),
		.proto_type = HTON( 0x0800 ),
		.hw_len = 6,
		.proto_len = 4,
		.opcode = HTON( 0x0002 )
	};
	struct ether_hdr eth_reply;
	

	DBG( 1, "%s: ETH: %02x:%02x:%02x:%02x:%02x:%02x"
			" %02x:%02x:%02x:%02x:%02x:%02x proto 0x%04x\n", __FUNCTION__,
			eth_request->dest[0], eth_request->dest[1], eth_request->dest[2],
			eth_request->dest[3], eth_request->dest[4], eth_request->dest[5],
			eth_request->src[0], eth_request->src[1], eth_request->src[2],
			eth_request->src[3], eth_request->src[4], eth_request->src[5],
			HTON( eth_request->proto ));


	/* is it for us or broadcast? */
	if ( !memeq( eth_request->dest, hwaddress, 6 ) && !memeq( eth_request->dest, bcast, 6 ))
		return -1;

	DBG( 1, "%s: opcode 0x%04x\n", __FUNCTION__, HTON( request->opcode ) );

	/* is it an arp request? */
	if ( request->opcode != HTON( 0x0001 ) )
		return -1;

	DBG( 1, "%s: got an arp request from IP %d.%d.%d.%d requesting %d.%d.%d.%d\n", __FUNCTION__,
			request->sender.ip[0],
			request->sender.ip[1],
			request->sender.ip[2],
			request->sender.ip[3],
			request->target.ip[0],
			request->target.ip[1],
			request->target.ip[2],
			request->target.ip[3]
			);

	/* requesting our IP ? */
	if ( !memeq( request->target.ip, clientipaddress, 4 ))
		return -1;

	DBG( 1, "%s: arp request for our IP\n", __FUNCTION__ );

	/* ethernet header */
	eth_reply.proto = HTON( 0x0806 );	/* ARP */
	memcpy( eth_reply.dest, eth_request->src, 6 );
	memcpy( eth_reply.src, hwaddress, 6 );

	/* arp reply */
	memcpy(reply.sender.hw, hwaddress, 6);
	memcpy(reply.sender.ip, clientipaddress, 4);
	memcpy(reply.target.hw, request->sender.hw, 6);
	memcpy(reply.target.ip, request->sender.ip, 4);

	memzero(arp_tx, sizeof(arp_tx));
	memcpy(arp_tx, &eth_reply, sizeof(struct ether_hdr));
	memcpy(arp_tx + sizeof(struct ether_hdr), &reply, sizeof(struct arp_packet));

	ether_snd(arp_tx, sizeof(arp_tx));
	return 0;
}
