/*
 *  ip_bits.c
 *
 *  Copyright 1996-2001 Russell King
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include <blob/util.h>
#include <blob/serial.h>
#include <net/net.h>
#include <net/ether.h>
#include <net/arp.h>
#include <net/ip_bits.h>

unsigned long ip_sequence = 0;

unsigned int checksum (unsigned char *ptr, int len, unsigned int cksum)
{
	int i;

	for (i = 0; i < len - 1; i += 2)
		cksum += (ptr[i] << 8) | ptr[i + 1];

	if (len & 1)
		cksum += ptr[len - 1] << 8;

	cksum = (cksum & 0xffff) + (cksum >> 16);
	cksum = (cksum & 0xffff) + (cksum >> 16);
	return cksum;
}

void printip (char *msg, unsigned char *ip)
{
	int i;
	printascii (msg);
	for (i = 0; i < 4; i++) {
		printf( "%02x", ip[i]);
		printch (i == 3 ? ' ' : '.');
	}
}

int ipbits_tx (unsigned char *ptr, unsigned int size)
{
	unsigned int cksum;

	ptr[OFF_ETHER + 12] = 8;
	ptr[OFF_ETHER + 13] = 0;
	ptr[OFF_UDP + 6] = 0;
	ptr[OFF_UDP + 7] = 0;
	ptr[OFF_IP + 4]  = ip_sequence >> 8;
	ptr[OFF_IP + 5]  = ip_sequence;
	ptr[OFF_IP + 10] = 0;
	ptr[OFF_IP + 11] = 0;
	ip_sequence ++;

	cksum = checksum (ptr + OFF_IP + 12, 8, 0x11 + size);
	cksum = ~checksum (ptr + OFF_UDP, size, cksum);
	if (cksum == 0)
		cksum = -1;
	ptr[OFF_UDP + 6] = cksum >> 8;
	ptr[OFF_UDP + 7] = cksum;

	cksum = ~checksum (ptr + OFF_IP, SIZE_IP, 0);
	if (cksum == 0)
		cksum = -1;
	ptr[OFF_IP + 10] = cksum >> 8;
	ptr[OFF_IP + 11] = cksum;

	if (size < 32)
		size = 32;
	return ether_snd (ptr, size + OFF_UDP);
}

/* handle broadcast packets. For arp requests */
int do_bcast( unsigned char *ptr, unsigned int size )
{
	struct ether_hdr *ether = (struct ether_hdr*)ptr;

	/* check whether or not ARP proto */
	if ( ether->proto != HTON(0x0806) )
		return 0;		/* not ARP */

	/* ok, its an arp. handle it. */
	return do_arp_reply( ptr, size );
}

int ipbits_rcv (unsigned char *ptr, unsigned int size, unsigned int *dsz, int sport, int dport)
{
	unsigned int cksum, hdrlen, off_udp;
	u8 bcast[6] = { [0 ... 5] 0xff };

	size = ether_rcv(ptr, size);
	if (!size)
		return 0;

	if (size < MIN_UDPPKTSIZE)	/* must be at least this size for UDP */
		return 0;

	if (memeq (ptr + OFF_ETHER, bcast, 6)) {
		return do_bcast( ptr, size );
	}

	if (!memeq (ptr + OFF_ETHER, hwaddress, 6))
		return 0;		/* not for us */

	if (((ptr[OFF_ETHER + 12] << 8) | ptr[OFF_ETHER + 13]) != 0x800)
		return 0;		/* not IP */

	hdrlen = (ptr[OFF_IP] - 0x40) * 4;
	if (hdrlen < 20 || hdrlen > 60)
		return 0;		/* not an IP packet we know about */

	cksum = checksum (ptr + OFF_IP, hdrlen, 0);
	if (cksum != 0xffff)
		return 0;		/* bad ip header checksum */

	if (ptr[OFF_IP + 9] != 0x11)
		return 0;		/* not UDP */

	off_udp = OFF_IP + hdrlen;
	hdrlen = (ptr[off_udp + 4] << 8) | ptr[off_udp + 5];

	if ((ptr[off_udp + 6] | ptr[off_udp + 7]) != 0) {
		cksum = checksum (ptr + OFF_IP + 12, 8, 0x11 + hdrlen);
		cksum = checksum (ptr + off_udp, hdrlen, cksum);
		if (cksum != 0xffff)
			return 0;	/* bad udp checksum */
	}

	if (sport != -1 && sport != ((ptr[off_udp] << 8) | ptr[off_udp + 1]))
		return 0;
	if (((ptr[off_udp + 2] << 8) | ptr[off_udp + 3]) != dport)
		return 0;		/* wrong UDP ports */
	*dsz = hdrlen - SIZE_UDP;
	return off_udp + SIZE_UDP;
}
