/*
 *  Copyright (C) Compaq Computer Corporation, 1998, 1999
 *  Copyright (C) Extenex Corporation, 2001
 *  Copyright (C) Intrinsyc, Inc., 2002
 *
 * (C) Copyright 2006 Marvell International Ltd.  
 * All Rights Reserved 
 *
 *  PXA USB controller core driver.
 *
 *  This file provides interrupt routing and overall coordination
 *  of the endpoints.
 *
 *  Please see:
 *    linux/Documentation/arm/SA1100/SA1100_USB 
 *  for more info.
 *
 *  02-May-2002
 *   Frank Becker (Intrinsyc) - derived from sa1100 usb_ctl.c
 *
 */

#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif
#include <stdio.h>
#include <types.h>
#include <blob/arch.h>
#include <blob/init.h>
#include <blob/types.h>
#include <blob/time.h>
#include <blob/pxa_usb.h>

#include "linux/errno.h"
#include "mhn_u2d_ctl.h"
#include "arch/mhn-regs.h"

#ifdef	ZYLONITE
#include "arch/zylonite.h"
#endif

#ifdef	LITTLETON
#include "arch/littleton.h"
#endif

#ifdef BENZINA
#include "arch/bez.h"
#endif

#include "xllp_u2d.h"

#undef DEBUG 
//#define DEBUG 3
#if DEBUG
static unsigned int usb_debug = DEBUG;
#else
#define usb_debug 0     /* gcc will remove all the debug code for us */
#endif

extern int u2d_info();


//////////////////////////////////////////////////////////////////////////////
// Prototypes
//////////////////////////////////////////////////////////////////////////////

int usbctl_next_state_on_event( int event );
void udc_int_hndlr(int, void *);
static void initialize_descriptors( void );
static void soft_connect_hook( int enable );
static void xcvr_init(void);
static void u2d_disable(void);
static void u2d_enable(void);
static void u2d_configure(int speed, int phase);
static void check_usb_speed(void);

#ifndef MMU_ENABLE
XLLP_DMAC_DESCRIPTOR_T	dma_desc_buf[DMA_DESC_NUM+1];
#else
XLLP_DMAC_DESCRIPTOR_T *dma_desc_buf;
int * ep1_buf;
int * ep2_buf;
#endif
P_XLLP_DMAC_DESCRIPTOR_T	ep1_dma_desc;
P_XLLP_DMAC_DESCRIPTOR_T	ep2_dma_desc;

#if CONFIG_PROC_FS
#define PROC_NODE_NAME "driver/mhnusb"
static int usbctl_read_proc(char *page, char **start, off_t off,
							int count, int *eof, void *data);
#endif

/////////////////////////////////////////////////////////////////////////////
int usb_connected = 0;
int usb_set_conf = 0;
int usb_set_intf = 0;
int usb_speed = USB_SPEED_FULL;


//////////////////////////////////////////////////////////////////////////////
// Globals
//////////////////////////////////////////////////////////////////////////////
static const char pszMe[] = "usbctl: ";
struct usb_info_t usbd_info;  /* global to ep0, usb_recv, usb_send */

/* device descriptors */
static desc_t desc;

#define MAX_STRING_DESC 16
static string_desc_t * string_desc_array[ MAX_STRING_DESC ];
static string_desc_t sd_zero;  /* special sd_zero holds language codes */

// called when configured
static usb_notify_t configured_callback = NULL;

enum {
    kStateZombie		= 0,
    kStateZombieSuspend		= 1,
    kStateDefault		= 2,
    kStateDefaultSuspend	= 3,
    kStateAddr			= 4,
    kStateAddrSuspend		= 5,
    kStateConfig		= 6,
    kStateConfigSuspend		= 7
};

XLLP_U2D_T	xllp_u2d;
#define XLLP_MFP_RM_DATABASE	((XLLP_MFP_RM_DB_ID_T)&(U2DDR0 + 4)) // reserved
//XLLP_MFP_RM_DB_ID_T XLLP_MFP_RM_DATABASE[XLLP_MFP_RM_DB_SIZE]; ?????

/*
 * FIXME: The PXA UDC handles several host device requests without user 
 * notification/intervention. The table could be collapsed quite a bit...
 */
static int device_state_machine[8][6] = {
//              suspend               reset          resume         adddr       config        deconfig
/* zombie */  { kStateZombieSuspend , kStateDefault, kStateZombie , kError    , kError      , kError },
/* zom sus */ { kStateZombieSuspend , kStateDefault, kStateZombie , kError    , kError      , kError },
/* default */ { kStateDefaultSuspend, kStateDefault, kStateDefault, kStateAddr, kStateConfig, kError },
/* def sus */ { kStateDefaultSuspend, kStateDefault, kStateDefault, kError    , kError      , kError },
/* addr */    { kStateAddrSuspend   , kStateDefault, kStateAddr   , kError    , kStateConfig, kError },
/* addr sus */{ kStateAddrSuspend   , kStateDefault, kStateAddr   , kError    , kError      , kError },
/* config */  { kStateConfigSuspend , kStateDefault, kStateConfig , kError    , kError      , kStateDefault },
/* cfg sus */ { kStateConfigSuspend , kStateDefault, kStateConfig , kError    , kError      , kError }
};

/* "device state" is the usb device framework state, as opposed to the
   "state machine state" which is whatever the driver needs and is much
   more fine grained
*/
static int sm_state_to_device_state[8] = { 
//  zombie            zom suspend       
USB_STATE_POWERED, USB_STATE_SUSPENDED, 
//  default           default sus
USB_STATE_DEFAULT, USB_STATE_SUSPENDED,
//  addr              addr sus         
USB_STATE_ADDRESS, USB_STATE_SUSPENDED, 
//  config            config sus
USB_STATE_CONFIGURED, USB_STATE_SUSPENDED
};

static char * state_names[8] =
{ "zombie", "zombie suspended", 
  "default", "default suspended",
  "address", "address suspended", 
  "configured", "config suspended"
};

static char * event_names[6] =
{ "suspend", "reset", "resume",
  "address assigned", "configure", "de-configure"
};

static char * device_state_names[] =
{ "not attached", "attached", "powered", "default",
  "address", "configured", "suspended" };

static int sm_state = kStateZombie;

//////////////////////////////////////////////////////////////////////////////
// Async
//////////////////////////////////////////////////////////////////////////////

/* The UDCCR reg contains mask and interrupt status bits,
 * so using '|=' isn't safe as it may ack an interrupt.
 */

void u2d_set_mask_U2DCR( int mask )
{
	U2DCR = (U2DCR & U2DCR_MASK_BITS) | (mask & U2DCR_MASK_BITS);
}

void u2d_clear_mask_U2DCR( int mask)
{
	U2DCR = (U2DCR & U2DCR_MASK_BITS) & ~(mask & U2DCR_MASK_BITS);
}

void
u2d_int_hndlr(int irq, void *dev_id)
{
	__u32 isr = U2DISR;
	__u32 icr = U2DICR;
	__u32 dma_int = U2DMAINT;
	int wait = 0;

	if (usb_debug) printf("%s U2DISR=%08x, U2DICR=%08x\n",
				pszMe, isr, icr);

	if ( !isr  && !dma_int) return;

	//printf("%s U2DISR=%08x, U2DINT=%08x\n",	pszMe, isr, dma_int);
	//U2DICR = 0x0;

	/* endpoint A, B DMA interrupt */
	if(dma_int & 0x06)	{
		U2DMAINT = dma_int & 0x06;
		if(dma_int & 0x2) {
			ep1_int_hndlr(0);
		}
			
		if(dma_int & 0x4) {
			//printf("%s int %x\n", pszMe, dma_int);

			ep2_int_hndlr(0);
		}
		
	}




	/* InterRupt SUSpend */
	if ( isr & U2DINT_SU )
	{
		/* FIXME */
		/* udc_ack_int_UDCCR( UDCCR_SUSIR); */
		U2DISR = U2DINT_SU;
		if (usb_debug) printf("%sSuspend...\n", pszMe);
		usbctl_next_state_on_event( kEvSuspend );
	}

	/* InterRupt ResUme */
	if ( isr & U2DINT_RU )
	{
		/* FIXME */
		/* udc_ack_int_UDCCR( UDCCR_RESIR); */
		U2DISR = U2DINT_RU;
		if( usb_debug) printf("%sResume...\n", pszMe);
		usbctl_next_state_on_event( kEvResume );
		usb_connected = 0;
	}

	/* InterRupt Configuration Changed */
	if (isr & U2DINT_CC) {
		if (usb_debug) printf("%sConfiguration Changed\n", pszMe);

		if(usb_set_conf){
			usb_set_intf = 1;
			usb_set_conf = 0;
		}else{
			usb_set_intf = 0;
			usb_set_conf = 1;
		}

		if(usb_set_intf){

		usb_connected = 1;

		/* Reset usb driver here */
		if ( usbctl_next_state_on_event(kEvReset) != kError ) {
			check_usb_speed();

			ep0_reset();
			ep1_reset();
			ep2_reset();
			
			usbctl_next_state_on_event(kEvConfig);

			usb_driver_reset();
		}

		
		}

		U2DISR = U2DINT_CC;
		U2DCR = (U2DCR & (U2DCR_UDE)) | U2DCR_SMAC | U2DCR_CC;	


	}

	/* ReSeT Interrupt Request - UDC has been reset */
	if (isr & U2DINT_RS) {
		if(usb_debug) printf("%sReset...\n", pszMe);
		U2DISR = U2DINT_RS;

		usb_connected = 0;
		usb_driver_reset();
	}
	else
	{	/* FIFO err */
		if ( isr & U2DINT_FIFOERR ) U2DISR = U2DINT_FIFOERR;
		/* ep0 int */
		if ( isr & (U2DINT_PACKETCMP | U2DINT_SPACKETCMP)) {
			if (usb_debug) printf("%sEP0...\n", pszMe);

			U2DISR = U2DINT_PACKETCMP | U2DINT_SPACKETCMP;
			ep0_int_hndlr((isr & (U2DINT_PACKETCMP | U2DINT_SPACKETCMP)));
		}

		if ( isr & (U2DINT_FIFOERR << 3) ) {
			U2DCSRA = U2DCSR_TRN;
			U2DISR = U2DINT_FIFOERR << 3;
		}
		
		/* transmit bulk */
		if ( isr & ((U2DINT_PACKETCMP | U2DINT_SPACKETCMP) << 3) ) {
			if (usb_debug) printf("%sEP1...\n", pszMe);

			U2DCSRA = isr & ((U2DINT_PACKETCMP | U2DINT_SPACKETCMP) << 3);
			ep1_int_hndlr(isr);
		}

		if ( isr & (U2DINT_FIFOERR << 6) )  {
			U2DCSRB = U2DCSR_TRN;
			U2DISR = U2DINT_FIFOERR << 6;
		}

#if 0
		/* receive bulk */
		if (isr & ((U2DINT_PACKETCMP | U2DINT_SPACKETCMP) << 6)) {
			if (usb_debug) printf("%sEP2...\n", pszMe);

			//printf("%s isr %x\n", pszMe, isr);
			U2DISR = isr & ((U2DINT_PACKETCMP | U2DINT_SPACKETCMP) << 6);
			ep2_int_hndlr(isr);
		}

		//while (U2DCSRB & U2DCSR_BNE) {
		//	ep2_int_hndlr(isr );
		//}
#endif
	}



	U2DICR = 0x9e000043;//icr;
	
}

void check_usb_speed(void)
{

		/* set usb speed */
		if(U2DCR & U2DCR_HS) usb_speed = USB_SPEED_HIGH;
		else usb_speed = USB_SPEED_FULL;
		//printf("change speed, U2DCR %x, usb_speed %d\n", U2DCR, usb_speed);

		usb_driver_speed(usb_speed);
		
		//u2d_configure(usb_speed, 1);
		u2d_configure(usb_speed, 2);

}

//////////////////////////////////////////////////////////////////////////////
// Public Interface
//////////////////////////////////////////////////////////////////////////////

/* Open PXA usb core on behalf of a client, but don't start running */

int
pxa_usb_open( const char * client )
{
	if (usb_debug) printf( "pxa_usb_open\n");
	usbd_info.client_name = (char*) client;
	memset(&usbd_info.stats, 0, sizeof(struct usb_stats_t));
	memset(string_desc_array, 0, sizeof(string_desc_array));

	/* hack to start in zombie suspended state */
	sm_state = kStateZombieSuspend;
	usbd_info.state = USB_STATE_SUSPENDED;

	/* create descriptors for enumeration */
	initialize_descriptors();

	ep1_init(0);
	if (usb_debug) printf( "%s%s registered.\n", pszMe, client );
	return 0;
}

/* Start running. Must have called usb_open (above) first */
int
pxa_usb_start( void )
{
	if (usb_debug) printf( "pxa_usb_start\n");

	if ( usbd_info.client_name == NULL ) {
		printf( "%s%s - no client registered\n",
				pszMe, __FUNCTION__ );
		return -EPERM;
	}

	//xcvr_init();
	/* start UDC internal machinery running 
	 * move to check_usb_connection, to only enable U2D
	 * when data transfer is desired 
	 */
	//u2d_enable();
	//msleep(1);

#ifdef MMU_ENABLE
	dma_desc_buf = (XLLP_DMAC_DESCRIPTOR_T *)DMA_AREA_START;
	ep1_buf = ((unsigned *)dma_desc_buf + DMA_DESC_NUM * sizeof(XLLP_DMAC_DESCRIPTOR_T));
	ep2_buf = ((unsigned *)ep1_buf + sizeof(DMA_BUF_SIZE));
#endif
	/* give endpoint notification we are starting */
	ep1_state_change_notify( USB_STATE_SUSPENDED );
	ep2_state_change_notify( USB_STATE_SUSPENDED );

	/* alloc the dma desc buffer for ep1, ep2 */
	ep1_dma_desc = ((unsigned int)(&(dma_desc_buf[0])+1) & 0xfffffff0);
	ep2_dma_desc = ep1_dma_desc + DMA_DESC_NUM/2;
	/* enable any platform specific hardware */
	//soft_connect_hook( 1 );

	//u2d_info();

	if (usb_debug) printf( "%sStarted %s\n", pszMe, usbd_info.client_name );
	return 0;
}

/* Stop USB core from running */
int
pxa_usb_stop( void )
{
	if ( usbd_info.client_name == NULL ) {
		printf( "%s%s - no client registered\n",
				pszMe, __FUNCTION__ );
		return -EPERM;
	}

	ep1_reset();
	ep2_reset();

	u2d_disable();
	if( usb_debug) printf( "%sStopped %s\n", pszMe, usbd_info.client_name );
	return 0;
}

/* Tell PXA core client is through using it */
int
pxa_usb_close( void )
{
	 if ( usbd_info.client_name == NULL ) {
		   printf( "%s%s - no client registered\n",
				  pszMe, __FUNCTION__ );
		  return -EPERM;
	 }
	 printf( "%s%s closed.\n", pszMe, (char*)usbd_info.client_name );
	 usbd_info.client_name = NULL;
	 return 0;
}

/* set a proc to be called when device is configured */
usb_notify_t pxa_set_configured_callback( usb_notify_t func )
{
	 usb_notify_t retval = configured_callback;
	 configured_callback = func;
	 return retval;
}

/*====================================================
 * Descriptor Manipulation.
 * Use these between open() and start() above to setup
 * the descriptors for your device.
 *
 */

/* get pointer to static default descriptor */
desc_t *
pxa_usb_get_descriptor_ptr( void ) { return &desc; }

/* optional: set a string descriptor */
int
pxa_usb_set_string_descriptor( int i, string_desc_t * p )
{
	 int retval;
	 if ( i < MAX_STRING_DESC ) {
		  string_desc_array[i] = p;
		  retval = 0;
	 } else {
		  retval = -EINVAL;
	 }
	 return retval;
}

/* optional: get a previously set string descriptor */
string_desc_t *
pxa_usb_get_string_descriptor( int i )
{
	 return ( i < MAX_STRING_DESC )
		    ? string_desc_array[i]
		    : NULL;
}

config_desc_t *
pxa_usb_get_config(int cfgval) 
{
	int i;
	desc_t * pdesc = pxa_usb_get_descriptor_ptr();
	config_desc_t *cfg = (config_desc_t*) (pdesc->cdb);

	for( i=0; i<pdesc->dev.bNumConfigurations; i++) {
		if( cfg->bConfigurationValue == cfgval ) return cfg;
		cfg = (config_desc_t*) ((unsigned char*)cfg + cfg->wTotalLength);
	}

	return NULL;
}

intf_desc_t *
pxa_usb_get_interface( config_desc_t *cfg, int idx)
{
	int i;
	intf_desc_t *intf = (intf_desc_t*) (cfg + 1);
	
	for( i=0; i < cfg->bNumInterfaces; i++) {
		if( idx == intf->bInterfaceNumber) return intf;
		intf++;
	}

	return NULL;
}


ep_desc_t *
pxa_usb_get_endpoint( intf_desc_t *intf, int idx)
{
	int i;
	ep_desc_t *ep = (ep_desc_t *) (intf+1);


	for( i=0; i< intf->bNumEndpoints; i++) {
		if( idx == (ep->bEndpointAddress & 0xF) ) return ep;
		ep++;
	}
	return NULL;
}

//////////////////////////////////////////////////////////////////////////////
// Exports to rest of driver
//////////////////////////////////////////////////////////////////////////////

/* called by the int handler here and the two endpoint files when interesting
   .."events" happen */

int
usbctl_next_state_on_event( int event )
{
	int next_state = device_state_machine[ sm_state ][ event ];
	if ( next_state != kError )
	{
		int next_device_state = sm_state_to_device_state[ next_state ];
		if (usb_debug) printf( "%s%s --> [%s] --> %s. Device in %s state.\n",
				pszMe, state_names[ sm_state ], event_names[ event ],
				state_names[ next_state ], device_state_names[ next_device_state ] );

		sm_state = next_state;
		if ( usbd_info.state != next_device_state )
		{
			if ( configured_callback != NULL
				 &&
				 next_device_state == USB_STATE_CONFIGURED
				 &&
				 usbd_info.state != USB_STATE_SUSPENDED
			   ) {
			  configured_callback();
			}
			usbd_info.state = next_device_state;

			ep1_state_change_notify( next_device_state );
			ep2_state_change_notify( next_device_state );
		}
	}
	else
		printf( "%s%s --> [%s] --> ??? is an error.\n",
				pszMe, state_names[ sm_state ], event_names[ event ] );
	return next_state;
}

//////////////////////////////////////////////////////////////////////////////
// Private Helpers
//////////////////////////////////////////////////////////////////////////////

/* setup default descriptors */

static void
initialize_descriptors(void)
{

	desc.dev.bLength               = sizeof( device_desc_t );
	desc.dev.bDescriptorType       = USB_DESC_DEVICE;
	desc.dev.bcdUSB                = 0x100; /* 1.0 */	// HS 0x200 ?????
	desc.dev.bDeviceClass          = 0xFF;	/* vendor specific */
	desc.dev.bDeviceSubClass       = 0;
	desc.dev.bDeviceProtocol       = 0;
	desc.dev.bMaxPacketSize0       = 64;	/* ep0 max fifo size, 16 orig ????? */
	desc.dev.idVendor              = 0;	/* vendor ID undefined */
	desc.dev.idProduct             = 0; 	/* product */
	desc.dev.bcdDevice             = 0; 	/* vendor assigned device release num */
	desc.dev.iManufacturer         = 0;	/* index of manufacturer string */
	desc.dev.iProduct              = 0; 	/* index of product description string */
	desc.dev.iSerialNumber         = 0;	/* index of string holding product s/n */
	desc.dev.bNumConfigurations    = 1;	/* configurations we have */

	desc.qua.bLength               = sizeof( qua_desc_t );
	desc.qua.bDescriptorType       = USB_DESC_QUALIFIER;
	desc.qua.bcdUSB                = 0x0200;
	desc.qua.bDeviceClass          = 0xFF;	/* vendor specific */
	desc.qua.bDeviceSubClass       = 0;
	desc.qua.bDeviceProtocol       = 0;
	desc.qua.bMaxPacketSize0       = 64;	/* ep0 max fifo size, 16 or 64 ????? */
	desc.qua.bNumConfigurations    = 1;
	desc.qua.bRESERVED             = 0;

	/*
	desc.b.cfg.bLength             = sizeof( config_desc_t );
	desc.b.cfg.bDescriptorType     = USB_DESC_CONFIG;
	desc.b.cfg.wTotalLength        = make_word_c( sizeof(struct cdb) );
	desc.b.cfg.bNumInterfaces      = 1;
	desc.b.cfg.bConfigurationValue = 1;
	desc.b.cfg.iConfiguration      = 0;
	desc.b.cfg.bmAttributes        = USB_CONFIG_BUSPOWERED;
	desc.b.cfg.MaxPower            = USB_POWER( 500 );

	desc.b.intf.bLength            = sizeof( intf_desc_t );
	desc.b.intf.bDescriptorType    = USB_DESC_INTERFACE;
	desc.b.intf.bInterfaceNumber   = 0; 
	desc.b.intf.bAlternateSetting  = 0;
	desc.b.intf.bNumEndpoints      = 2;
	desc.b.intf.bInterfaceClass    = 0xFF; 
	desc.b.intf.bInterfaceSubClass = 0;
	desc.b.intf.bInterfaceProtocol = 0;
	desc.b.intf.iInterface         = 0;
	*/

/*
 * FIXME...
 * The host usbnet driver expects EP1=out EP2=in. On the PXA UDC EP1=in, EP2=out
 */
	/*
	desc.b.ep1.bLength             = sizeof( ep_desc_t );
	desc.b.ep1.bDescriptorType     = USB_DESC_ENDPOINT;
	desc.b.ep1.bEndpointAddress    = USB_EP_ADDRESS( 1, USB_IN );
	desc.b.ep1.bmAttributes        = USB_EP_BULK;
	desc.b.ep1.wMaxPacketSize      = make_word_c( 64 );
	desc.b.ep1.bInterval           = 0;

	desc.b.ep2.bLength             = sizeof( ep_desc_t );
	desc.b.ep2.bDescriptorType     = USB_DESC_ENDPOINT;
	desc.b.ep2.bEndpointAddress    = USB_EP_ADDRESS( 2, USB_OUT );
	desc.b.ep2.bmAttributes        = USB_EP_BULK;
	desc.b.ep2.wMaxPacketSize      = make_word_c( 64 );
	desc.b.ep2.bInterval           = 0;
	*/

// FIXME: Add support for all endpoint...
	usb_driver_reset();

	/* set language */
	/* See: http://www.usb.org/developers/data/USB_LANGIDs.pdf */
	sd_zero.bDescriptorType = USB_DESC_STRING;
	sd_zero.bLength         = sizeof( string_desc_t );
	sd_zero.bString[0]      = make_word_c( 0x409 ); /* American English */
	pxa_usb_set_string_descriptor( 0, &sd_zero );
}

/* soft_connect_hook()
 * Some devices have platform-specific circuitry to make USB
 * not seem to be plugged in, even when it is. This allows
 * software to control when a device 'appears' on the USB bus
 * (after Linux has booted and this driver has loaded, for
 * example). If you have such a circuit, control it here.
 */
static void
soft_connect_hook( int enable )
{
	if (enable) {
		U2DCR &= ~U2DCR_ADD;
		msleep(6);
	} else {
		U2DCR |= U2DCR_ADD;
		msleep(6);
	}
}

/* configure udc endpoints */
static void
u2d_configure(int speed, int phase)
{
	P_XLLP_U2D_EP_CONFIG ep_config;

	/* endpoint A~ BULK, IN, MPS: 64B(FS)/512B (HS),
	 * configuration 1, interface 0, altsetting 0, endpoint 1 */
	ep_config = &xllp_u2d.ConfigTable[1];

	if(phase == 1)	{
		ep_config->ET = USB_ENDPOINT_XFER_BULK;
		ep_config->ED = 1;
		ep_config->EN = 1;
		
		ep_config->CN = 1;
		ep_config->InterfaceNumber = 0;
		ep_config->AISN = 0;

		ep_config->HBW = 0;
		//ep_config->bufferSize = (BULK_FIFO_SIZE % 8)?1:0;
		//ep_config->bufferSize += (BULK_FIFO_SIZE / 8);
		ep_config->bufferSize = BULK_FIFO_SIZE;

		//U2DEPCR(i) = (ep_config->bufferSize << 1 )|U2DEPCR_EE;

		//XllpU2DConfigureOneEndpoint(&dev->xllp_u2d); not actual config
		XllpU2DEndpointSetConfig(&xllp_u2d, 1, ep_config);
	}else if(phase == 2){
		ep_config->MPS = ((speed)==USB_SPEED_HIGH)?((unsigned)512):((unsigned)64);
		XllpU2DEndpointSetInfo(&xllp_u2d, 1, ep_config);
	}

	/* endpoint B~ BULK, OUT, MPS: 64B(FS)/512B (HS),
	 * configuration 1, interface 0, altsetting 0, endpoint 1 */
	ep_config = &xllp_u2d.ConfigTable[2];

	if(phase == 1)	{
		ep_config->ET = USB_ENDPOINT_XFER_BULK;
		ep_config->ED = 0;
		ep_config->EN = 2;
		
		ep_config->CN = 1;
		ep_config->InterfaceNumber = 0;
		ep_config->AISN = 0;

		ep_config->HBW = 0;	// 
		//ep_config->bufferSize = (BULK_FIFO_SIZE % 8)?1:0;
		//ep_config->bufferSize += (BULK_FIFO_SIZE / 8);
		ep_config->bufferSize = BULK_FIFO_SIZE;

		//U2DEPCR(i) = (ep_config->bufferSize << 1 )|U2DEPCR_EE;

		//XllpU2DConfigureOneEndpoint(&dev->xllp_u2d); not actual config
		XllpU2DEndpointSetConfig(&xllp_u2d, 2, ep_config);
	}else if(phase == 2){
		ep_config->MPS = ((speed)==USB_SPEED_HIGH)?((unsigned)512):((unsigned)64);
		XllpU2DEndpointSetInfo(&xllp_u2d, 2, ep_config);
	}


#if 0
	if( speed == USB_SPEED_HIGH ){
		/* endpoint A ~ BULK, IN, MPS: 64bytes, Double-buffering, configuration 1, interface 0,
		 * altsetting 0, endpoint 1
		 */
		U2DEPCRA = 0x0200d103;
		U2DENA = ;

		/* endpoint B ~ BULK, OUT,MPS: 64bytes, Double-buffering, configuration 1, interface 0,
		 * altsetting 0, endpoint 2
		 */
		U2DEPCRB = 0x02014103;
		U2DENB = ;
	} else {

		
	}
#endif
}


/* disable the UDC at the source */
static void
u2d_disable(void)
{
	soft_connect_hook( 0 );
	/* clear UDC-enable */
	//udc_clear_mask_UDCCR( UDCCR_UDE); 

	U2DICR = 0x0;
	U2DISR = 0xfeffffff;
#ifdef CONFIG_CPU_MONAHANS_LV
	//U2DICR2 = 0x0;
	//U2DISR2 = 0xfeffffff;
	U2DOTGCR &= ~U2DOTGCR_ULE;
	U2DOTGICR = 0x0;
	U2DOTGISR = 0x00037f7f;
#elif defined(CONFIG_CPU_TAVOR_P)
	U2DOTGCR &= ~U2DOTGCR_ULE;
	U2DOTGCR &= ~U2DOTGCR_IESI;
#endif
	XllpU2DDisable(&xllp_u2d);

#if defined(CONFIG_CPU_MONAHANS_LV) || defined(CONFIG_CPU_TAVOR_P)
	U2DOTGCR |= U2DOTGCR_UTMID;
#endif

    	/* Disable clock for USB device */
	//*(unsigned long*)0x4134000c &= ~0x100000;
	XllpClockEnable(CLK_BASE, XLLP_CLK_USB2, XLLP_FALSE);
}

/*-------------------------------------------------------------------------*/
#if defined(CONFIG_CPU_MONAHANS_LV) || defined(CONFIG_CPU_TAVOR_P)

static int ulpi_reg_read(u32 reg, u8 *value)
{	int i=10000000;

	U2DOTGUCR = U2DOTGUCR_RUN | U2DOTGUCR_RNW | (reg<<U2DOTGUCR_ADDR_S);
	
	while((U2DOTGUCR & U2DOTGUCR_RUN) && i--);

	if(i<=0){
		printf("Read ULPI register Time out, reg %x\n", reg);
		return -1;
	}
	
	*value=(u8)(U2DOTGUCR & U2DOTGUCR_RDATA);

	return 0;
	
}

static int ulpi_reg_write(u32 reg, u8 value)
{	int i=10000000;

	U2DOTGUCR = U2DOTGUCR_RUN | (reg<<U2DOTGUCR_ADDR_S) | (value<<U2DOTGUCR_WDATA_S);
	
	while((U2DOTGUCR & U2DOTGUCR_RUN) && i--);

	if(i<=0){
		printf("Write ULPI register Time out, reg %x val %x\n", reg, value);
		return -1;
	}

	return 0;
}

static void 
xcvr_init(void)
{	u8 val;
	u32 count=2000;

	/* reset the transceiver */
	ulpi_reg_read(ULPI_FUNCTION_CONTROL, &val);
	val |= ULPI_FC_RESET;
	ulpi_reg_write(ULPI_FUNCTION_CONTROL_SET, val);
	do{
		ulpi_reg_read(ULPI_FUNCTION_CONTROL, &val);
	} while((count--) && (val & ULPI_FC_RESET));
}
#else
static void xcvr_init(void){}
#endif

static void u2d_soft_dis(int enable)
{
#if 0
#ifdef xjCONFIG_CPU_MONAHANS_LV
	if (enable) {
		/* enable D+/D- pulldown resistor */
		ulpi_reg_write(ULPI_OTG_CONTROL_SET, ULPI_OC_DPPULLDOWN);
		ulpi_reg_write(ULPI_OTG_CONTROL_SET, ULPI_OC_DMPULLDOWN);
	} else {
		/* disable D+/D- pulldown resistor */
		ulpi_reg_write(ULPI_OTG_CONTROL_CLEAR, ULPI_OC_DPPULLDOWN);
		ulpi_reg_write(ULPI_OTG_CONTROL_CLEAR, ULPI_OC_DMPULLDOWN);
	}
#else
	if (enable) {
		U2DCR |= U2DCR_ADD;
		msleep(6);
	} else {
		U2DCR &= ~U2DCR_ADD;
		msleep(6);
	}
#endif
#endif
}

/*  enable the udc at the source */
static void
u2d_enable(void)
{
	ASCR &=0x7fffffff;
    /* Enable clock for USB device */
	//CKENA |= 0x40;
	//CKENB |= 0x200;
	
	XllpU2DInitHandle(&xllp_u2d, (P_XLLP_U2D_REGISTERS_T)0x54100000, (P_XLLP_VUINT32_T)0x40E10000, (P_XLLP_GPIO_T)0x40E00000, (P_XLLP_MFP_RM_DB_ID_T)0x54100304);
		       	//U2D_BASE, MFP_BASE, GPIO_BASE, XLLP_MFP_RM_DATABASE);
	/* Bit 39 is HSIO2 clock. */
	OSCC |= 0x800;
	XllpClockEnable(CLK_BASE, 32+9, XLLP_TRUE);
	XllpClockEnable(CLK_BASE, XLLP_CLK_USB2, XLLP_TRUE);
	XllpU2DClearAllConfigure(&xllp_u2d);

	/* configure the MFP */
	XllpU2DPlatformInit(&xllp_u2d);
	
	/* U2D Endpoints */
	u2d_configure(usb_speed, 1);

	/* enable endpoint 0, A, B's Packet Complete Interrupt. */
	U2DICR = U2DINT_CC | U2DINT_RS | U2DINT_RU | U2DINT_SU | U2DINT_DPE |  \
		(U2DINT_SPACKETCMP | U2DINT_PACKETCMP /*| U2DINT_FIFOERR*/);

	/* clear the interrupt status/control registers */
	U2DISR = 0xffffffff;
	
	/* set UDC-enable */
	//u2d_set_mask_U2DCR( U2DCR_UDE); 
	U2DEN0 = 0x02000000;
	
	u2d_soft_dis(0);

#ifdef CONFIG_CPU_MONAHANS_LV	
	U2DOTGCR = U2DOTGCR_ULAF|U2DOTGCR_ULE;
#elif defined(CONFIG_CPU_TAVOR_P)
	U2DOTGCR = U2DOTGCR_ULE;
#endif

	//U2DCR |= ((7<<U2DCR_HSTC_S)&U2DCR_HSTC_MASK) | ((7<<U2DCR_FSTC_S)&U2DCR_FSTC_MASK);
	XllpU2DDisable(&xllp_u2d);
	XllpU2DEnable(&xllp_u2d);
	//XllpU2DEnable(&xllp_u2d);

	if (U2DCR & U2DCR_EMCE) {
		printf("%s:Endpoint Memory Config Error\n", __FUNCTION__);
	}

	if( (U2DCR & U2DCR_UDA) == 0) {
		/* There's a reset on the bus,
		 * clear the interrupt bit and keep going	?????
		 */
		if (usb_debug) printf("reset on bus\n");
		/* udc_ack_int_UDCCR( UDCCR_RSTIR); */
	}
}

void u2d_start(void)
{
	u2d_enable();
	soft_connect_hook(1);
}
