/**********************************************************************
 * pcmcia framework
 *
 * Provides an raw pcmcia/cf framework for blob. Based on original
 * work from Brad Parker.
 *
 * Copyright (C) 2001, Brad Parker (brad@heeltoe.com)
 * Copyright (C) 2002, Stefan Eletzhofer <stefan.eletzhofer@eletztrick.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: pcmcia.c,v 1.8 2003/08/20 23:12:22 seletz Exp $
 *
 */

/**********************************************************************
 * Includes
 */
#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/types.h>
#include <blob/errno.h>
#include <blob/util.h>
#include <blob/command.h>
#include <blob/time.h>
#include <blob/serial.h>
#include <blob/arch.h>
#include <blob/pcmcia.h>

#ifdef SA1111_BASE
#include <blob/proc/sa1111.h>
#endif

/**********************************************************************
 * Defines / Makros
 */
#define WEAK_SYM __attribute__ (( weak ))

#define MEMb(adr)	(*((u8 *)(adr)))

#define PCMCIA_DEBUG 1

#ifdef PCMCIA_DEBUG
#	define DBG( x, args... )	if ( pcmcia_dbg>x ) printf( args );
#else
#	define DBG( x, args... )	
#endif

/**********************************************************************
 * types
 */
typedef struct {
	u16				status;
	int				vcc;
	int				vpp;
	u32				attr_base;
	u32				base;
	volatile u8	*ident;
	volatile u8	*feature_p[PCMCIA_MAX_FEATURES];
	int				n_features;
	u16				config_base;
} pcmcia_slot_t;

/**********************************************************************
 * program globals
 */

pcmcia_slot_t pcmcia_slots[ PCMCIA_MAX_SLOTS ] = {
	{
		status:		0,
		vcc:		0,
		vpp:		0,
		attr_base:	PCMCIA_S0_ATTR_BASE,
		base:		PCMCIA_S0_BASE,
	},{
		status:		0,
		vcc:		0,
		vpp:		0,
		attr_base:	PCMCIA_S1_ATTR_BASE,
		base:		PCMCIA_S1_BASE,
	},
};


/**********************************************************************
 * module globals
 */
#ident "$Id: pcmcia.c,v 1.8 2003/08/20 23:12:22 seletz Exp $"

#ifdef	PCMCIA_DEBUG
static int pcmcia_dbg = 1;
#else
static int pcmcia_dbg = 0;
#endif

/**********************************************************************
 * prototypes
 */
static int pcmcia_print_fixed(volatile u8 *p);
static int pcmcia_print_funcid(int func);

/**********************************************************************
 * exported functions
 */

void pcmcia_dbg_set( int level )
{
	pcmcia_dbg = level;
}

/* do whatever is needed to get pcmcia initialized */
int WEAK_SYM pcmcia_init()
{
	return 0;
}

/* set pcmcia slot voltage levels */
int WEAK_SYM pcmcia_voltage_set( int slot, int vcc, int vpp )
{
	return 0;
}

/* detect slot */
int WEAK_SYM pcmcia_slot_detect( int slot )
{
	int	detect = 0;
#ifdef SA1111_BASE
	unsigned int flags = PCSR;

	switch ( slot ) {
		case 0:
			detect = (flags & PCSR_S0_DETECT) == 0;
			break;
		case 1:
			detect = (flags & PCSR_S1_DETECT) == 0;
			break;
		default:
			detect = -1;
			goto DONE;
			break;
	}

	if ( detect ) {
		pcmcia_slots[slot].status |= PCMCIA_SS_DETECT;
	} else {
		pcmcia_slots[slot].status &= ~PCMCIA_SS_DETECT;
	}

	DBG( 1, "%s: slot %d status: %x\n", __FUNCTION__, slot, pcmcia_slots[slot].status );
DONE:
#else
#warning "FIXME: non sa1111 systems not yet supported"
#endif
	return detect;
}

/* enable slot */
int WEAK_SYM pcmcia_slot_enable( int slot )
{
	int ret = 0;

#ifdef SA1111_BASE
	switch ( slot ) {
		case 0:
			PCCR = (PCCR & ~PCCR_S0_PSE) | PCCR_S0_FLT | PCCR_S0_PWAITEN;
			break;
		case 1:
			PCCR = (PCCR & ~PCCR_S1_PSE) | PCCR_S1_FLT | PCCR_S1_PWAITEN;
			break;
		default:
			ret = -1;
			goto DONE;
			break;
	}

	pcmcia_slots[slot].status |= PCMCIA_SS_ENABLED;

	DBG( 1, "%s: slot %d status: %x\n", __FUNCTION__, slot, pcmcia_slots[slot].status );
DONE:
#else
#warning "FIXME: non sa1111 systems not yet supported"
#endif
	return ret;
}

/* disable slot */
int WEAK_SYM pcmcia_slot_disable( int slot )
{
	int ret = 0;

#ifdef SA1111_BASE
	switch ( slot ) {
		case 0:
			PCCR = PCCR | PCCR_S0_PSE;
			break;
		case 1:
			PCCR = PCCR | PCCR_S1_PSE;
			break;
		default:
			ret = -1;
			goto DONE;
			break;
	}

	pcmcia_slots[slot].status &= ~PCMCIA_SS_ENABLED;

	DBG( 1, "%s: slot %d status: %x\n", __FUNCTION__, slot, pcmcia_slots[slot].status );
DONE:
#else
#warning "FIXME: non sa1111 systems not yet supported"
#endif
	return ret;
}

/* reset pcmcia slot */
int WEAK_SYM pcmcia_slot_reset(int slot)
{
	int ret = 0;

#ifdef SA1111_BASE
	switch ( slot ) {
		case 0:
			PCCR |= PCCR_S0_RST;
			msleep(20);
			PCCR &= ~PCCR_S0_RST;
			msleep(20);
			break;
		case 1:
			PCCR |= PCCR_S1_RST;
			msleep(20);
			PCCR &= ~PCCR_S1_RST;
			msleep(20);
			break;
		default:
			ret = -1;
			break;
	}

	DBG( 1, "%s: slot %d reset\n", __FUNCTION__, slot );
#else
#warning "FIXME: non sa1111 systems not yet supported"
#endif
	return ret;
}

/* dump slot HW line stati */
int pcmcia_dump_stati( void )
{
#ifdef SA1111_BASE
	int status = PCSR;

	printf("s0: ");
	if (status & PCSR_S0_READY) printf("ready "); else printf("      ");
	if (status & PCSR_S0_DETECT) printf("detect "); else printf("       ");
	if (status & PCSR_S0_VS1) printf("vs1 "); else printf("    ");
	if (status & PCSR_S0_VS2) printf("vs2 "); else printf("    ");
	if (status & PCSR_S0_WP) printf("wp "); else printf("   ");
	if (status & PCSR_S0_BVD1) printf("bvd1 "); else printf("     ");
	if (status & PCSR_S0_BVD2) printf("bvd2 "); else printf("     ");
	printf(";  ");

	printf("s1: ");
	if (status & PCSR_S1_READY) printf("ready "); else printf("      ");
	if (status & PCSR_S1_DETECT) printf("detect "); else printf("       ");
	if (status & PCSR_S1_VS1) printf("vs1 "); else printf("    ");
	if (status & PCSR_S1_VS2) printf("vs2 "); else printf("    ");
	if (status & PCSR_S1_WP) printf("wp "); else printf("   ");
	if (status & PCSR_S1_BVD1) printf("bvd1 "); else printf("     ");
	if (status & PCSR_S1_BVD2) printf("bvd2 "); else printf("     ");
	printf("\n");
#else
#warning "FIXME: non sa1111 systems not yet supported"
#endif

	return 0;
}

/* identify pcmcia card */
int pcmcia_identify( volatile unsigned char *p )
{
	unsigned char id_str[PCMCIA_MAX_IDENT_CHARS];
	unsigned char data;
	unsigned char *t;
	int i, done;

	if (p == NULL)
		return (0);     /* Don't know */

	t = id_str;
	done =0;

	for (i=0; i<=4 && !done; ++i, p+=2) {
		while ((data = *p) != '\0') {
			if (data == 0xFF) {
				done = 1;
				break;
			}
			*t++ = data;
			if (t == &id_str[PCMCIA_MAX_IDENT_CHARS-1]) {
				done = 1;
				break;
			}
			p += 2;
		}
		if (!done)
			*t++ = ' ';
	}
	*t = '\0';
	while (--t > id_str) {
		if (*t == ' ')
			*t = '\0';
		else
			break;
	}
	DBG( 1, "%s: id_str='%s'\n", __FUNCTION__, id_str);

	return (1);     /* don't know */
}

/* parse cis tuples of card in slot <slot> */
int pcmcia_cis_parse( int slot )
{
	int	ret					= 0;
	void *cfg_mem_addr		= NULL;
	volatile u8 *ident	= NULL;
	volatile u8 *p		= NULL;
	volatile u8 *start	= NULL;
	int n_features			= 0;
	u8 func_id			= ~0;
	u8 len				= 0;
	u8 code				= 0;
	u16 config_base			= 0;
	int found				= 0;
	int i					= 0;
	volatile u8 **feature_p;

	/* sanity check */
	if ( slot < 0 ) {
		printf( "pcmcia: invalid slot slot %d\n", slot );
		ret = -1;
		goto DONE;
	}
	if ( slot > PCMCIA_MAX_SLOTS )  {
		printf( "pcmcia: invalid slot slot %d\n", slot );
		ret = -1;
		goto DONE;
	}

	if ( !( pcmcia_slots[slot].status & PCMCIA_SS_ENABLED ) ||
			!( pcmcia_slots[slot].status & PCMCIA_SS_DETECT )) {
		printf( "pcmcia: slot %d not detected and enabled\n", slot );
		ret = -1;
		goto DONE;
	}

	cfg_mem_addr = (void *)pcmcia_slots[slot].attr_base;
	feature_p = pcmcia_slots[slot].feature_p;
	ident = pcmcia_slots[slot].ident;


	DBG( 1, "%s: PCMCIA MEM: %p\n", __FUNCTION__, cfg_mem_addr);

	start = p = (volatile u8 *)cfg_mem_addr;

	while ((p - start) < PCMCIA_MAX_TUPEL_SZ) {

		code = *p; p += 2;

		if (code == 0xFF) { /* End of chain */
			break;
		}
		len = *p; p += 2;

#if PCMCIA_DEBUG
		if (pcmcia_dbg > 10) {
			volatile u8 *q = p;
			printf("\nTuple code %2x  length %d\n\tData:",
					code, len);

			for (i = 0; i < len; ++i) {
				printf (" %2x", *q);
				q+= 2;
			}
		}
#endif

		switch (code) {
			case CISTPL_VERS_1:
				ident = p + 4;
				DBG( 1, "%s: ident=%p\n", __FUNCTION__, ident );
				break;
			case CISTPL_FUNCID:
				func_id = *p;
				DBG( 1, "%s: func_id=%d\n", __FUNCTION__, func_id );
				break;
			case CISTPL_FUNCE:
				if (n_features < PCMCIA_MAX_FEATURES)
					feature_p[n_features++] = p;
				break;
			case CISTPL_CONFIG:
				config_base = (*(p+6) << 8) + (*(p+4));
				pcmcia_slots[slot].config_base = config_base;
				DBG( 1, "%s: config_base=0x%04x\n", __FUNCTION__, config_base );
			default:
				break;
		}
		p += 2 * len;
	}

	found = pcmcia_identify( ident );

	if (func_id != ((u8)~0)) {
		pcmcia_print_funcid(func_id);

		if (func_id == CISTPL_FUNCID_FIXED)
			found = 1;
	}

	if ( found ) {
		for (i=0; i<n_features; ++i) {
			pcmcia_print_fixed (feature_p[i]);
		}
	}

	ret = 0;
DONE:
	return ret;
}

int pcmcia_slot_address_get( int slot, u32 *base, u32 *attr )
{
	int ret = 0;

	/* sanity check */
	if ( slot < 0 ) {
		printf( "pcmcia: invalid slot slot %d\n", slot );
		ret = -1;
		goto DONE;
	}
	if ( slot > PCMCIA_MAX_SLOTS )  {
		printf( "pcmcia: invalid slot slot %d\n", slot );
		ret = -1;
		goto DONE;
	}

	if ( !( pcmcia_slots[slot].status & PCMCIA_SS_ENABLED ) ||
			!( pcmcia_slots[slot].status & PCMCIA_SS_DETECT )) {
		printf( "pcmcia: slot %d not detected and enabled\n", slot );
		ret = -1;
		goto DONE;
	}

	DBG( 1, "%s: base=%p\n", __FUNCTION__, (void *)pcmcia_slots[slot].base);
	DBG( 1, "%s: attr_base=%p\n", __FUNCTION__, (void *)pcmcia_slots[slot].attr_base);

	if ( base )	*base = pcmcia_slots[slot].base;
	if ( attr )	*attr = pcmcia_slots[slot].attr_base;

	ret = 0;
DONE:
	return ret;
}

int pcmcia_slot_cfg_reg_get( int slot, u16 *cfg_base )
{
	int ret = 0;

	/* sanity check */
	if ( slot < 0 ) {
		printf( "pcmcia: invalid slot slot %d\n", slot );
		ret = -1;
		goto DONE;
	}
	if ( slot > PCMCIA_MAX_SLOTS )  {
		printf( "pcmcia: invalid slot slot %d\n", slot );
		ret = -1;
		goto DONE;
	}

	if ( !( pcmcia_slots[slot].status & PCMCIA_SS_ENABLED ) ||
			!( pcmcia_slots[slot].status & PCMCIA_SS_DETECT )) {
		printf( "pcmcia: slot %d not detected and enabled\n", slot );
		ret = -1;
		goto DONE;
	}

	DBG( 1, "%s: cfg_base=0x%03x\n", __FUNCTION__, pcmcia_slots[slot].config_base);

	if ( cfg_base )	*cfg_base = pcmcia_slots[slot].config_base;

	ret = 0;
DONE:
	return ret;
}

int pcmcia_slot_attr_read( int slot, u16 offset, u8 *value )
{
	int ret = 0;

	/* sanity check */
	if ( slot < 0 ) {
		printf( "pcmcia: invalid slot slot %d\n", slot );
		ret = -1;
		goto DONE;
	}
	if ( slot > PCMCIA_MAX_SLOTS )  {
		printf( "pcmcia: invalid slot slot %d\n", slot );
		ret = -1;
		goto DONE;
	}
	if ( !value ) {
		printf( "pcmcia: value==NULL\n" );
		ret = -1;
		goto DONE;
	}

	if ( !( pcmcia_slots[slot].status & PCMCIA_SS_ENABLED ) ||
			!( pcmcia_slots[slot].status & PCMCIA_SS_DETECT )) {
		printf( "pcmcia: slot %d not detected and enabled\n", slot );
		ret = -1;
		goto DONE;
	}

	*value = MEMb(
			pcmcia_slots[slot].attr_base +
			pcmcia_slots[slot].config_base +
			offset
			);

	DBG( 1, "%s: ATTR_REG(0x%02x) => 0x%02x\n", __FUNCTION__,
			offset, *value );

	ret = 0;
DONE:
	return ret;
}

int pcmcia_slot_attr_write( int slot, u16 offset, u8 value )
{
	int ret = 0;

	/* sanity check */
	if ( slot < 0 ) {
		printf( "pcmcia: invalid slot slot %d\n", slot );
		ret = -1;
		goto DONE;
	}
	if ( slot > PCMCIA_MAX_SLOTS )  {
		printf( "pcmcia: invalid slot slot %d\n", slot );
		ret = -1;
		goto DONE;
	}

	if ( !( pcmcia_slots[slot].status & PCMCIA_SS_ENABLED ) ||
			!( pcmcia_slots[slot].status & PCMCIA_SS_DETECT )) {
		printf( "pcmcia: slot %d not detected and enabled\n", slot );
		ret = -1;
		goto DONE;
	}

	DBG( 1, "%s: ATTR_REG(0x%02x) <= 0x%02x\n", __FUNCTION__,
			offset, value );

	MEMb( pcmcia_slots[slot].attr_base +
			pcmcia_slots[slot].config_base +
			offset
			) = value;

	ret = 0;
DONE:
	return ret;
}

/**********************************************************************
 * static functions
 */

/* print function id strings */
static int pcmcia_print_funcid(int func)
{
	printf("\t     ");
	switch (func) {
		case CISTPL_FUNCID_MULTI:
			printf("Multi-Function");
			break;
		case CISTPL_FUNCID_MEMORY:
			printf("Memory");
			break;
		case CISTPL_FUNCID_SERIAL:
			printf("Serial Port");
			break;
		case CISTPL_FUNCID_PARALLEL:
			printf("Parallel Port");
			break;
		case CISTPL_FUNCID_FIXED:
			printf("Fixed Disk");
			break;
		case CISTPL_FUNCID_VIDEO:
			printf("Video Adapter");
			break;
		case CISTPL_FUNCID_NETWORK:
			printf("Network Adapter");
			break;
		case CISTPL_FUNCID_AIMS:
			printf("AIMS Card");
			break;
		case CISTPL_FUNCID_SCSI:
			printf("SCSI Adapter");
			break;
		default:
			printf("Unknown");
			break;
	}
	printf(" Card\n");

	return 0;
}

/* print info for FIXED_DISK function type card */
static int pcmcia_print_fixed(volatile u8 *p)
{
	if (p == NULL)
		return -1;

	printf("\t    ");

	switch (*p) {
		case CISTPL_FUNCE_IDE_IFACE:
			{   u8 iface = *(p+2);

				printf((iface == CISTPL_IDE_INTERFACE) ? " IDE" : " unknown");
				printf(" interface ");
				break;
			}
		case CISTPL_FUNCE_IDE_MASTER:
		case CISTPL_FUNCE_IDE_SLAVE:
			{   u8 f1 = *(p+2);
				u8 f2 = *(p+4);

				printf((f1 & CISTPL_IDE_SILICON) ? " [silicon]" : " [rotating]");

				if (f1 & CISTPL_IDE_UNIQUE)
					printf(" [unique]");

				printf((f1 & CISTPL_IDE_DUAL) ? " [dual]" : " [single]");

				if (f2 & CISTPL_IDE_HAS_SLEEP)
					printf(" [sleep]");

				if (f2 & CISTPL_IDE_HAS_STANDBY)
					printf(" [standby]");

				if (f2 & CISTPL_IDE_HAS_IDLE)
					printf(" [idle]");

				if (f2 & CISTPL_IDE_LOW_POWER)
					printf(" [low power]");

				if (f2 & CISTPL_IDE_REG_INHIBIT)
					printf(" [reg inhibit]");

				if (f2 & CISTPL_IDE_HAS_INDEX)
					printf(" [index]");

				if (f2 & CISTPL_IDE_IOIS16)
					printf(" [IOis16]");

				break;
			}
	}
	printf("\n");

	return 0;
}

