/*
 * serial-s3c2500.c: Samsung s3c2500 serial port driver
 *
 * Copyright (C) 2002  Erik Mouw <J.A.K.Mouw@its.tudelft.nl>
 * Copyright (C) 2003  Russ Dill <Russ.Dill@asu.edu>
 *
 * $Id: serial-s3c2500.c,v 1.1 2003/09/04 18:03:10 russd Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ident "$Id: serial-s3c2500.c,v 1.1 2003/09/04 18:03:10 russd Exp $"

#ifdef HAVE_CONFIG_H
# include <blob/config.h>
#endif

#include <blob/arch.h>
#include <blob/errno.h>
#include <blob/arch.h>
#include <blob/serial.h>
#include <blob/types.h>
#include <blob/util.h>

static void s3c2500_serial_idle_queue(void);


/* This convaluted ring thingy is necessary because we have a poll
 * based serial driver, with a one byte fifo. Spinning on the transmit
 * fifo makes it really easy to lose incoming characters */

/* somewhere to put stuff while we are spinning on a send */
#define RING_SIZE 128
int receive_ring[RING_SIZE]; /* int, not char, so we can store errors */
int head, tail, error_in_poll;
static int ring_overflow;


/* return -1 on full, 0 otherwise */
static int ring_push(int c)
{
	if (tail == (head + 1) % RING_SIZE) {
		ring_overflow = 1;
		return -ESERIAL;
	}

	/* push it into the ring if not */
	receive_ring[head] = c;
	head = (head + 1) % RING_SIZE;

	return 0;
}


/* only call this if head != tail */
static int ring_pull(void)
{
	int rv = receive_ring[tail];

	tail = (tail + 1) % RING_SIZE;

	/* handle the ring overflow */
	if (tail == (head + 2) % RING_SIZE && ring_overflow) {
		ring_push(-ESERIAL);
		ring_overflow = 0;
	} else {
		/* see if there is a char to replace this one */
		s3c2500_serial_idle_queue();
	}

	return rv;
}


/* check if there is a character available to read. returns 1 if there
 * is a character available, 0 if not, and negative error number on
 * failure */
static int s3c2500_serial_real_poll(void)
{
	u32 volatile tmp = CUSTAT;

	/* check for errors */
	if (tmp & (CUSTAT_PER | CUSTAT_FER | CUSTAT_OER)) {
		CUSTAT = CUSTAT_PER | CUSTAT_FER | CUSTAT_OER;
		if (tmp & CUSTAT_PER) return -ESERIAL;
		else if (tmp & CUSTAT_FER) return -ESERIAL;
		else return -ESERIAL;
	} else if (tmp & CUSTAT_RDV) return 1;
	
	return 0;
}


/* First, check the ring, then, check the port */
static int s3c2500_serial_poll(void)
{
	if (head != tail) {
		if (receive_ring[tail] >= 0) return 1;
		else return ring_pull();
	}
	return s3c2500_serial_real_poll();
}


/* flush serial input queue. returns 0 on success or negative error
 * number otherwise
 */
static int s3c2500_serial_flush_input(void)
{
	volatile u32 tmp;

	/* kill the queue */
	head = tail;
	ring_overflow = 0;

	/* keep on reading as long as the receiver is not empty */
	while(((tmp = CUSTAT) & CUSTAT_RDV) | !(tmp & CUSTAT_RXIDLE)) {
		if(tmp & (CUSTAT_PER | CUSTAT_FER | CUSTAT_OER)) {
			CUSTAT = CUSTAT_PER | CUSTAT_FER | CUSTAT_OER;
			continue;
		}

		if (tmp & CUSTAT_RDV) tmp = CURXBUF;
	}

	return 0;
}


/* try to push something into the queue if we are just sitting idle */
static void s3c2500_serial_idle_queue(void)
{
	u32 tmp;
	switch (tmp = s3c2500_serial_real_poll()) {
	case 0: /* no character ready */
		break;
	case 1:
		/* read the character */
		ring_push(CURXBUF & 0xff);
		break;
	default:
		/* clear the error and record it */
		CUSTAT = CUSTAT_PER | CUSTAT_FER | CUSTAT_OER;
		error_in_poll = 1;
		ring_push(tmp);
	}
}


/* flush output queue. returns 0 on success or negative error number
 * otherwise
 */
static int s3c2500_serial_flush_output(void)
{
	/* wait until the transmitter is no longer busy */
	while(!(CUSTAT & CUSTAT_TI)) s3c2500_serial_idle_queue();

	return 0;
}


/* initialise serial port at the request baudrate. returns 0 on
 * success, or a negative error number otherwise
 */
#define DIVISOR (29491200 / 16)
static int s3c2500_serial_init(serial_baud_t baud)
{
	head = tail = ring_overflow = error_in_poll = 0;

	if (baud >= baud_COUNT) return -ERANGE;

	/* Clear any errors */
	CUSTAT = CUSTAT_BSD | CUSTAT_FER | CUSTAT_PER | CUSTAT_OER | CUSTAT_CCD;

	/* flush it out completely */
	s3c2500_serial_flush_output();
	
	/* Disable interrupts */
	CUINT = 0;
	
	/* Set baud rate */
	CUBRD = ((DIVISOR / baud_rates[baud] - 1) << 4);

	/* turn the receiver and transmitter on */
	CUCON  = CUCON_8BITS | CUCON_SCSEL | CUCON_RMODE | CUCON_TMODE;

	return 0;
}


/* read one character from the serial port. return character (between
 * 0 and 255) on success, or negative error number on failure. this
 * function is blocking */
static int s3c2500_serial_read(void)
{
	u32 tmp;

	if (head != tail) return ring_pull();

	for(;;) {
		switch (tmp = s3c2500_serial_real_poll()) {
		case 0: break;
		case 1: return CURXBUF & 0xff;
		default:
			CUSTAT = CUSTAT_PER | CUSTAT_FER | CUSTAT_OER;
			return tmp;
		}
	}
}


/* write character to serial port. return 0 on success, or negative
 * error number on failure. this function is blocking
 */
static int s3c2500_serial_write(int c)
{
	/* wait for room in the transmit FIFO */
	do {
		s3c2500_serial_idle_queue();
	} while (!(CUSTAT & CUSTAT_THE));

	CUTXBUF = (u32)(c & 0xff);

	return 0;
}




/* export serial driver */
serial_driver_t s3c2500_serial_driver = {
	init:		s3c2500_serial_init,
	read:		s3c2500_serial_read,
	write:		s3c2500_serial_write,
	poll:		s3c2500_serial_poll,
	flush_input:	s3c2500_serial_flush_input,
	flush_output:	s3c2500_serial_flush_output
};
